/******************************************************************************** 
 * File:   Debug.h
 * Author: skidmarks
 *
 * Created on March 3, 2009, 11:02 AM
********************************************************************************/


//********************************************************************************
// PURPOSE:
//    Provide debug facilities.
//    
// DESCRIPTION:
//    The design allows incorporation of the debug toolset through inheritance
//    and by static calls. The debug output indents entries based on the 
//    dynamic textual level and outputs a text string. The textual level
//    is established through calls to DebugEnter and Debug Exit. All static
//    calls indent and output the supplied text. Class specialization uses
//    DData() and provides a means for class specific output to be generated
//    via <class>.DData().
//    
//    Debugging is controlled by a single, global, Debug Flag. The flag can
//    be set, reset, and interrogated during execution. This allows debug
//    to be controlled (on a global basis) and specific external conditions
//    used to control debug output.
//    
//    All debug is output to the Log stream.
// 
// ARCHITECTURE:
//    Publically accessible functions either modifiy, interrogate, or use
//    the global debug flag. Publically accessible output functions inhibit
//    output if the Debug flag is false.
//
//      Functions     Description
//      DebugEnter    Conditionally Output the user string and increment indent
//      DebugExit     Conditionally Decrement indent and output the user string
//      DebugData     Conditionally Output the user string
//      DEnter        Unconditionally Output the user string and increment indent
//      DExit         Unconditionally Decrement indent and output the user string
//      DData(string) Output user supplied data
//      DData()       Output class specific data
//      setFlag       Set the Debug Flag
//      FlipFlag      Toggle the Debug Flag
//      isDebug       Interrogate the Debug Flag
//    
// NOTES:
//    1. If cDebug is inherited then DData() must be constructed.
//********************************************************************************

#ifndef _DEBUG_H
#define	_DEBUG_H

# include "stdafx.h"

# include <algorithm>
# include <iostream>
# include <fstream>
# include <iomanip>
# include <string>
# include "global.h"

using namespace std;

class cDebug {
private:
    static bool         DebugFlag;               // Debug On/Off
    static long         Indent;                  // Indent debug output
    static bool         ReadyFlag;               // Debug initialized
    static const long   DelIndent;               // Indent increment
    static const long   MaxIndent;               // Maximum indentation amount
    static const string Blanks;                  // Blank line
    
    static void DEnter(const string& text) { string str("Enter " + text); DData(str);  Indent += DelIndent; };
    static void DExit (const string& text) { string str("Exit  " + text); Indent -= DelIndent; DData(str);  };
protected:
    static void DData (const string& text) { Log << setw(max(long(0), min(Indent, MaxIndent))) << "" << text << endl;
                                             Log.flush(); }
    virtual
    void DData () = 0;
public:
    cDebug()                       {  };
    ~cDebug()                      {  }
    static bool setFlag(bool flag) { bool f = DebugFlag; DebugFlag = flag; return f; }
    static bool flipFlag()         { bool f = DebugFlag; DebugFlag = !DebugFlag; return f; }
    static void DebugEnter(const string& text) { if ( is_Debug() ) DEnter(text); }
    static void DebugExit (const string& text) { if ( is_Debug() ) DExit (text); }
    static void DebugData (const string& text) { if ( is_Debug() ) DData (text); }
    
    static void DebugEnter(const char text[])  { if ( is_Debug() ) { string str(text); DEnter(str); }}
    static void DebugExit (const char text[])  { if ( is_Debug() ) { string str(text); DExit (str); }}
    static void DebugData (const char text[])  { if ( is_Debug() ) { string str(text); DData (str); }}
    
    void DebugData ()                    { if ( is_Debug() ) DData ();     }
    
    static bool is_Debug()                 { return DebugFlag; }
}; // cDebug

#endif	/* _DEBUG_H */

