/***********************************************************************\
 *
 * File:           RngStream.h for multiple streams of Random Numbers
 * Language:       C++ (ISO 1998)
 * Copyright:      Pierre L'Ecuyer, University of Montreal
 * Notice:         This code can be used freely for personal, academic,
 *                 or non-commercial purposes. For commercial purposes, 
 *                 please contact P. L'Ecuyer at: lecuyer@iro.umontreal.ca
 * Date:           14 August 2001
 *
\***********************************************************************/
#ifndef _RNGSTREAM_H
#define	_RNGSTREAM_H

# include <string>

using namespace std;

class RngStream
{
public:
  RngStream (const char *name = "");
  //  This constructor creates a new stream with (optional) descriptor name. 
  //  It initializes its seed Ig, and sets Bg and Cg to Ig. It also sets its 
  //  anti and incPrec switches to false. The seed Ig is 13 equal to the 
  //  initial seed of the package if this is the first stream created; otherwise
  //  it is Z steps ahead of the seed of the most recently created stream.

  static bool SetPackageSeed (const unsigned long seed[6]);
  //  Sets the initial seed s0 of the package to the six integers in the vector
  //  seed. The first 3 integers in the seed must all be less than 
  //  m1 = 4294967087, and not all 0; and the last 3 integers must all be less 
  //  than m2 = 4294944443, and not all 0. If this method is not called, the 
  //  default initial seed is (12345, 12345, 12345, 12345, 12345, 12345). Returns
  //  false for invalid seeds, and true otherwise.

  void ResetStartStream ();
  //  Reinitializes the stream to its initial state: Cg and Bg are set to Ig.

  void ResetStartSubstream ();
  //  Reinitializes the stream to the beginning of its current substream: Cg is 
  //  set to Bg.

  void ResetNextSubstream ();
  //  Reinitializes the stream to the beginning of its next substream: Ng is 
  //  omputed, and Cg and Bg are set to Ng.

  void SetAntithetic (bool a);
  //  If a = true, the stream will start generating antithetic variates, 
  //  i.e., 1 - U instead of U, until this method is called again with a = false.

  void IncreasedPrecis (bool incp);
  //  After calling this method with incp = true, each call to the generator 
  //  (direct or indirect) for this stream will return a uniform random number 
  //  with more bits of resolution (53 bits if machine follows IEEE 754 standard)
  //  instead of 32 bits, and will advance the state of the stream by 2 steps 
  //  instead of 1. More precisely, if s is a stream of the class RngStream, 
  //  in the nonantithetic case, the instruction \u = s.RandU01()" will be 
  //  equivalent to \u = (s.RandU01() + s.RandU01() * fact) % 1.0" where the 
  //  constant fact is equal to 2-24. This also applies when calling RandU01 
  //  indirectly (e.g., via RandInt, etc.). By default, or if this method is 
  //  called again with incp = false, each call to RandU01 for this stream 
  //  advances the state by 1 step and returns a number with 32 bits of resolution.

  bool SetSeed (const unsigned long seed[6]);
  //  Sets the initial seed Ig of the stream to the vector seed. The vector seed 
  //  should contain valid seed values as described in SetPackageSeed. The state 
  //  of the stream is then reset to this initial seed. The states and seeds of 
  //  the other streams are not modified. As a result, after calling this method,
  //  the initial seeds of the streams are no longer spaced Z values apart. We 
  //  discourage the use of this method; proper use of the Reset* methods is 
  //  preferable. Returns false for invalid seeds, and true otherwise.

  void AdvanceState (long e, long c);
  //  Advances the state by n steps (see below for the meaning of n), without 
  //  modifying the states of other streams or the values of Bg and Ig in the 
  //  current object. If e > 0, then n = 2e + c; if e < 0, then n = -2-e +c; 
  //  and if e = 0, then n = c. Note: c is allowed to take negative values. We 
  //  discourage the use of this method.

  void GetState (unsigned long seed[6]) const;
  // Returns in seed[0..5] the current state Cg of this stream. This is convenient 
  //  if we want to 14 save the state for subsequent use.

  void WriteState () const;
  //  Writes (to standard output) the current state Cg of this stream.

  void WriteStateFull () const;
  //  Writes (to standard output) the value of all the internal variables of this 
  //  stream: name, anti, incPrec, Ig, Bg, Cg.

  double RandU01 ();
  //  Normally, returns a (pseudo)random number from the uniform distribution over 
  //  the interval (0; 1), after advancing the state by one step. The returned number 
  //  has 32 bits of precision in the sense that it is always a multiple of 1=(232-208). 
  //  However, if IncreasedPrecis(true) has been called for this stream, the state is 
  //  advanced by two steps and the returned number has 53 bits of precision.

  int RandInt (int i, int j);
  //  Returns a (pseudo)random number from the discrete uniform distribution over the 
  //  integers fi; i + 1; : : : ; jg. Makes one call to RandU01.
private:
  const string name;
  // User supplied Name of random number stream;
  
  double Cg[6], Bg[6], Ig[6];
  // Vectors to store the current seed, the beginning of the current block (substream) 
  //  and the beginning of the current stream.

  bool anti, incPrec;
  //  Variables to indicate whether to generate antithetic or increased precision 
  //  random numbers. std::string name; String to store the optional name of the 
  //  current RngStream object.

  static double nextSeed[6];
  //  Static vector to store the beginning state of the next RngStream to be created 
  //  (instantiated).

  double U01 ();
  //  The backbone uniform random number generator.

  double U01d ();
  //  The backbone uniform random number generator with increased precision.
};
#endif