/******************************************************************************** 
 * File:   cFixedPartition.h
 * Author: skidmarks
 *
 * Created on March 3, 2009, 11:02 AM
********************************************************************************/



//********************************************************************************
// PURPOSE:
//    Create partitions of fixed size.
//    
// DESCRIPTION:
//    A generated partition:
//    1. Is half-open, [), and
//    2. Contains a lower bound and upper bound, [low, high), and
//    3. Has zero or more associated input intervals, [low, high) <- 1, 2, ... .
//    4. Each input interval has bounds on a closed interval, [low, high].
//
//    A generated fized size partition which satisfies the following 
//    requirement:
//       Each referenced input interval is partially contained within the
//       generated partition. That is for each referenced input interval
//       [R.low <= G.high, R.high >= G.low)
//
//       where R is the referenced input interval, and
//             G is the generated partition
//    
//    The functions contained in the class generates ideal partitions. The 
//    algoritm makes a single pass through the input data. If the current
//    input interval bounds are within the range of the current partition
//    bounds, then the input interval is put onto the ambiguity stack. If
//    the input interval low bound exceeds the generated partition high
//    bound then the ambiguity stack populates the generated partitiion
//    list of references and intervals which terminate in the current
//    generated partition are removed. The current interval is pushed onto
//    the ambiguity stack and the next interval is generated.
//    
//    Total size estimate is: (N == number of input intervals)
//     O( N) input intervals
//     O(2N) sorted bounds of input intervals
//     O( N) ambiguity stack (high bounds of input intervals are not stored)
//   -------
//     O(4N)
//    
//    Total time estimate is: (N == number of input intervals)
//     O( N)      input input intervals
//     O(2N)      number of input bounds to process
//     O(2N lg2N) sort separated input bounds
//     O(N + EP*EE) ambiguity stack operations 
//                  where EP == the expected number of partitions, O(2N)
//                        EE == expected number of references / partition
//      = O(N + 2N*EE) = O(N(1 + 2*EE))
//
//    Total number of generated partitions :
//       (min(input interval low bouond) - max(input interval high bound)) / partition size
//
// ARCHITECTURE:
//    Class functions are:
//    
//    Function          Description
//    cIdealPartition   Constructor
//    ~cIdealPartition  Descructor
//    Binary            Output partitions
//    Create_Partitions Create Ideal Partition 
//    Run               Public interface and driver to process
//    Validate          Validate correctness of generated partitions
//    
// NOTES:
//    1. The description of the algorithm is approximate.
//********************************************************************************

# ifndef _FIXEDPARTITION_H
# define	_FIXEDPARTITION_H

# include "create.h"

class cFixedPartition : public cCreate {
private:
   void Create_Partitions(cBound_List& SortArray, cPartition_List& Partition, double P_Size);
   void PopDel(cPartition& Partition, cStack& Stack);
   long ValidatePartitions(cBound_List& SortArray, cPartition_List& Partition);
public:
   cFixedPartition(const char* Filename, const string Name, const double Bandwidth, cBound_List* Bounds) 
                  : cCreate(Filename, Name, Bandwidth, Bounds) { };
   ~cFixedPartition(void);
   virtual void Binary() { BinaryOutput(".fix"); }
   void Run(double FixedSize) {        
       DebugEnter("cCreate::RunFixed");
       long Size = (long)((Bandwidth + FixedSize - FixedSize/10.)/FixedSize);
       Partitions  = new  cPartition_List(Size, "FixedPartitions");
       Create_Partitions(*Bounds, *Partitions, FixedSize);
       DebugExit("cCreate::RunFixed");
    }
    void Validate() { ValidatePartitions(*Bounds, *Partitions); }
}; // class cFixedPartition
# endif
