/******************************************************************************** 
 * File:   cIdealPartition.h
 * Author: skidmarks
 *
 * Created on March 3, 2009, 11:02 AM
********************************************************************************/


//********************************************************************************
// PURPOSE:
//    Create partitions that contain no wasted space.
//    
// DESCRIPTION:
//    A generated partition:
//    1. Is half-open, [), and
//    2. Contains a lower bound and upper bound, [low, high), and
//    3. Has one or more associated input intervals, [low, high) <- 1, 2, ... .
//    4. Each input interval has bounds on a closed interval, [low, high].
//
//    A generated partition with no wasted space satisfies the following 
//    requirement:
//       Each referenced input interval is wholly contained within the
//       generated partition. That is for each referenced input interval
//       [R.low <= G.low, G.high <= R.high)
//
//       where R is the referenced input interval, and
//             G is the generated partition
//    
//    The functions contained in the class generates ideal partitions. The algorithm
//    foundation is based on the relation between a sorted list of input bound,
//    R.low and R.high. The sorted list is constructed by separating the input
//    interval bounds. If the input list size is N, then the separated list size
//    is 2N. The constructed list is sorted, a O(2N lg2N) operation, and the 
//    current bound is compared to the preceding bound on the list. Depending on
//    the relation between the two bounds, the current entry is put on a stack
//    (the ambiguity stack) of references which can be contained in the next
//    generated partition, or the current entry is used to delete a reference in
//    the ambiguity stack, or a new partition is constructed and a reference list
//    of entries in the ambiguity stack wholly contained within the partition is
//    constructed.
//    
//    Total size estimate is: (N == number of input intervals)
//     O( N) input intervals
//     O(2N) sorted bounds of input intervals
//     O( N) ambiguity stack (high bounds of input intervals are not stored)
//   -------
//     O(4N)
//    
//    Total time estimate is: (N == number of input intervals)
//     O( N)      input input intervals
//     O(2N)      number of input bounds to process
//     O(2N lg2N) sort separated input bounds
//     O(N + EP*EE) ambiguity stack operations 
//                  where EP == the expected number of partitions, O(2N)
//                        EE == expected number of references / partition
//      = O(N + 2N*EE) = O(N(1 + 2*EE))
//
//    Total number of generated partitions < 2N
//
// ARCHITECTURE:
//    Class functions are:
//    
//    Function          Description
//    cIdealPartition   Constructor
//    ~cIdealPartition  Destructor
//    Binary            Output partitions
//    Create_Partitions Create Ideal Partition 
//    Run               Public interface and driver to process
//    Validate          Validate correctness of generated partitions
//    
// NOTES:
//    
//********************************************************************************

# ifndef _IDEALPARTITION_H
# define _IDEALPARTITION_H

# include "create.h"

class cIdealPartition : public cCreate {
private:
   void Create_Partitions(cBound_List& SortArray, cPartition_List& Partition);
   long ValidatePartitions(cBound_List& SortArray, cPartition_List& Partition);
public:
   cIdealPartition(const char* Filename, const string Name, double Bandwidth, cBound_List* Bounds) 
                  : cCreate(Filename, Name, Bandwidth, Bounds) { };
   ~cIdealPartition(void);
   virtual void Binary() { BinaryOutput(".idl"); }
   void   Run() { 
      DebugEnter("cCreate::RunIdeal");
      Partitions  = new  cPartition_List((*Bounds).GetSize(), "IdealPartitions");
      Create_Partitions(*Bounds, *Partitions);
      DebugExit("cCreate::RunIdeal");
   }
    void Validate() { ValidatePartitions(*Bounds, *Partitions); }
};
# endif