/******************************************************************************** 
 * File:   common.h
 * Author: skidmarks
 *
 * Created on March 3, 2009, 12:01 PM
********************************************************************************/


//********************************************************************************
// PURPOSE:
//    A common set of classes used to support partitioning.
//
// DESCRIPTION:
//    A general collection of classes in support of partitioning. The classes
//    can be divided into lists (<class>_List), classes specific to interval
//    handling and manipulation, and general support.
//    
//    Interval Handling provides support for the remembering of Input Interval
//    References (cRange and cList) and the creation and maintenance of the
//    generated partitions, Output Intervals (cBound, cBound_List, cPartition and
//    cPartition_List). Within this realm of idealized possibilities:
//    
//      cBound         Provides the defiinition of one side of the Input Interval,
//                     either the LOW or the HIGH bound of an Input Interval. The
//                     separation of an Input Interval into it's LOW and HIGH bound
//                     is fundamental to processing.
//      cBound_List    A list of bounds. The list is sorted by <bound, type>. A
//                     single pass is made through this list to discover all the
//                     disjoint partitions, for Ideal Partition processing, and
//                     all the fixed sized partitions, for Fixed Partition
//                     processing.
//      cPartition     The definition of a partition. Each partition contains an
//                     upper bound and a lower bound [upper, lower), and a list
//                     of references to Input Intervals wholly or partially 
//                     contained in the generated partitions. For Ideal partitions,
//                     an Input Interval is wholly contained within a Generated
//                     Partition. For Fixed Size Partitions, an Input Interval
//                     may be partially contained within the Generated Partitino.
//      cPartition_List The list of Generated Partitions.At process completion, 
//                     all Generated Partitions are contained within this list.
//    
//    The support classes are:
//      cCell          A single, doubly linked, cell. The contents of the cell
//                     are heterogeneous, two types of entries are supported.
//                     All operations required to support manipulation of a
//                     single cell are supported included delettion and insertion
//                     of a single cell within a list of cells.
//      cStack         A stack implementation using doubly linked cells. The cost
//                     of use of a cell rather than a simple array is warranted
//                     given the random nature of insertion and deletion. In 
//                     similar fashion, the use of a doubly-linked list rather
//                     than it's spacewise more efficient singly-linked list is
//                     warranted given the issues involved with insertion and
//                     deletion.
//    
//
// ARCHITECTURE:
//    see above.
//    
// NOTES:
//    1. <class>_List is a list of <class>es. See cList for an exception.
//    2. All classes are derived from cDebug and provide support for 
//       <class>.DData() to output class based instance data.
//    3. All non-list classes have a friend (as we should all have) output
//       class. The friend is used to support <class>.DebugData().
//    4. <class>_List has an operator{} to allow array access to list
//       elements. Operator[] is bounds checked.
//    5. cStack has an operator[] function.
//    6. All lists and cStack have a GetSize function to return the list size.
//    7. A deliberate attempt is made to make class functions in-line.
//********************************************************************************

#ifndef _COMMON_H
#define	_COMMON_H

# include "stdafx.h"

# include <cFLoat>
# include <climits>
# include <cstdlib>
# include <iostream>
# include <iomanip>
# include <sstream>

# include "Debug.h"

using namespace std;

    //----------------------------------------
    //         CLASSES & STRUCTURES
    //----------------------------------------

  class cRange : public cDebug {                 // Input Interval and Reference
  protected:
    virtual
    void DData () { ostringstream stream(ios::out);
                    stream << *this;
                    cDebug::DData(stream.str());
                  }
  public:
      double      RLo;                           // Low  bound in range [RLo, RHi]
      double      RHi;                           // High bound in range [RLo, RHi]
      long        Datum;                         // User Datum

      cRange() : RLo(LDBL_MIN), RHi(-LDBL_MIN), Datum(LONG_MIN) { };
      cRange(double myRLo, double myRHi, long myDatum)
               : RLo(myRLo), RHi(myRHi), Datum(myDatum) { };
      long&   Info() { return Datum; }
      cRange& Init() { RLo = LDBL_MIN; RHi = -LDBL_MIN; Datum = LONG_MIN; return *this; }
      friend ostream& operator<<(ostream& stream, cRange& Range);
  }; // cRange
  
  class cCell : public cDebug {                  // Doubly Linked List Cell
  public:
     enum eType { eRange, eCell, eValue };       // Cell Types
     union Datum {
         cRange* Range;                          // <RLo, RHi, Datum>
         cCell*  Cell;                           // Cell pointer
         long    Value;                          // An integer value
     };
  private:
     cCell*  Left_Link;                          // Left Link
     cCell*  Right_Link;                         // Right Link
     Datum   Datum;                              // Cell datum
     eType   Type;                               // Cell Type
  protected:
      virtual
      void DData () { ostringstream stream(ios::out); 
                      stream << "<" << this 
                                    << ": <- "  << Left_Link 
                                    << " -> "   << Right_Link << "  ";
                      if (Type == eRange)     stream <<  *Datum.Range << ">";
                      else if (Type == eCell) stream << (Datum.Cell->Info()) << ">";
                      else                    stream << hex << "0x" << Datum.Value << ">";
                      DebugData(stream.str());
      }
  public:
      cCell()            : Left_Link(NULL), Right_Link(NULL), Type(eRange) { Datum.Range = NULL; };
      cCell&   Delete();
      cRange&  Info()     { return (cRange&)*Datum.Range; }
      long&    Value()    { return Datum.Value;           }
      cCell*   Cell()     { return Datum.Cell;            }
      cCell&   Init()     { Left_Link = Right_Link = NULL;
                            Datum.Range = NULL; Type = eRange; 
                            return *this;
                          }
      cCell&   InfoCell() { return (cCell&)*Datum.Cell; }
      cCell&   InLeft(cCell &Cell);
      cCell&   InRight(cCell &Cell);
      eType    isType()   { return Type; }
      cCell&   Left()     { return (cCell &)*Left_Link; }
      cCell&   Right()    { return (cCell &)*Right_Link;}
      void     SetType(eType myType) { Type = myType; }
      friend   ostream& operator<<(ostream& stream, cCell& Cell);
      friend   class cStack;
  }; // class cCell

  class cBound : public cDebug {                 // Single bound of range [Low, High]
  protected:
    void DData () { ostringstream stream(ios::out); 
                    stream << this;
                    cDebug::DData(stream.str());
    }
  public:
    enum      eBoundary { Low, High, Error };    // Boundary type (Low, High, Guard)
    static string const Boundary[];              // String equivalent for eBoundary
    
    double    Bound;                             // Single bound of range
    eBoundary Type;                              // Bound is low or high
    cRange*   Datum;                             // <RLo, RHo, Datum> Bound == RLo or RHi
  public:
    cBound() : Bound(-LDBL_MIN), Type(Error), Datum(NULL) { };
    cBound(double myBound, eBoundary myType, cRange* myDatum)
                                            : Bound(myBound)
                                            , Type(myType)
                                            , Datum(myDatum) { };
    cRange* Info() { return Datum; }
    void    Init() { Bound = -LDBL_MIN, Type = Error; Datum = NULL; }
    friend  ostream &operator<<(ostream &stream, cBound& Bound);
  }; // cBound

  class cBound_List : public cDebug {            // List of Input Intervals
  private:
    string const Name;
    long         Size;                           // Number of bounds in list
    cBound*      List;                           // List of Ranges
  protected:
    virtual
    void DData () { ostringstream stream(ios::out); 
                    stream << "<<" << Name << ">> Size =" << setw(6) << Size;
                    cDebug::DData(stream.str());
                    for(long i = 0; i < Size; i++) {
                      stream.flush();
                      stream << setw(6) << i << ": " << &List[i];
                      cDebug::DData(stream.str());
                    }
    }
  public:
     cBound_List(string myName) :  Size(0), Name(myName) { List = NULL; }
     cBound_List(long mySize, string myName) : Size((mySize > 0) ? mySize : 0)
                                             , Name(myName) {
        List = new cBound[Size + 1];
        for (long i = 0; i <= Size; i++) List[i].Init();
     }
     ~cBound_List() { delete[] List; }
     long GetSize() {  return Size;  }
     void Init(long mySize) {
        Size = (mySize > 0)? mySize: 0;
        if (!List) delete [] List;
        List = new cBound[Size + 1];
        for (long i = 0; i <= Size; i++) List[i].Init();
     }
     cBound& operator[](long Ndx) {
        if ((Ndx >= 0) && (Ndx < Size)) return List[Ndx];
        else                            Log << "INTERNAL ERROR: Index out of range for " << Name << ": " << Ndx << endl;
                                        exit(2048);
                                        return List[Size];
     }
   }; // cBound_List
     
  class cList : public cDebug {                  // List of Input Intervals
      long  const Size;                          // Number of items in list
      cRange**    List;                          // List
  public:
      cRange static ErrorItem;                   // Erroneous value
  protected:
      virtual
      void DData () { long i;
                      ostringstream stream(ios::out); 
                      stream << setw(9) << Size << ": ";
                      for(i = 0; i < min(Size, (long) 9); i++) 
                          stream << setw(10) << *List[i];
                      cDebug::DData(stream.str());
                      for(long j = i; i < Size; i = j) {
                          stream.flush();
                          stream << "           ";
                          for(j = i; j < min(Size, i + 9); j++) 
                              stream << setw(10) << *List[j];
                          cDebug::DData(stream.str());
                      }
  }
  public:
      cList(long mySize) : Size(mySize) { List = new cRange*[Size]; }
      cList(long mySize, cRange** myList) : Size(mySize), List(myList) { };
     ~cList()                           { delete[] List; }
     long GetSize()                     { return Size; }
     cRange& operator[](long Ndx) { 
        if ((Ndx >= 0) && (Ndx < Size)) return *List[Ndx];
        else                            Log << "INTERNAL ERROR: Index out of range for cList : " << Ndx << endl;
                                        exit(2049);
                                        return ErrorItem;
     }

  }; // cList

  class cPartition : public cDebug {             // [RLo, RHi), List
  private:
    static string const GlobalName;              // "Partition Record"
    string              Name;                    // Local Name
    
    string ConstructName(long myName)
                          { ostringstream stream(ios::out);
                            stream << myName;
                            Name = stream.str();
                            return Name;
                          };
  protected:
    virtual
    void DData () { ostringstream stream(ios::out); 
                    stream << *this;
                    cDebug::DData(stream.str());
    }
  public:
    double  RLo;                                 // Lower Range Bound
    double  RHi;                                 // Upper Range Bound
    cList*  List;                                // Input Ranges in this partition
    
    cPartition()           { Init(); }
    ~cPartition()          { if (List) delete [] List; }
    cPartition(double myLo) : RLo(myLo), RHi(-LDBL_MIN), Name("") { };
    cPartition(double myLo, long myName) : RLo(myLo), RHi(-LDBL_MIN)
                           { Name = ConstructName(myName).c_str();  };
    cPartition(double myLo, double myHi, long myName) : RLo(myLo), RHi(myHi)
                           { Name = ConstructName(myName).c_str();  };
    long GetSize()         { return List->GetSize();     }
    void Init()            { RLo = RHi = -LDBL_MIN;      }
    void Init(double myLo) { RLo = myLo; RHi = LDBL_MIN; }
    friend ostream& operator<<(ostream& stream, cPartition& Partition);
    friend ostream& operator<<(ostream& stream, cPartition* Partition);
    cList&  operator=(cList* myList) { return (cList &)*(List = myList); }
    cRange& operator[](long Ndx)     { return (*List)[Ndx]; }
  }; // cPartition
  
  class cPartition_List : public cDebug {
  private:
    string const Name;                           // Name of Partition
    long         Size;                           // Size of cPartition_List
    long         Count;                          // Number of active partitions
    cPartition** Partition;                      // Array of partitions
  protected:
    virtual
    void DData () { ostringstream stream(ios::out); 
                    stream << "<<" << Name << ">> Size =" << setw(6) << Size;
                    cDebug::DData(stream.str());
                    for(long i = 0; i < Size; i++ ) (*Partition[i]).DebugData();
    }
  public:
     cPartition_List(string myName) : Size(0), Name(myName) { };
     cPartition_List(long mySize, string myName) : Size(mySize), Name(myName), Count(0)
     {  Partition = new cPartition *[Size+1];
        for(long i = 0; i < Size + 1; i++ ) Partition[i] = NULL;
     }
     ~cPartition_List() { if (Partition) delete[] Partition; }
     long         GetSize()             { return Count; }
     void         Init(long mySize) {
        Partition = new cPartition *[Size+1];
        for(long i = 0; i < Size + 1; i++ ) Partition[i] = NULL;
     }
     void         New(double RLo, long Ndx) { Partition[Ndx] = new cPartition(RLo, Ndx); }
     void         New(double RLo, double RHi, long Ndx) { Partition[Ndx] = new cPartition(RLo, RHi, Ndx); }
     cPartition&  operator[](long Ndx)  { return *(Partition[Ndx]);  };
     long&        operator++()          { Count++; return Count; }
     long&        operator++(int)       { Count++; return Count; }
  }; // class cPartition_List

  class cStack : public cDebug {                 // List manager
  private:
      string const        Name;                  // Name
      long const          Size;                  // Number of list elements
      cCell              *Space;                 // List as an array
      cCell               AVSL;                  // Available space list
      cCell               Head;                  // Active List Head
      cCell&              NextCell;              // Used by stack iterator
      
      cCell& NUCELL();                           // Get cell from AVSL
  private:
     void    PrintList(string Name, cCell* Head);
     cCell&  Push();
  protected:
      virtual
      void DData () { ostringstream stream(ios::out); 
                      stream << Name << " Stack "
                             << ": <" << setw(4) << Head.Value()
                             << ", "  << setw(9) << AVSL.Value()
                             << ">";
                      cDebug::DData(stream.str());
                      PrintList(Name, &Head); };
  public:
      cStack(int Size, string myName);
     ~cStack()                    { delete[] Space; }
     cCell&  Delete(cCell& Cell);
     cCell&  Find(cRange& myInfo);
     cCell&  Find(cRange* myInfo) { return Find(*myInfo);   };
     cCell&  Find(cCell&  Cell) { return Find(Cell.Info()); }
     long    GetSize()          { return Head.Value(); }
     cCell&  Next()             { NextCell = NextCell.Right();
                                  return (&NextCell == &Head)? Space[Size]:  NextCell; }
     cCell&  Peek();
     cCell&  Peek(long Ndx);
     cCell&  Push(cCell&  Cell);
     cCell&  Push(cRange& myInfo);
     cCell&  Push(cRange* myInfo)  { return Push(*myInfo); };
     cCell&  Pop();
     void    PrintAVSL();
     void    PrintStack();
     void    PrintLists()       { if (is_Debug()) { PrintStack(); PrintAVSL(); } }
     cCell&  ResetNext()        { return Head; }
     cCell&  Top()              { return (&Head.Right() != &Head)? Head.Right(): Space[Size]; }
     cCell&  operator[](long Ndx){
        if ((Ndx >= 0) && (Ndx < Size)) return Space[Ndx];
        else                            Log << "INTERNAL ERROR: Index out of range for " << Name << ": " << Ndx << endl;
                                        exit(2050);
                                        return Space[Size];
     }
}; // cStack
  
    //----------------------------------------
    //               Streams
    //----------------------------------------
  
  ostream& operator<<(ostream& stream, cRange& Range);
  ostream& operator<<(ostream& stream, cRange* Range);
  ostream& operator<<(ostream& stream, cCell& Cell);
  ostream& operator<<(ostream& stream, cBound& Bound);
  ostream& operator<<(ostream& stream, cPartition& Partition);
  ostream& operator<<(ostream& stream, cPartition* Partition);

#endif	/* _COMMON_H */

