/******************************************************************************** 
 * File:   generate.h
 * Author: skidmarks
 *
 * Created on March 3, 2009, 11:14 AM
********************************************************************************/

//********************************************************************************
// PURPOSE:
//    Support for the generation of input intervals to the partitioning functions.
//
// DESCRIPTION:
//    Artifical range intervals are generated for input to the partitioning
//    functions. The ranges are generated based on input statistics. The
//    statistics can either be input via a file or a limited subset can be 
//    given as an option.
//    
//    A generate statistic is of the form:
//      [low, high, percent, width]
//    where
//      low     is the lowest number in the range
//      high    is the highest number in the range
//      percent is the percent of generated member in the range
//      width   is the expected width of any generated interval
//    
//    No member of a set of statistic ranges can overlap:
//      given any two members in the set of inputs, i,j:
//      [low(i), high(i)], [low(j), high(j)], i <> j, low <= high, then:
//      either: high(i) < low(j), or
//              high(j) < low(i)
//    
//    If statistics come from the command line then the generation interval
//    is continuous on [0, 10,0000] with the width required on input.
//    
//    Processing consists of the following steps:
//    1. The statistics are input and bounds (low, high), width and percent
//       are verified for correctness. 
//    2. The statistics are sorted by range and they are checked for overlap.
//    3. The statistics are sorted by percent and if the total percent
//       is less than 100%, the following is done:
//       a. The total range [min(low), max(high)] is considered as continuous.
//       a. All disjoint ranges missing from the continuum are created.
//       a. All missing widths are inserted into the statistic.
//       a. Missing percentages are inserted into the statistics.
//    4. The statistics are sorted by percent.
//    5. Partition intervals are generated. A pseudo-random number generator
//       (RNG) is  used throughout. Processing proceedes as:
//       a. The RNG is used to generate a percent between 0 .. 100.
//       a. The statistics are searched for a cumulative percent greater than
//          or equal the generated one and interval preceding this one selected.
//       a. The RNG is used to generate statistic low <= low <= statistic high.
//       a. The RNG is used to generate a number low <= high <= statistic high.
//       a. The Partition Interval [low, high] is output.
//
// ARCHITECTURE:
//    Two basic classes/structures are defined:
//    sGenerate Defines the Options passed from the command line
//    cGenerate Generate functions and internal classes
//       sData   Generation Statistic Intervals [[low, high], percent width]
//       cGenInt List of all Generation Statistic Intervals
//    
//    During input of the Generate Statistic File, each new Generate Statistic
//    creates and new instance of sData. This new instance is appended to 
//    cGenInt. The validated list of sData objects is used for all subsequent
//    processing.
//    
// NOTES:
//    1. A maximum of 512 distinct input and constructed Generation Statistic
//       Intervals are supported.
//    2. CompareData and ComparePercent are shallow functions designed to all
//       access to cGenerate variables and pass them to qsort interfacable
//       comparison functions.
//********************************************************************************


#ifndef _GENERATE_H
#define	_GENERATE_H

# include <cFLoat>
# include <fstream>
# include <iostream>
# include <sstream>
# include <stdlib.h>
# include <string>
# include "Debug.h"
# include "RngStream.h"
# include "utility.h"

struct sGenerate {                               // General structure for GenOptions
   long            Count;                        // Number of intervals to generate
   unsigned long   Seed;                         // Start seed
   double          Width;                        // Bandwidth of each interval
   sGenerate() : Count(0), Seed(12345), Width(0.0) { }
}; // sGenerate

class cGenerate : public cDebug {
private:
   struct sData : public cDebug {                // Interval ranges
   protected:
      virtual
      void   DData () { ostringstream stream(ios::out);
                        stream << "[[" << Flow  << ", " << FHigh  << "], "
                                       << Width << ", " << Percent << "]]";
                        DebugData(stream.str()); };
   public:
      double Flow;                               // Lower interval bound
      double FHigh;                              // Upper interval bound
      double Width;                              // Maximum width of generated intervals
      double Percent;                            // Percent of output in this range
      sData(double myLow, double myHigh, double myWidth, double myPercent) 
                                                      : Flow(myLow)
                                                      , FHigh(myHigh)
                                                      , Width(myWidth)
                                                      , Percent(myPercent)
                                                      { }
      sData() : Flow(-LDBL_MIN),FHigh(-LDBL_MIN), Width(-LDBL_MIN),Percent(-LDBL_MIN) { }
   }; // sData

   class cGenInt : public cDebug {               // Interval list for range generation
   private:
      long   MaxNdx;                             // High water mark
      const
      long   Size;                               // Maximum number of entries
      sData* Data;                               // Generation Interval Descriptions
   protected:
      virtual
      void   DData () { ostringstream stream(ios::out);
                        cDebug::DData("cGenInt");
                        for(long i = 0; i < MaxNdx; i++) {
                           stream << "  " << Data[i];
                           cDebug::DData(stream.str());
                        }
                      };
   public:
      cGenInt(long mySize) : MaxNdx(0), Size(mySize) { Data = new sData[mySize]; }
      ~cGenInt() { if (Data) delete [] Data; Data = NULL; }
      long GetSize() {  return MaxNdx;  }
      sData& operator=(sData& myData) { if (MaxNdx < Size) return Data[MaxNdx++] = myData;
                                    cout << "INTERNAL ERROR: Not enough room for new Data " << MaxNdx << endl;
                                    exit(0x101);
                                    return Data[0]; }
      sData& operator[](long Ndx) { if ((Ndx >=0) && (Ndx < MaxNdx)) return Data[Ndx];
                                    cout << "INTERNAL ERROR: GenData Index " << Ndx << endl;
                                    exit(0x100);
                                    return Data[0]; }
   }; // cGenInt

   const string Filename;                        // Output filename
   const string IntervalFile;                    // Input Interval File
   long        Count;                            // Number of items to generate
   sData       Default;                          // Default sData when no Interval File
   sData       DefaultStat;                      // Default InputStat when no Interval File
   sData       InputStat;                        // min(low) max(high), Total Width, total percent
   cGenInt     Statistic;                        // Statistic Intervals
   RngStream   Rng;                              // Pseudo-random number generator
   unsigned
   long        Seed[6];                          // Pseudo-random number generator seed
   double      Width;                            // Bandwidth for interval
private:
   bool   CheckInput();                          // Check for overlapping intervals
   bool   CreateIntervals();                     // Generate intervals for 100% coverage
   void   Generate();                            // Generate output ranges
   bool   Parse(string& ParseLine);              // Parse input line
   bool   ReadIntervals();                       // Read input file
   void   WriteHeader();                         // Output banner
protected:
   virtual
   void   DData () {  };
public:
   static int ComparesData(const void* One, const void* Two);
   static int ComparePercent(const void* One, const void* Two);
   cGenerate(sGenerate Options, sFiles& File);
   ~cGenerate(void);
   bool Run();
   friend ostream& operator<<(ostream& stream, sData& Data);
}; // class cGenerate
  
    //----------------------------------------
    //               Streams
    //----------------------------------------

ostream& operator<<(ostream& stream, cGenerate::sData& Data);

#endif