/******************************************************************************** 
 * File:   irange.h
 * Author: Arthur Schwarz
 *
 * Created on April 13, 2009, 1:23 PM
********************************************************************************/

//********************************************************************************
// PURPOSE:
//    Input the Partition Interval data.
//
// DESCRIPTION:
//    This is the antecedent operation to partitioning the input intervals. This
//    class inputs and validates all of the user provided or system generated
//    intervals, and saves the resultant values into a user accessible data 
//    structure. The overall system design requires that this class be retained
//    until partitioning (fixed or ideal) is complete. The output of input
//    processing is a flag, indicating that processing was successful, diagnostic
//    message, output to the Log file for invalid input, and the input values.
//    
//    Diagnostic checks are made at several levels:
//    1. Parse: Each input field must be correct.
//    2. Semantic: All inputs in the same interval must be correct.
//    
//    At the end of data input, the intervals (low, high) are separated into
//    two entries and tagged 
//       (bound, low )->(low, high, identifier) and 
//       (bound, high)->(low, high, identifier) and
//    sorted by bound and tag (bound, low) < (bound, high). This sorted result
//    is used by the partitioning algorithms.
//    
//    Processing allows for an unbounded number of input intervals (limited
//    only by memory constraints in the host computer). This is done by:
//    1. Allocated a fixed size in-memory buffer (nominally 10,000 intervals).
//    2. Flushing the buffer to a temporary file when full.
//    3. Termination processing:
//       a. If the buffer has not been flushed, continue, otherwise
//       b. Flush the buffer to a temporary file.
//       c. Allocate an in-memory buffer of sufficient size for all the data.
//       d. Input data from the temmporary file to the new buffer.
//    
//    The input interval file must be a line-oriented, ASCII text file. Each
//    line must contain a:
//      low bound, high bound, identifer
//    
//      with low and high bound floating point numbere
//           the identifier a integer
//           and the ',' separators required
//    
//    The bounds may be positive or negative but low bound <= high bound.
//    
// ARCHITECTURE:
//    sRange      Base structure for ranges. Contains the minimum, required
//                data for any range, [low, high]. Initializes the basic
//                data to a known bad value on use.
//    sRange_3    Partition Interval, [[low, high], identifier]. Initializes
//                data to a known bad value on use.
//    iRange_List Base class for input processing and partitioning preparation.
//       =        Update index, flush buffer and assign range.
//       []       Index range list.
//       Chkin    Check input Partition Interval for validity.
//       Create_Bounds 
//       DData    Output debug data to Log file.
//       Flush    Flush Partition Range Buffer.
//       GetErrorCount Number of input errors found.
//       GetBounds Return pointer to bounds array.
//       GetSize  Number of input Partition Intervals.
//       Next     Assign input Partition Interval to buffer, flush buffer if full.
//       Read     Input Partition Interval.
//
//    ciRange_3   Processing and definitions for required Partition Interval
//                input data and processing.
//       =        Update index, flush buffer and assign range.
//       Parse    Parse input Partition Interval.
//       Read     Input Partition Interval.
//    
// NOTES:
//    1. Read() is the entry point for the class functions.
//********************************************************************************

# ifndef _IRANGE_H
# define _IRANGE_H

# include "stdafx.h"
# include <climits>
# include <cmath>
# include <fstream>
# include <istream>
# include <ostream>
# include "common.h"
# include "Debug.h"

struct  sRange { 
   double  RLo;                                  // Low  range value
   double  RHi;                                  // High range value
   sRange() : RLo(-LDBL_MIN), RHi(-LDBL_MIN) { };
};

struct  sRange_3 : public sRange {
   long    Datum;                                // User supplied datum
   sRange_3() : sRange(), Datum(LONG_MIN) { };
};

class iRange_List : public cDebug {              // input ranges
private:
    // Processing of range buffer: <Ndx, Size, BUFFERSIZE, Ranges, TempStream>
    //  Ranges[Ndx] : Ndx<0..BUFFERSIZE-1>
    //  if (Ndx == BUFFERSIZE) TempStream.write(Ranges[Ndx]; Size += BUFFERSIZE; Ndx = 0;
    static
    long const   BUFFERSIZE;                     // Input buffer size
    cBound_List* Bounds;                         // <bound, type, &Range>
    cRange       ErrorRange;
    long         Ndx;                            // Index in Ranges buffer
    long         Size;                           // Total number ranges used
    fstream      TempStream;
protected:
    long         ErrorCount;                     // Number of errors found
    cRange*      Ranges;                         // All input ranges
    string       TempFileName;                   // Constructed tempporary file name
    static
    long const  ERRORCOUNTLIMIT;                 // Maximum I/O Errors allowed before termination
public:
protected:
    iRange_List(string Path);
    iRange_List(const iRange_List& orig);
    virtual ~iRange_List();
    virtual
    void   DData () { string str(80, ' ');
                      ostringstream stream(str);
                      stream << "iRange_List<"
                             << setw(9) << Ndx  << ", "
                             << setw(9) << Size << ", "
                             << setw(9) << ErrorCount << ">" << endl;
                      cDebug::DData(str);
                    };
    bool   ChkIn(cRange Range);
    void   Flush();
    long   GetErrorCount()               { return ErrorCount; }
    long   Next() { if ( ++Ndx < BUFFERSIZE) return Ndx;
                    if (!TempStream.is_open()) {
                      Log << "ERROR: Temporary File not open" << endl;
                      exit(0x2010);
                    }
                    TempStream.write((char*)Ranges, sizeof(Ranges));
                    Size += BUFFERSIZE;
                    return (Ndx = 0);
                  };
public:
    double Create_Bounds();
    long   GetSize()  { return Size; }
    cBound_List* GetBounds() { return Bounds; }
    virtual bool Read() = 0;
    virtual
    cRange& operator=(cRange& Range) { return (Ranges[Next()] = Range); }
    cRange& operator[](long Ndx) { return ((Ndx >=0 ) && (Ndx < Size))? Ranges[Ndx]: ErrorRange; }
}; // class iRange_List

class ciRange_3 : public iRange_List {           // <low, high, id>
private:
   bool   Parse(string& ParseLine);              // Parse input line
private:
    cRange& operator=(cRange& Range) { return (Ranges[Next()] = Range); }
protected:
public:
    ciRange_3(string Path) : iRange_List(Path) { }
    virtual bool Read();
}; // class ciRange_3


    //----------------------------------------
    //            Debug Streams
    //----------------------------------------

ostream& operator<<(ostream& stream, sRange& Range);
ostream& operator<<(ostream& stream, sRange_3& Range);

#endif	/* _IRANGE_H */
