/******************************************************************************** 
 * File:   option.h
 * Author: Arthur Schwarz
 *
 * Created on April 17, 2009, 10:48 AM
********************************************************************************/

//********************************************************************************
// PURPOSE:
//    Provide classes and function necessary to process command line options.
//
// DESCRIPTION:
//    Command line options are described as options defined in a list of options.
//    The Option class (cOption) contains the fundamental properties of each
//    options, The Option List (cOption_List) contains a list of all command line
//    options. Processing consists in identifying each command in the command line
//    with it's associated option by searching the option list for a match.
//    
//    The options fundamental properties consists of the options:
//      long name   '--<long name>'
//      short name  '-<letter>'
//      function    parsing and processing function
//      value       string containing the command line text recognized as an option
//      void flag   set if an option value has been detected
//
//    The option list contains functions necessary to process the long and short
//    form inputs and retrieve the input options. The driver for option recognition
//    and value retrieval is 'commandline'.
//
// ARCHITECTURE:
//    Each <class>_List contains members of <class>.
//    Each <class>_List contains a GetSize funcion returning the number of 
//        items of type <class> in it.
//    
//    cOption contains option specific fields. The option processing function
//        'Function' is a non-class function created in the main driver.
//    
//    cOption_List contains a list of cOption items, the driver function,
//        (CommandLine), and routines to process the long and short form
//        option versions (GetLong, GetShort), and the retrieve an option
//        (GetOpt).
//
// NOTES:
//    
//********************************************************************************


#ifndef _OPTION_H
#define	_OPTION_H

# include "stdafx.h"

# include <sstream>
# include "Debug.h"

class cOption : public cDebug {                  // description of argument
private:
   char  const empty= '\0';
   
protected:
   virtual
   void DData () { ostringstream stream(ios::out);
                   stream << *this;
                   DebugData(stream.str());
   }
public:
   typedef long (*OptionFun)(cOption& Option);
   enum eFlag { eOptional    = 0                 // Optional option w/no arguments
              , eArgument    = 1                 // Argument allowed (=0 argument not allowed)
              , eArgument_Req= 2                 // Argument required (-x arg)
              , eRequired    = 4                 // Option required
              };
   char  const  ShortName;                       // [-]<letter>
   char  const* LongName;                        // [-]'-'<string>
   bool         Empty;                           // argument found (false)
   long  const  Flag;                            // type of option/argument
   OptionFun    Function;                        // Option handling function
   char  const* Val;                             // string representing argument value
   cOption() : ShortName('\0'), LongName(&empty), Empty(false), Flag(0), Val(&empty) { };
   cOption(char myShort, char* myLong, long myFlag, OptionFun myFunction)
                                      : ShortName(myShort)
                                      , LongName(myLong)
                                      , Empty(true)
                                      , Flag(myFlag)
                                      , Function(myFunction)
                                      , Val(NULL) { }
   long   ChkArg(char* Argument, long& Ndx, bool& ArgFlag);
   bool   ParseOption(string Str, long& Count, string* Value);
   friend  ostream& operator<<(ostream& stream, cOption& Option);
}; // cOption

class cOption_List : public cDebug {
private:
    static  cOption ErrorOption;                 // Faulty return
    cOption**  Option;                           // List of options
    long const Size;                             // Number of items in the list
protected:
      virtual
      void DData () { ostringstream stream(ios::out);
                      stream << "Option List";
                      DebugData(stream.str());
                      for(long i = 0; i < Size; i++) {
                          stream.flush();
                          stream << "      " << Option[i];
                          DebugData(stream.str());
                      }
                    }
public:
    cOption_List(long const mySize, cOption* myOption) : Size(mySize) {
        Option = new cOption*[Size];
        for(long i = 0; i < Size; i++) Option[i] = &myOption[i];
    }
    ~cOption_List() { delete [] Option; }
    bool     CommandLine(long const Argc, char* Argv[], string& Filename);
    bool     ChkRequired();
    long     GetLong(char  Argv[], char* Argument, long& Ndx);
    long     GetOpts(long const Argc, char* Argv[], string* List[]);
    long     GetSize() { return Size; }
    long     GetShrt(char  Argv[], char* Argument, long& Ndx);
    cOption& operator[](long Ndx) { return ((Ndx < 0) || (Ndx >= Size))? ErrorOption: *Option[Ndx]; }
}; // cOption_List

    //----------------------------------------
    //            DEBUG STREAMS
    //----------------------------------------
ostream& operator<<(ostream& stream, cOption& Option);


#endif	/* _OPTION_H */

