# include "../Headers/StdAfx.h"
# include "../Headers/common.h"
# include "../Headers/cFixedPartition.h"


cFixedPartition::~cFixedPartition(void) {
} // cFixedPartition::~cFixedPartition
//
//******************************************************************************
//
//  NAME: Create_Partitions
//  Arguments
//    INPUT  SortArray
//           P_Size
//    OUTPUT Partition
//
//       |       fi     |   | Partition |
//   PHI |  f | LO | HI |sig|    Type   | Action
//  -----+----+----+----+---|-----------+--------------------------------------------
//    f  |  < |  x |    |000| RLo < end | push RLo on List
//    f  |  < |    |  x |001| RHi < end | push RHi on Del
//  -----+----+----+----+---+-----------|
//    f  |  = |  x |    |010| RLo = end | push RLo on List
//    f  |  = |    |  x |011| RHi = end | push RHi on Del
//  -----+----+----+----+---+-----------|
//    f  |  > |  x |    |100| RLo > end | List, pop Del, new Partition, push RLo
//    f  |  > |    |  x |101| RHi > end | List, pop Del, new Partition, push RHi
//  ---------------------------------------------------------------------------------
//
//   LEGEND
//
//   fi             Component in SortArray
//   f              Bound of Component in SortArray
//   LO             Bound range boundary type = Low
//   HI             Bound range boundary type = High
//   sig            Signature. Binary encoding of
//                    o The relation between bounds
//                       <  == 000
//                       =  == 010
//                       >  == 100
//                    o The fi.Boundary_Type
//                       LO == 000
//                       HI == 001
//   <               f.bound < partition end, weight = 000
//   =               f.bound = partition end, weight = 010
//   >               f.bound > partition end, weight = 100
//   Partition Type  Partition decision being made
//
//******************************************************************************

void   cFixedPartition::Create_Partitions(cBound_List&     SortArray
                                         ,cPartition_List& Partition
                                         ,double           P_Size) {
        
     //---------------------------
     //        OBJECTS
     //---------------------------
    static  //             SortArray.Bound       Partition
    string const  SignatureText[] = { "low    < upper Bound <low> "
                                    , "high   < upper Bound <high>"
                                    , "log   == upper Bound <low> "
                                    , "high  == upper Bound <high>"
                                    , "low    > upper Bound <low> "
                                    , "high   > upper Bound <high>"
                                    , "low   >= upper Bound <low>  ERROR"
                                    , "high  >= upper Bound <high> ERROR"
                                    };

    static const string Name = "Create_Fixed_Partitions";
    DebugEnter(Name);

    long const StackSize = SortArray.GetSize();  // Maximum records ptrs in stack
    cStack  Stack(StackSize, "Data Stack");      // Runtime stack
    cCell*  Cell;                                // Cell returned form Stack
    double  Bound = SortArray[0].Bound + P_Size; // Upper bound of partition
    long    k = 0;                               // Partition index
    long    signature;                           // Type of partition
        
     //---------------------------
     //          CODE
     //---------------------------


    if (is_Debug()) {
       ostringstream stream(ios::out); 
       stream << "SortArray[0.."
              << SortArray.GetSize()
              << "], Partition[0.."
              << Partition.GetSize()
              << "]";
       DebugData(stream.str());
    }

    Partition.New(SortArray[0].Bound, Bound, 0);
    Cell = &Stack.Push(SortArray[0].Datum);

    for (long i = 1; i < StackSize; i++) {
      signature = 4 * (SortArray[i].Bound >  Bound)
                + 2 * (SortArray[i].Bound == Bound)
                + 1 * (SortArray[i].Type  == cBound::High);
      
      if (is_Debug()) {
         ostringstream stream(ios::out);
         stream << setw(9) << i << ": Signature(0x" 
                << setw(2) << hex << setfill('0') << signature 
                <<  ") " << SignatureText[signature]
                <<  "  " << setfill(' ')
                << setw(9) << SortArray[i].Bound << ": ["
                << setw(9) << Partition[k].RLo   << ", "
                << setw(9) << Partition[k].RHi   << "] "
                << SortArray[i];
//                << cBound::Boundary[SortArray[i].Type];
         DebugData(stream.str());
      }

      switch (signature) {  //
        case  0x0:  // 2#000#   RLo < Bound | push RLo on List
          Cell = &Stack.Push(SortArray[i].Datum);
          break;

        case  0x01: // 2#001#   RHi < Bound | push RHi on Del
          break;

        case  0x02: // 2#010#   RLo = Bound | push RLo on List
          Cell = &Stack.Push(SortArray[i].Datum);
          break;

        case  0x03: // 2#011#   RHi = Bound | push RHi on Del
          break;

        case  0x04: // 2#100#   RLo > Bound | pop Del, new Partition, push RLo
          do {
             Partition[k].List = Create_DataList(Stack);
             Partition++;
             PopDel(Partition[k], Stack);
             k              = k + 1;
             Partition.New(Bound, Bound + P_Size, k);
             Bound += P_Size;
          } while (SortArray[i].Bound > Bound);
          Cell = &Stack.Push(SortArray[i].Datum);
          break;

       case  0x05: // 2#101#   RHi > Bound | pop Del, new Partition, push RHi
          do {
             Partition[k].List = Create_DataList(Stack);
             Partition++;
             PopDel(Partition[k], Stack);
             k              = k + 1;
             Partition.New(Bound, Bound + P_Size, k);
             Bound += P_Size;
          } while (SortArray[i].Bound > Bound);
          break;

        default: 
          Stdout << "INTERNAL ERROR: " << setw(9) << i
                 << " Signature out of bounds = 0x"
                 << setw(2) << signature << endl;
          break;
      }
//      Cell->DebugData();
    };

    while ( Stack.GetSize() > 0 ) {
        Partition[k].List = Create_DataList(Stack);
        k                     = k + 1;
        Partition.New(Partition[k-1].RHi, Partition[k-1].RHi + P_Size, k );
        //Partition[k]          = * (new cPartition(Partition[k-1].RHi
        //                                        ,Partition[k-1].RHi + P_Size
        //                                        ,k ));
        PopDel(Partition[k], Stack);
    }

    //Partition++;
    //Partition[k].RLo =  Partition[k].RHi = -LDBL_MIN;
    
    //if (is_Debug())
    //   if (ValidatePartitions(SortArray, Partition)) 
    //      Stdout << "INTERNAL ERROR: Fixed Partitions failed to validate" << endl;

    DebugExit(Name);
} // cFixedPartition::Create_Partition(cBound_List SortArray, cPartition_List Partition, long P_Size)


//*****************************************************************************
// NAME: cFixedPartition::PopDel
//
// INPUTS  none
//
// OUTPUT  none
//
// RETURN  none
//
// DESCRIPTION
//    
// NOTE:
//
// UPDATE HISTORY
//     Date        Name         Description
// 19 Apr 2009  A. Schwarz      OIS
//*****************************************************************************

void cFixedPartition::PopDel(cPartition& Partition, cStack& Stack) {
   cCell*     Cell   = &Stack.Top();
   long const Size   = Stack.GetSize();

   ostringstream Legend(ios::out);
   Legend << "PopDel(" << Size << ")";
   DebugEnter(Legend.str());
   if (Size > 0) Stack.DebugData();

   for(long i = 0; i < Size; i++) {
      cCell& Next= Cell->Right();
      //if (is_Debug()) {
      //   ostringstream stream(ios::out); 
      //   stream << "Cell " << Cell << " Ref"  << &Ref;
      //   DebugData(stream.str()); 
      //}
      Cell->DebugData();
      if ((Cell->Info()).RHi < Partition.RHi ) {
         Stack.Delete(*Cell);
      }
      Cell = &Next;
   }

   DebugExit("PopDel");
}; // cFixedPartition::PopDel(cPartition& Partition, cStack& Stack, cStack& Delta)

//*****************************************************************************
// NAME: cFixedPartition::ValidatePartitions
//
// INPUTS  none
//
// OUTPUT  none
//
// RETURN  none
//
// DESCRIPTION
//    Check the partitioning algorithm for correctness. The output of the
//    partitioning process is checked for errors. The checks are exhaustive
//    and determine whether:
//     1. All input ranges are in all partitions with overlapping ranges.
//     2. All partitions contain ranges.
//     3. All partitions contain ranges which are partially contained within it.
//     4. No partitions ranges overlap (I1:high < I2:low)
//     5. All partition ranges are correct (low <= high)
//    
//  NOTE:
//   1. The input partitions are disjoint and sorted.
//
// UPDATE HISTORY
//     Date        Name         Description
// 19 Apr 2009  A. Schwarz      OIS
//*****************************************************************************

long   cFixedPartition::ValidatePartitions(cBound_List& SortArray, cPartition_List& Partition) {

     //---------------------------
     //  OBJECTS
     //---------------------------

    long   ErrorFlag       = 0;                   // total number of errors
    long const Size        = Partition.GetSize(); // Active Partitions
                                                  // Statistics
    long  BoundaryCount   = 0;                    // Number of boundaries
    long  BoundsCount     = 0;                    // Number of boundary value errors
    long  DuplicateCount  = 0;                    // Duplicate entries in a partition
    long  MissingPartitionCount = 0;              // Number of missing partitions
    long  MissingRangeCount = 0;                  // Input range not in disjoint partition
    long  NullListCount   = 0;                    // Number of partitions user list
    long  PartitionCount  = 0;                    // Number of partitions w/illegal bounds
    long  ZeroLengthCount = 0;                    // Number of partitions w/empty user list

    long  i, j, k;                                // Indexing variables

//  begin -- Validate_IdealPartitions
      DebugEnter("cFixedPartition::Validate");

      Log   << endl << "Validate fixed partitions" << endl;
      
// Check necessary conditions
      for(k = 0; k < Size; k++ ) {
        // Validate Partition
        if ( Partition[k].RLo > Partition[k].RHi ) {
          BoundaryCount++;
          ErrorFlag++;
          Log    << "INTERNAL ERROR: Partition[" << k
                 << "] Low > High "
                 << Partition[k] << endl;
        }

        if ( Partition[k].List == NULL ) {
          NullListCount++;
          ErrorFlag++;
          Log    << "INTERNAL ERROR: Partition[" << k
                 << "] Ambiguity list not assigned "
                 << Partition[k] << endl;
        //} else if ( Partition[k].List->GetSize() == 0 ) {
        //  ZeroLengthCount++;
        //  Log    << "Information: Partition[" << k
        //         << "] empty Ambiguity list "
        //         << Partition[k] << endl;
        } else {
          // Validate Range Intervals are partially contained in the partition
           for(i = 0; i < Partition[k].List->GetSize(); i++) {
              long LowError = 0;
              if (Partition[k][i].RLo > Partition[k].RHi) {
                 LowError = 1;
                 ErrorFlag++;
                 Log    << "INTERNAL ERROR: Partition[" << k << "][" << i
                    << "].Low > Partition.High "
                    << Partition[k][i] << Partition[k] << endl;
              }
              long HighError = 0;

              if (Partition[k][i].RHi < Partition[k].RLo) {
                 HighError = 1;
                 ErrorFlag++;
                 Log    << "INTERNAL ERROR: Partition[" << k << "]["  << i
                    << "].High < Partition.Low "
                    << Partition[k][i] << " " << Partition[k] << endl;
              }
              for(i = 0; i < Partition[k].List->GetSize() - 1; i++) {
                 if (Partition[k][i].Datum == Partition[k][i+1].Datum) {
                    DuplicateCount++;
                    ErrorFlag++;
                    Log    << "INTERNAL ERROR: Duplicate Entry Partition[" << k << "]["  << i
                           << "] == Partition[" << k << "][" << i + 1 << "] "
                           << Partition[k][i] << " " << Partition[k][i+1] << endl;
                 }
              }
              PartitionCount += LowError | HighError;
              if (ErrorFlag >= MAXIMUM_ERROR_COUNT) break;
          }
          if (ErrorFlag >= MAXIMUM_ERROR_COUNT) break;
        }
        Partition[k].DebugData();
      }

// Validate that the bounds for 'i' are strictly less than the bounds for 'i+1'
      for(k = 0; k < Size - 1; k++ ) {
         if ( Partition[k].RLo >= Partition[k+1].RLo ) {
            ErrorFlag++;
            BoundsCount++;
            Log    << "INTERNAL ERROR: Partition[" << k
                   << "].Low >= Partition[" << k+1
                   << "].Low "
                   << Partition[k] 
                   << Partition[k+1] << endl;
         } else if ( Partition[k].RHi > Partition[k+1].RLo ) {
            BoundsCount++;
            ErrorFlag++;
            Log    << "INTERNAL ERROR: Partition[" << k
                   << "].High > Partition[" << k+1
                   << "].Low "
                   << Partition[k] 
                   << Partition[k+1] << endl;
         } else if ( Partition[k].RHi > Partition[k+1].RHi ) {
            BoundsCount++;
            ErrorFlag++;
            Log    << "INTERNAL ERROR: Partition[" << k
                   << "].High > Partition[" << k+1
                   << "].High "
                   << Partition[k] 
                   << Partition[k+1] << endl;
         }
         if (ErrorFlag >= MAXIMUM_ERROR_COUNT) break;
      }

// Validate necessary condition
//    Each input interval must be in every partition contained in it's
//    range. We checked that each range contained in a partitione is 
//    contained within the partition. We must check (here) that the input
//    range is represented in each partition.
//
//    The checks are most easily done by observing that SortArray contains
//    a sorted list of the start and stop of a range (eLow & eHigh), and
//    that each SortArray Object has a pointer to the original input range
//    contained within it. The following observations are made:
//     1: Both boundaries of the input range are separately contained in
//        SortArray. Only one needs to be used.
//     2: A single scan of the SortArray is required.
//     3: Each scan of the partition array starts:
//        a. At or one more than the previous scan.
//        b. Terminates when the high bound of the input range is reached.
//           (we have already determined that the input ranges are 
//           contained in each partition range above and need not be
//           rechecked).

      for(i = 0, j = 0; i < SortArray.GetSize(); i++) {
         if (SortArray[i].Type = cBound::Low) {
            while(SortArray[i].Low  < Partition[j].RLo) j++;
            long ll = j;
            while(SortArray[i].High <= Partition[ll].RHi) {
               for(k = 0; k < Partition[ll].GetSize(); k++)
                  if (Partition[ll][k].Datum == (*SortArray[i].Datum).Datum) break;
               if (k >= Partition[ll].GetSize()) {
                  MissingRangeCount++;
                  ErrorFlag++;
                  Log    << "INTERNAL ERROR: Input interval [" << i 
                         << "] not found in Partitition[" << k
                         << "] " << Partition[k] << endl;
               
               }
               if (++ll >= Partition.GetSize()) break;
            }
         }
         if (ErrorFlag >= MAXIMUM_ERROR_COUNT) break;
      }

      if (ErrorFlag)
        Log    << " Errors  Description" << endl
               << setw(7) << ErrorFlag   << endl
               << setw(7) << BoundaryCount << " Boundaries between partitions"    << endl
               << setw(7) << BoundsCount   << " Entries not in bound"             << endl
               << setw(7) << DuplicateCount<< " Duplicate entries in partition"   << endl
               << setw(7) << NullListCount << " Partitions with no assigned list" << endl
               << setw(7) << ZeroLengthCount << " Partitions with no entries"     << endl
               << setw(7) << MissingPartitionCount << " Ranges with a partition"  << endl
               << setw(7) << MissingRangeCount     << " Ranges not found in partition" << endl;
      else
         Log    << " No Errors Found" << endl;
      DebugExit("cFixedPartition::Validate");
      return ErrorFlag;
}; // cFixedPartition::ValidatePartitions(cBound_List& SortArray, cPartition_List& Partition)
