#include "../Headers/StdAfx.h"
#include "../Headers/cIdealPartition.h"

cIdealPartition::~cIdealPartition(void) {
} // cIdealPartition::~cIdealPartition(void)

//
//******************************************************************************
//
//  NAME: Create_Partitions
//
//  Arguments
//    INPUT  cBound_List SortArray Sorted array of range bounds. Each range [Low, High}
//                                 is mapped to two bounds, one for low and one
//                                 for high. The mapped list is sorted.
//    OUTPUT sRange_List Partition Ordered list of partitions
//
//  DEFINITIONS:
//    Legend
//     Pi            Partition  'i'
//     Pi:[fx .. fy] Partition  'i': [ low .. high  ]
//     fjL           Range      'j' Low
//     fjH           Range      'j' High
//     Ei            User Datum 'i'
//     push fi       Push   onto ambiguity stack for partition range fi
//     pop  fi       Remove from ambiguity stack for partition range fi
//
//     Partition Membership:  fiL <= f < fiH == [ fiL, fiH )
//
//
//  DESCRIPTION:     Given a sorted range list, range partitions will be
//                   generated such that for each partition:
//                   o the partition is unique - no two partitions have
//                     the same range boundaries.
//                     Pi == Px:[ fiL .. fiH ]
//                     Pj == Px:[ fjL .. fjH ]
//                     fiL  !=  fjL  for i != j
//                     fiH  !=  fjH  for i != j
//
//                   o all partitions are disjoint - no two partitions
//                     share any portion of a range (Low, High).
//                     Pi == Pi:[ fiL .. fiH ]
//                     Pj == Pj:[ fjL .. fjH ]
//
//                     fiL  <   fjL  for i < j
//                     fiH  <=  fjL  for i == j - 1 because the partition is right side open
//                     fiH  <   fjL  for i < j - 1
//                     fiH  <   fjH  for i < j
//
//                   o all User Data associated with a partition will be
//                     included in the partition.
//
//                   o all User Data in the sorted range array will be included
//                     in at least one partition.
//
// ANALYSIS:
//
//
//    DISJOINT RANGES:
//
//
//     |      o-------------o      o-------------o
//     |      |             |      |             |
//     |      |             |      |             |
//     |      |             |      |             |
//     |      |             |      |             |
//     |------+-------------+------+-------------+--
//           f0            f0     f1            f1   Bound ->
//
//
//        input       operation   stack     partition          User Data
//     Range Datum
//      f0L   E0      push E0     { E0 }
//      f0H   E0      pop  E0     ()        P0:[ f0L .. f0H ]  E0
//      f1L   E1      push E1     { E1 }
//      f1H   E1      pop  E1     ()        P1:[ f1L .. f1H ]  E1
//
//     NUMBER OF PARTITIONS: == N (N: number of Bounduency ranges)
//
//
//
//     |  o-----------------------------------------o
//     |  |                                         |
//     |  |                                         |
//     |  |   o-------------o      o-------------o  |
//     |  |   |             |      |             |  |
//     |  |   |             |      |             |  |
//     |  |   |             |      |             |  |
//     |  |   |             |      |             |  |
//     |--+---+-------------+------+-------------+--+--
//        f2 f0            f0     f1            f1  f2 Bound ->
//
//
//        input       operation   stack     partition          User Data
//     Range Datum
//      f2L   E2      push E2     ( E2 )
//      f0L   E0      push E0     { E2 E0 } P2:[ f2L .. f0L ]  E2
//      f0H   E0      pop  E0     ( E2 )    P0:[ f0L .. f0H ]  E0 E2
//      f1L   E1      push E1     { E2 E1 } P3:[ f0H .. f1L ]  E2
//      f1H   E1      pop  E1     ( E2 )    P1:[ f1L .. f1H ]  E1 E2
//
//     NUMBER OF PARTITIONS: == N + 3 (N: number of frequency ranges)
//
//
//    OVERLAPPING RANGES:
//
//     |                                  o--------------------o
//     |                                  |                    |
//     |                                  |                    |
//     |                           o------+------o             |
//     |                           |      |      |             |
//     |      o-------------o      |      |      |             |
//     |      |             |      |      |      |             |
//     |      |             |      |      |      |             |
//     |      |      o------+------+------+------+------o      |
//     |      |      |      |      |      |      |      |      |
//     |------+------+------+------+------+------+------+------+--
//           f0     f1     f0     f2     f3     f2     f1     f3   Bound ->
//
//
//        input       operation   stack     partition          User Data
//     Range Datum
//      f0L   E0      push E0     { E0 }
//      f1L   E1      push E1     ( E0 E1 }    P0:[ f0L .. f1L ]  E0
//      f0H   E0      pop  E0     { E1 }       P1:[ f1L .. f0H ]  E0 E1
//      f2L   E2      push E2     ( E1 E2 }    P2:[ f0H .. f2L ]  E1
//      f3L   E3      push E3     { E1 E2 E3 } P3:[ f2L .. f3L ]  E1 E2
//      f2H   E2      pop  E2     ( E1 E3 }    P4:[ f3L .. f2H ]  E1 E2 E3
//      f1H   E1      pop  E1     { E3 }       P5:[ f2H .. f1H ]  E1 E3
//      f3H   E3      pop  E3     (}           P6:[ f1H .. f3H ]  E3
//
//     NUMBER OF PARTITIONS: <= O(2N) (N: number of frequency ranges)
//
//
//
//    HIGH - LOW (Fi.High == Fj.Low):
//
//
//     |      o------o
//     |      |      |
//     |      |      |
//     |      |      +------o
//     |      |      |      |
//     |------+------+------+--
//           f0     f1     f1    Bound ->
//                  f0
//
//
//        input       operation   stack     partition          User Data
//     Range Datum
//      f0L   E0      push E0     { E0 }
//      f1L   E1      push E1     ( E0 E1 }    P0:[ f0L .. f1L ]  E0
//      f0H   E0      pop  E0     { E1 }       P1:[ f1L .. f0H ]  E0 E1
//      f1H   E1      pop  E1     ( E1 }       P2:[ f1L .. f1H ]  E1
//
//     NUMBER OF PARTITIONS: <= O(2N + N - 1) (N: number of frequency ranges
//                                              + number of overlapping end points)
//
//
//
//    HIGH - HIGH (Fi.High == Fj.HIGH):
//
//
//     |      o-------------o
//     |      |             |
//     |      |             |
//     |      |      +------o
//     |      |      |      |
//     |------+------+------+--
//           f0     f1     f1    Bound ->
//                         f0
//
//
//        input       operation   stack     partition          User Data
//     Range Datum
//      f0L   E0      push E0     { E0 }
//      f1L   E1      push E1     ( E0 E1 }    P0:[ f0L .. f1L ]  E0
//      f0H   E0      pop  E0     { E1 }       P1:[ f1L .. f0H ]  E0
//      f1H   E1      pop  E1     (    }       P2:[ f0L .. f0H ]  E0 E1
//
//     NUMBER OF PARTITIONS: <= O(N)  (N: number of Bounduency ranges
//
//
//
// ALGORITHM:
//
//    1. Create sortable records: Each range (RLo  , RHi  ) becomes
//       a single record: <RLo  , Low , User Datum> &
//                        <RHi  , high, User Datum>
//
//       b. Since fiL <= f < fiH == [ fiL, fiH ),
//            setting Record.RHi    = RHi   + 1
//            guarantees that the partition contains the upper bound.
//
//    2. Sort the records, RLo   < RHi   for fiL == fjH
//
//    3. Form partitions from the sorted records. NOTE: Indices refer to
//       the index values after sorting.
//
//       a.   Start: P0:[ f1L .. ] E1, f1L is the first frequency after sorting.
//            f1L is the lowest frequency in the UDF. Advance to the next
//            frequency (f2). After each step, advance the index
//
//       b.   fi == fiL
//            fj == fjL
//            i  == j - 1
//            if fiL == fjL { push User Datum Ej
//
//            if fiL < fjL {
//               Pk:  [     .. fjL ]
//               Pk+1:[ fjL ..     ] push User Datum Ej
//
//       c.   fi == fiL
//            fj == fjH
//            i  == j - 1
//            if fiL == fjH {
//               Pk:  [     .. fjH ] pop  User Datum Ej
//               Pk+1:[ fjH ..     ] push User Datum Ei
//
//            if fiL < fjH {
//               Pk:  [     .. fjH ] pop User Datum Ej
//               Pk+1:[ fjH ..     ]
//
//       d.   fi == fiH
//            fj == fjH
//            i  == j - 1
//            if fiH == fjH { pop User Datum Ej
//
//            if fiH < fjH {
//               Pk:  [ fiH .. fjH ] pop User Datum Ej
//               Pk+1:[ fjH ..     ]
//
//       e.   fi == fiH
//            fj == fjL
//            i  == j - 1
//            if fiH == fjL { (see c. fiL == fjH)
//
//            if fiH < fjL {
//              if [fiH .. fiL ] populated with User Data
//                 Pk:  [ fiH .. fiL ]
//                 Pk+1:[ fiL ..     ] push User Datum j
//            else
//               Pk:  [ fiH ..     ] delete
//               Pk:  [ fjL ..     ] push User Datum j
//
//    4. Clean-Up: if the last partition has no User Data, delete it.
//
//******************************************************************************

void cIdealPartition::Create_Partitions(cBound_List& SortArray, cPartition_List& Partition) {
    //  ----------------------------------------------------------------------
    //  |         i == j - 1                                                 |
    //  |    |       fj     |    |   |                 |                     |
    //  | fi |  f | LO | HI | fi |sig|    SortArray    |    Partition Type   |
    //  |----+----+----+----+----|---+-----------------+---------------------|
    //  | f  |  < |  x |    | LO |000|   {     fiL fjL }  OVERLAPPING RANGES |
    //  | f  |  < |    |  x | LO |001|   { fjL fiL fjH }  OVERLAPPING RANGES |
    //  | f  |  < |  x |    | HI |010|   {     fiH fjL }  DISJOINT RANGES    |
    //  | f  |  < |    |  x | HI |011|   { fjL fiH fjH }  OVERLAPPING RANGES |
    //  |----+----+----+----+----+---+-----------------+---------------------|
    //  | f  | == |  x |    | LO |100|   {     fiL fjL }  OVERLAPPING RANGES |
    //  | f  | == |    |  x | LO |101|   { fjL fiL fjH }  HIGH - LOW         |
    //  | f  | == |  x |    | HI |110|   {     fiH fjL }  can not occur      |
    //  | f  | == |    |  x | HI |111|   { fjL fiH fjH }  HIGH - HIGH        |
    //  |----+----+----+----+----+---+-----------------+---------------------|
    //  | f  |  > |  x |    | LO |---|                 |  can not occur      |
    //  | f  |  > |    |  x | LO |---|                 |  can not occur      |
    //  | f  |  > |  x |    | HI |---|                 |  can not occur      |
    //  | f  |  > |    |  x | HI |---|                 |  can not occur      |
    //  ----------------------------------------------------------------------
    //
    //   LEGEND
    //
    //   fi             Component in inList. i == j - 1
    //   fj             Component in inList
    //   f              Range of Component in SortedArray
    //   LO             Range boundary type == Low
    //   HI             Range boundary type == High
    //   sig            Signature. Binary encoding of
    //                    o The relation between ranges
    //                       <  == 000
    //                      ==  == 100
    //                    o The fi.Boundary_Type
    //                       LO == 000
    //                       HI == 010
    //                    o The fj.Boundary_Type
    //                       LO == 000
    //                       HI == 001
    //   <               fi.range  < fj.range, weight == 000
    //   ==              fi.range == fj.range, weight == 100
    //   >               Not possible because input sorted
    //   SortedArray     Position of Frequency_Record objects in
    //                   sorted array. The last two items in the
    //                   list represent the items in the table.
    //                   Where there are three items, the first
    //                   items represents the precedence relation
    //                   needed to satisfy the Partition Type.
    //                   The first item may not immediately precede
    //                   the second and third items.
    //   Partition Type  The Partition Type as described in the package
    //                   preamble.
    //
    //******************************************************************************


    //---------------------------
    //         OBJECTS
    //---------------------------
    static
    string const  SignatureText[] = { "fi != fj & <Low , Low >"
                                    , "fi != fj & <Low , High>"
                                    , "fi != fj & <High, Low >"
                                    , "fi != fj & <High, High>"
                                    , "fi == fj & <Low , Low >"
                                    , "fi == fj & <Low , High>"
                                    , "fi == fj & <High, Low > ERROR"
                                    , "fi == fj & <High, High>"
    }; // Signature Text

    long const StackSize = SortArray.GetSize();  // Maximum record ptrs in stack
    cStack  Stack(StackSize/2, "Data Stack");    // Runtime stack
    long    k = 0;                               // Partition index
    long    signature;                           // Type of partition
    static const string Name = "Create_IdealPartitions";
    ostringstream stream(ios::out); 

    //---------------------------
    //          CODE
    //---------------------------

    DebugEnter(Name);

    if (is_Debug()) {
       ostringstream stream(ios::out);
       stream << "SortArray[0 .. "
              << setw(9) << SortArray.GetSize()
              << "], Partition[0 .. "
              << Partition.GetSize()
              << "]";
       DebugData(stream.str());
    }

    Partition.New(SortArray[0].Bound, 0);

    Stack.Push(SortArray[0].Info());
    Stack.PrintStack();

    for (long i = 1; i < StackSize; i++) {
        signature = 4 * (SortArray[i - 1].Bound == SortArray[i].Bound)
                  + 2 *  SortArray[i - 1].Type
                  + 1 *  SortArray[i    ].Type;

        if (is_Debug()) {
           ostringstream stream(ios::out);
           stream << setw(9) << i << ": Signature(0x" 
                  << setw(2) << hex << setfill('0') << signature 
                  <<  ") " << SignatureText[signature]
                  << setfill(' ')
                  <<  "  " << SortArray[i - 1];
           DebugData(stream.str());
        }

        switch(signature) { // fi = i - 1   fj = i
            case 0x0:  // 2#000# OVERLAPPING RANGES: fi < fj & fi.LO & fj.LO
                Partition[k].RHi  = SortArray[i].Bound;
                Partition[k].List = Create_DataList(Stack);
                Partition[k].DebugData();
                k = k + 1;
                Partition++;
                Partition.New(SortArray[i].Bound, k);
                Stack.Push(SortArray[i].Datum);
                Partition[k - 1].DebugData();
                break;

            case 0x1: // 2#001# OVERLAPPING RANGES: fi < fj & fi.LO & fj.HI
                Partition[k].RHi  = SortArray[i].Bound;
                Partition[k].List = Create_DataList(Stack);
                Partition[k].DebugData();
                Stack.Delete((Stack.Find(SortArray[i].Datum)));
                k = k + 1;
                Partition++;
                Partition.New(SortArray[i].Bound, k);
                Partition[k - 1].DebugData();
                break;

            case 0x2: // 2#010# DISJOINT RANGES   : fi < fj & fi.HI & fj.LO
                if (Stack.GetSize() > 0) {
                    Partition[k].RHi = SortArray[i].Bound;
                    Partition[k].List = Create_DataList(Stack);
                    Partition[k].DebugData();
                    Partition[k].DebugData();
                    k = k + 1;
                    Partition++;
                    Partition.New(SortArray[i].Bound, k);
                } else
                    Partition[k].RLo = SortArray[i].Bound;
                Stack.Push(SortArray[i].Info());
                break;

            case 0x3: // 2#011# OVERLAPPING RANGES: fi < fj & fi.HI & fj.HI
                Partition[k].RHi = SortArray[i].Bound;
                Partition[k].List = Create_DataList(Stack);
                Partition[k].DebugData();
                Stack.Delete((Stack.Find(SortArray[i].Datum)));
                k = k + 1;
                Partition++;
                Partition.New(SortArray[i].Bound, k);
                Partition[k - 1].DebugData();
                break;

            case 0x4: // 2#100# OVERLAPPING RANGES: fi = fj & fi.LO & fj.LO
                Stack.Push(SortArray[i].Info());
                break;

            case 0x5: {// 2#101# HIGH - LOW        : fi = fj & fi.LO & fj.HI
                Partition[k].RHi = SortArray[i].Bound;
                Partition[k].List = Create_DataList(Stack);
                Partition[k].DebugData();
                Stack.Delete((Stack.Find(SortArray[i].Datum)));
                k = k + 1;
                Partition++;
                int  Exp;                        // Calculate the number for the least
                frexp(SortArray[i].Bound, &Exp); // signicant bit with the given exp
                double Delta = ldexp(1.0L, Exp - 53);
//union { double f; unsigned long l[2]; } x, y, z;
//x.f = SortArray[i].Bound;
//y.f = SortArray[i].Bound + Delta;
//z.f = Delta;
//Log << setfill(' ') <<dec << setw(8) << Exp << " " << setw(8) << Exp - 52
//    << setfill('0') << "   0x" << hex << setw(8) << x.l[1] << hex << setw(8) << x.l[0]
//                    << " + 0x" << hex << setw(8) << z.l[1] << hex << setw(8) << z.l[0] 
//                    << " = 0x" << hex << setw(8) << y.l[1] << hex << setw(8) << y.l[0] << endl;
                Partition.New(SortArray[i].Bound + Delta, k);
                Partition[k - 1].DebugData();
                break;
            }
            case 0x6: // 2#110# can not occur     : fi = fj & fi.HI & fj.LO
                Stdout << "INTERNAL ERROR: " << setw(9) << i
                       << ": Signature not in bounds. 2#0110#" << endl;
                break;

            case 0x7: // 2#111# HIGH - HIGH       : fi = fj & fi.HI & fj.HI
                Stack.Delete((Stack.Find(SortArray[i].Datum)));
                break;

            default: 
                Stdout << "INTERNAL ERROR: " << setw(9) << i
                       << ": Signature not in bounds. "
                       << setw(4) << hex << signature << endl;
                break;
        }
//        Stack.PrintStack();
    }

    //if (is_Debug())
    //   if (ValidatePartitions(SortArray, Partition)) 
    //      Stdout << "INTERNAL ERROR: Ideal Partitions failed to validate" << endl;

    DebugExit(Name);
 } // cIdealPartition::Create_Partitions(cBound_List SortArray, cPartition_List Partition)

//*****************************************************************************
// NAME: cIdealPartition::ValidatePartitions
//
// INPUTS  none
//
// OUTPUT  none
//
// RETURN  none
//
// DESCRIPTION
//    Check the partitioning algorithm for correctness. The output of the
//    partitioning process is checked for errors. The checks are exhaustive
//    and determine whether:
//     1. All input ranges are in all partitions with overlapping ranges.
//     2. All partitions contain ranges.
//     3. All partitions contain ranges which are wholly contained within it.
//     4. No partitions ranges overlap (I1:high < I2:low)
//     5. All partition ranges are correct (low <= high)
//    
//  NOTE:
//   1. The input partitions are disjoint and sorted.
//
// UPDATE HISTORY
//     Date        Name         Description
// 19 Apr 2009  A. Schwarz      OIS
//*****************************************************************************

long   cIdealPartition::ValidatePartitions(cBound_List& SortArray, cPartition_List& Partition) {

     //---------------------------
     //  OBJECTS
     //---------------------------

    long   ErrorFlag       = 0;                   // total number of errors
    long const Size        = Partition.GetSize(); // Active Partitions
                                                  // Statistics
    long  BoundaryCount   = 0;                    // Number of boundaries
    long  BoundsCount     = 0;                    // Number of boundary value errors
    long  EntryCount      = 0;                    // Number of entries out of bounds
    long  DuplicateCount  = 0;                    // Duplicate entries in a partition
    long  MissingPartitionCount = 0;              // Number of missing partitions
    long  MissingRangeCount = 0;                  // Input range not in disjoint partition
    long  NullListCount   = 0;                    // Number of partitions user list
    long  PartitionCount  = 0;                    // Number of partitions w/illegal bounds
    long  ZeroLengthCount = 0;                    // Number of partitions w/empty user list

    long  i, j, k, l;                             // Indexing variables

    DebugEnter("Validate_IdealPartitions");

    Log   << "Validate ideal partitions" << endl;
      
// Check necessary conditions
      for(k = 0; k < Size; k++ ) {
        // Validate Partition
        if ( Partition[k].RLo > Partition[k].RHi ) {
          BoundaryCount++;
          Log    << "INTERNAL ERROR: Partition[" << k
                 << "] Low > High "
                 << Partition[k] << endl;
        }

        if ( Partition[k].List == NULL ) {
          NullListCount++;
          Log    << "INTERNAL ERROR: Partition[" << k
                 << "] Ambiguity list not assigned "
                 << Partition[k] << endl;
        } else if ( Partition[k].List->GetSize() == 0 ) {
          ZeroLengthCount++;
          Log    << "INTERNAL ERROR: Partition[" << k
                 << "] empty Ambiguity list "
                 << Partition[k] << endl;
        } else {
           // Validate Range Intervals are fully contained in the partition
           for(i = 0; i < Partition[k].List->GetSize(); i++) {
              long LowError = 0;
              if (Partition[k][i].RLo > Partition[k].RHi) {
                 EntryCount += 1;
                 Log    << "INTERNAL ERROR: Partition[" << k << "][" << i
                        << "].RLo > Partition.RHi "
                        << Partition[k][i] << Partition[k] << endl;
              } else if ((Partition[k][i]).RLo > Partition[k].RLo) {
                 EntryCount += 1;
                 Log    << "INTERNAL ERROR: Partition[" << k << "][" << i
                        << "].Low > Partition.Low "
                        << Partition[k][i] << " " << Partition[k] << endl;
              }
              long HighError = 0;

              if (Partition[k][i].RHi < Partition[k].RLo) {
                 EntryCount += 1;
                 Log    << "INTERNAL ERROR: Partition[" << k << "][" << i
                        << "].High < Partition.Low "
                        << Partition[k][i] << " " << Partition[k] << endl;
              } else if (Partition[k][i].RHi < Partition[k].RHi) {
                 EntryCount += 1;
                 Log    << "INTERNAL ERROR: Partition[" << k << "][" << i
                        << "].High > Partition.High "
                        << Partition[k][i] << " " << Partition[k] << endl;
              }
              for(i = 0; i < Partition[k].List->GetSize() - 1; i++) {
                 if (Partition[k][i].Datum == Partition[k][i+1].Datum) {
                    DuplicateCount++;
                    ErrorFlag++;
                    Log    << "INTERNAL ERROR: Duplicate Entry Partition[" << k << "]["  << i
                           << "] == Partition[" << k << "][" << i + 1 << "] "
                           << Partition[k][i] << " " << Partition[k][i+1] << endl;
                 }
              }
           }
        }
        Partition[k].DebugData();
      }

// Validate that the bounds for 'i' are strictly less than the bounds for 'i+1'
      for(k = 0; k < Size - 1; k++ ) {
         if ( Partition[k].RLo >= Partition[k+1].RLo ) {
            BoundsCount++;
            Log    << "INTERNAL ERROR: Partition[" << k
                   << "].Low >= Partition[" << k+1
                   << "].Low "
                   << Partition[k] 
                   << Partition[k+1] << endl;
         } else if ( Partition[k].RHi > Partition[k+1].RLo ) {
            BoundsCount++;
            Log    << "INTERNAL ERROR: Partition[" << k
                   << "].High > Partition[" << k+1
                   << "].Low "
                   << Partition[k] 
                   << Partition[k+1] << endl;
         } else if ( Partition[k].RHi > Partition[k+1].RHi ) {
            BoundsCount++;
            Log    << "INTERNAL ERROR: Partition[" << k
                   << "].High > Partition[" << k+1
                   << "].High "
                   << Partition[k] 
                   << Partition[k+1] << endl;
         }
      }

// Validate necessary condition
//    Each input interval must be in every partition contained in it's range
//    and each input range must wholly contained within the partition.
//    We checked that each range contained in a partitione is wholly
//    contained within the partition. We must check (here) that the
//    input range is represented in each partition.
//
//    The checks are most easily done by observing that SortArray contains
//    a sorted list of the start and stop of a range (eLow & eHigh), and
//    that each SortArray Object has a pointer to the original input range
//    contained within it. The following observations are made:
//     1: Both boundaries of the input range are separately contained in
//        SortArray. Only one needs to be used.
//     2: A single scan of the SortArray is required.
//     3: Each scan of the partition array starts:
//        a. At or one more than the previous scan.
//        b. Terminates when the high bound of the input range is reached.
//           (we have already determined that the input ranges are fully
//           contained in each partition range above and need not be
//           rechecked).

      for(i = 0, k = 0; i < SortArray.GetSize(); i++) {
         if (SortArray[i].Type == cBound::Low) {
            cRange* Range = SortArray[i].Info();
            while(Range->RLo > Partition[k].RHi) k++;
            for(j = k; Range->RLo < Partition[j].RLo;) j++;
            for(; Range->RHi < Partition[j].RLo; j++) {
               for(l = 0; Partition[j].GetSize(); l++)
                  if (Range == &Partition[j][l]) break;
               if (l >= Partition[l].GetSize()) {
                  MissingRangeCount++;
                  Log    << "INTERNAL ERROR: Input range missing in partition "
                     << Range 
                     << " Not in "
                     << Partition[j] << endl;
               }
            }
         }
      }

      ErrorFlag = (BoundaryCount        + BoundsCount       + EntryCount + DuplicateCount
                + MissingPartitionCount + MissingRangeCount + NullListCount
                + ZeroLengthCount);
      if (ErrorFlag)
         Log    << " Errors  Description" << endl
                << setw(7) << ErrorFlag   << endl
                << setw(7) << BoundaryCount     << " Boundaries between partitions"   << endl
                << setw(7) << BoundsCount       << " Partitions with illegal bounds"  << endl
                << setw(7) << DuplicateCount    << " Duplicate entries in partition"  << endl
                << setw(7) << EntryCount        << " Entries not fully in parition"   << endl
                << setw(7) << MissingPartitionCount << " Interval not in any partition" << endl
                << setw(7) << MissingRangeCount << " Ranges not found in partition"   << endl
                << setw(7) << NullListCount     << " Partitions with no assigned entries" << endl
                << setw(7) << ZeroLengthCount   << " Partitions with no entries"      << endl;
      else
         Log    << " No Errors Found" << endl;
      DebugExit("Validate_IdealPartitions");
      return ErrorFlag;
}; // cIdealPartition::ValidatePartitions(cBound_List& SortArray, cPartition_List& Partition)
