
# include "../Headers/stdafx.h"

# include "../Headers/common.h"
# include <climits>

using namespace std;

cRange       cList::ErrorItem   = ErrorItem.Init();
string const cBound::Boundary[] = { "low", "high", "error" };
string const cPartition::GlobalName = "Partition Record";

//*****************************************************************************
// NAME: cCell::InLeft
//
// INPUTS  Cell   Cell to be inserted
//
// OUTPUT  none
//
// RETURN  cCell& Cell to be inserted
//
// DESCRIPTION
//    Inserts the input cell to the left of *this
//
// UPDATE HISTORY
//     Date        Name         Description
// 19 Apr 2009  A. Schwarz      OIS
//*****************************************************************************

cCell& cCell::InLeft(cCell& Cell) {
    Cell.Right_Link       = this;
    Cell.Left_Link        = Left_Link;
    Left_Link->Right_Link = &Cell;
    Left_Link             = &Cell;
    return Cell;
} // cCell& cStack::InLeft(cCell &Cell)

//*****************************************************************************
// NAME: cCell::InRight
//
// INPUTS  Cell   Cell to be inserted
//
// OUTPUT  none
//
// RETURN  cCell& Cell to be inserted
//
// DESCRIPTION
//    Insert cell to the right of *this.
//
// UPDATE HISTORY
//     Date        Name         Description
// 19 Apr 2009  A. Schwarz      OIS
//*****************************************************************************

cCell& cCell::InRight(cCell& Cell) {
    Cell.Left_Link        = this;
    Cell.Right_Link       = Right_Link;
    Right_Link->Left_Link = &Cell;
    Right_Link            = &Cell;
    return Cell;
} // cCell& cCell::InRight(cCell &Cell)

//*****************************************************************************
// NAME: cCell::Delete
//
// INPUTS  none
//
// OUTPUT  none
//
// RETURN  cCell&  Current cell.
//
// DESCRIPTION
//    Delete the current cell from the current list.
//
// UPDATE HISTORY
//     Date        Name         Description
// 19 Apr 2009  A. Schwarz      OIS
//*****************************************************************************

cCell& cCell::Delete() {
    Left_Link->Right_Link  = Right_Link;
    Right_Link->Left_Link  = Left_Link;
    Left_Link = Right_Link = NULL;
    Type      = eValue;
    Value() = 0;
    return (cCell& )*this;
} // cCell& cCell::Delete(cCell &Cell)

//*****************************************************************************
// NAME: cStack::cStack
//
// INPUTS  mySize  Number of stack cells to allocate
//         myName  Name to be used for the staack
//
// OUTPUT  none
//
// RETURN  none
//
// DESCRIPTION
//    Constructor for the stack. 
//     1. Allocates space from the heap.
//     2. Initializes the space (cells connected onto the avaiable space list
//        AVSL) as a linked list.
//     3. Create a null used stack (head points to itself).
//     4. Create an initial free space list (AVSL) points the heap.
//     5. Initialize the AVSL to the list size and the used space to zero.
//
// NOTE:
//  1. The list implementation includes the list header (AVSL or Head) as 
//     part of the list. A null list is indicated by the list header pointing
//     to itself and the list count as zero (this list count is contained as
//     the list header Datum).
//
// UPDATE HISTORY
//     Date        Name         Description
// 19 Apr 2009  A. Schwarz      OIS
//*****************************************************************************

cStack::cStack(int mySize, string myName) : Size((mySize > 0)? mySize: 0)
                                          , Name(myName)
                                          , NextCell(Head)
{
    DebugEnter("cStack::cStack");
    Head.Left_Link    = &Head;                   // Initialize header to point
    Head.Right_Link   = &Head;                   // to itself
                                                 // Change (cRange*)Head.Datum to long
    Head.Value()      = 0;                       // Number of cells in active list
    Head.Type         = cCell::eValue;

    Space             = new cCell[Size+1];       // Allocate available space list
    AVSL.Right_Link   = &Space[0];               // Initialize header to point to
    AVSL.Left_Link    = &Space[Size-1];          // the first and last cell in AVSL
                                                 // Change (cRange*)AVSL.Datum to long
    AVSL.Value()      = Size;                    // Number of cells in available space list
    AVSL.Type         = cCell::eValue;

    for(long i = 0; i < Size; i++) {             // Initialize space list
        Space[i].Left_Link   = &Space[i-1];
        Space[i].Right_Link  = &Space[i+1];
        Space[i].Value()     = i;
        Space[i].Type        = cCell::eValue;
    }
    Space[0].Left_Link       = &AVSL;
    Space[Size-1].Right_Link = &AVSL;
    
    Space[Size].Right_Link   = Space[Size].Left_Link = NULL;
    Space[Size].Value()      = LONG_MIN;
    Space[Size].Type         = cCell::eValue;
    if (is_Debug()) PrintStack();
    DebugExit("cStack::cStack");
}; // cStack::cStack(int Size, string name)

//*****************************************************************************
// NAME: cStack::NUCELL
//
// INPUTS  none
//
// OUTPUT  none
//
// RETURN  cCell& Reference to acquired cell
//
// DESCRIPTION
//    Return the next avaiable cell from the AVSL and decrease the number of
//    free cells remaining by one. If no cell exists, return a known false
//    cell (in lieu of aborting).
//
//    Deletion of a cell from the AVSL decrements the AVSL count of cells by 1.
//
// UPDATE HISTORY
//     Date        Name         Description
// 19 Apr 2009  A. Schwarz      OIS
//*****************************************************************************

cCell&  cStack::NUCELL() {
//    DebugEnter("cStack::NUCELL");
    if (&AVSL.Right() == &AVSL) {
      DebugExit("cStack::NUCELL(ERROR)");
      return Space[Size];
    }
    AVSL.Value()--;                               // Update AVSL count
    cCell& Cell = (AVSL.Right()).Delete();
//    if (is_Debug()) PrintAVSL();
//    DebugExit("cStack::NUCELL");
    return Cell;
} // cCell&  cStack::NUCELL()

//*****************************************************************************
// NAME: cStack::Delete
//
// INPUTS  Cell   Reference to cell to be deleted
//
// OUTPUT  none
//
// RETURN  cCell& Reference to cell to be deleted
//
// DESCRIPTION
//    Delete a cell from the current list and insert it into the AVSL.
//    Decrement the Active list count by one, increment the AVSL list
//    count by one.
//
// UPDATE HISTORY
//     Date        Name         Description
// 19 Apr 2009  A. Schwarz      OIS
//*****************************************************************************

cCell&  cStack::Delete(cCell& Cell) {
    DebugEnter("cStack::Delete");
    Cell.DebugData();
    Head.Value()--;                              // Update count of active cells
    AVSL.Value()++;                              // Update AVSL count
    AVSL.InRight(Cell.Delete());
    if (is_Debug()) PrintStack();
    DebugExit("cStack::Delete");
    return Cell;
} // cCell&  cStack::Delete(cCell& Cell)

//*****************************************************************************
// NAME: cStack::Find
//
// INPUTS  myInfo  Search datum
//
// OUTPUT  none
//
// RETURN  cCell& Found cell
//
// DESCRIPTION
//    Search the active stack for a Datum field with an ID equal to the ID
//    contained in the input myInfo. If found, return it, if not found,
//    return a cell with known bad properties - in lieu of aborting.
//
// NOTE:
//  1. This algorithm is depedant on program assumptions It is not general.
//  2. It is assumed that the input range (cRange&) Datum value is being
//     sought. It is assumed that the cell Datum is a pointer to a cRange
//     object, which may not be the same as myInfo, with contains a Datum
//     field.
//
// UPDATE HISTORY
//     Date        Name         Description
// 19 Apr 2009  A. Schwarz      OIS
//*****************************************************************************

cCell&  cStack::Find(cRange& myInfo) {
    DebugEnter("cStack::Find");
    long    i;
    cCell*  Cell = &Head.Right();
    for(i = 0; i < GetSize(); Cell = &Cell->Right(), i++)
        if (myInfo.Datum == Cell->Info().Datum) break;
    if (i >= GetSize()) Cell = &Space[Size];
    Cell->DebugData();
    DebugExit("cStack::Find");
    return *Cell;;
} // cCell&  cStack::Find(long myInfo)

//*****************************************************************************
// NAME: cStack::Peek
//
// INPUTS  none
//
// OUTPUT  none
//
// RETURN  cCell&  Reference to first cell on stack
//
// DESCRIPTION
//    Shell function. 
//
// UPDATE HISTORY
//     Date        Name         Description
// 19 Apr 2009  A. Schwarz      OIS
//*****************************************************************************

inline
cCell&  cStack::Peek() { return (&Head.Right() != &Head)? (Head.Right())
                                                        : Space[Size]; }

//*****************************************************************************
// NAME: cStack::Peek
//
// INPUTS  Ndx    Index for cell to be retrieved
//
// OUTPUT  cCell& Retrieved cell
//
// RETURN  cCell&  Reference to first cell on stack
//
// DESCRIPTION
//    Retrieves the ith cell in the stack, starting at 0. If there is no
//    ith cell, returns a cell of known bad value in lieu of aborting.
//
// UPDATE HISTORY
//     Date        Name         Description
// 19 Apr 2009  A. Schwarz      OIS
//*****************************************************************************

cCell&  cStack::Peek(long Ndx) {
    if ((Ndx < 0) || (Ndx > GetSize())) return Space[Size];
    
    cCell&  Cell = Head.Right();
    for(long i = Ndx; i > 0; i++) Cell = Cell.Right();
    return Cell;
} // cCell&  cStack::Peek(long Ndx)

//*****************************************************************************
// NAME: cStack::Pop
//
// INPUTS  none
//
// OUTPUT  none
//
// RETURN  cCell&  Popped cell.
//
// DESCRIPTION
//    Remves the top cell in the stack and returns it to the AVSL. The active
//    cell count of the stack is decremented and the active cell count of the
//    AVSL is incremented. If the stack is empty, returns a cell of known bac
//    value in lieu of aborting.
//
// UPDATE HISTORY
//     Date        Name         Description
// 19 Apr 2009  A. Schwarz      OIS
//*****************************************************************************

cCell&  cStack::Pop() {
    DebugEnter("cStack::Pop");
    if (&Head.Right() == &Head) {
       DebugExit("cStack::Pop");
       return Space[Size];
    }
    Head.Value()--;                              // Update count of active cells
    AVSL.Value()++;                              // Update AVSL count
    cCell& Cell = AVSL.InRight((Head.Right()).Delete());
    if (is_Debug()) PrintStack();
    DebugExit("cStack::Pop");
    return Cell;
} // cCell&  cStack::Pop()

//*****************************************************************************
// NAME: cStack::PrintAVSL
//
// INPUTS  none
//
// OUTPUT  none
//
// RETURN  none
//
// DESCRIPTION
//    (Debug) Outputs the contents of the AVSL.
//
// UPDATE HISTORY
//     Date        Name         Description
// 19 Apr 2009  A. Schwarz      OIS
//*****************************************************************************

void cStack::PrintAVSL() {
   if (is_Debug()) PrintList("AVSL", &AVSL);
} // cStack::PrintAVSL()

//*****************************************************************************
// NAME: cStack::PrintStack
//
// INPUTS  none
//
// OUTPUT  none
//
// RETURN  none
//
// DESCRIPTION
//    (Debug) Outputs the contents of the active list.
//
// UPDATE HISTORY
//     Date        Name         Description
// 19 Apr 2009  A. Schwarz      OIS
//*****************************************************************************

void cStack::PrintStack() {
   if (is_Debug()) PrintList("STACK", &Head);
} // cStack::PrintStack()

//*****************************************************************************
// NAME: cStack::PrintList
//
// INPUTS  none
//
// OUTPUT  none
//
// RETURN  none
//
// DESCRIPTION
//    (Debug) Formats and outputs the contents of a list to the log file.
//
// UPDATE HISTORY
//     Date        Name         Description
// 19 Apr 2009  A. Schwarz      OIS
//*****************************************************************************

void cStack::PrintList(string Name, cCell* Head) {
   if (!is_Debug()) return;
   DebugData(Name);
   Head->DebugData();
   for(cCell* Cell = &Head->Right(); Cell != Head; Cell = &Cell->Right()) {
     Log << "  ";
     Cell->DebugData();
   }
} // cStack::PrintList(string Name, cCell* Head)

//*****************************************************************************
// NAME: cStack::Push
//
// INPUTS  none
//
// OUTPUT  none
//
// RETURN  cCell&  Cell pushed onto the stack
//
// DESCRIPTION
//    Shell function. Removes a cell from the AVSL and pushes it onto the
//    active list. The active list counter is incremented.
//
// UPDATE HISTORY
//     Date        Name         Description
// 19 Apr 2009  A. Schwarz      OIS
//*****************************************************************************

cCell&  cStack::Push() {
//    myInfo.DebugData();
    cCell* Cell = &NUCELL();
    if (Cell == &Space[Size]) {
      DebugExit("cStack::Push(ERROR) No More Space");
      return Space[Size];
    }
    Head.Value()++;                              // Update count of active cells
    Head.InRight(*Cell);
    return *Cell;
} // cCell&  cStack::Push()

//*****************************************************************************
// NAME: cStack::Push
//
// INPUTS  myCell  Cell to be pushed onto the stack
//
// OUTPUT  none
//
// RETURN  cCell&  Cell pushed onto the stack
//
// DESCRIPTION
//    Copies the contents of myCell into a new cell and pushes the ocpy onto
//    the stack.
//
// UPDATE HISTORY
//     Date        Name         Description
// 19 Apr 2009  A. Schwarz      OIS
//*****************************************************************************

cCell&  cStack::Push(cCell&  myCell) {
    DebugEnter("cStack::Push(cCell&)");
    cCell& Cell = Push();
    Cell.SetType(cCell::eCell);
    Cell.Datum.Cell = &myCell;
    if (is_Debug()) PrintStack();
    DebugExit("cStack::Push(cCell&)");
    return Cell;
}; // cCell&  Push(cCell&  Cell)

//*****************************************************************************
// NAME: cStack::Push
//
// INPUTS  myInfo  Data to be pushed onto the stack
//
// OUTPUT  none
//
// RETURN  cCell&  Reference to cell pushed onto the stack
//
// DESCRIPTION
//    Retrieves a new cell, copies the data into the new cell, and pushes the
//    new cell onto the stack.
//
// UPDATE HISTORY
//     Date        Name         Description
// 19 Apr 2009  A. Schwarz      OIS
//*****************************************************************************

cCell&  cStack::Push(cRange& myInfo) {
    DebugEnter("cStack::Push(cRange&)");
    cCell& Cell = Push();
    Cell.SetType(cCell::eRange);
    Cell.Datum.Range  = &myInfo;
    if (is_Debug()) PrintStack();
    DebugExit("cStack::Push(cRange&)");
    return Cell;
} // cCell&  cStack::Push(long *myInfo)

    //----------------------------------------
    //            Debug Streams
    //----------------------------------------

  inline ostream& operator<<(ostream& stream, cRange& Range) {
    stream << setfill(' ') 
           << "<"  << setw(9) << fixed << setprecision(3) << Range.RLo
           << ", " << setw(9) << fixed << setprecision(3) << Range.RHi
           << ", " << setw(6) << dec                      << Range.Datum
           << ">";
  return stream;
  }; // ostream& operator(ostream& stream, cRange& Range)

  ostream& operator<<(ostream& stream, cRange* Range) { return operator<<(stream, *Range); }

  inline ostream& operator<<(ostream& stream, cCell& Cell) {
    stream << "<" ;
    if (Cell.Type == cCell::eRange) stream <<  Cell.Datum.Range << ">";
    else                            stream <<  Cell.Cell() << ">";
//    else                            stream << (Cell.Cell()->Info()) << ">";
//    else                            stream << (Cell.Datum.Cell->Info()) << ">";
  return stream;
  }; // ostream& operator<<(ostream& stream, cCell& Cell)

  ostream& operator<<(ostream& stream, cBound& Bound) {
    stream << setfill(' ') 
           << '<'  << setw(9) << fixed << setprecision(3) << Bound.Bound
           << "  " << setw(5) << cBound::Boundary[Bound.Type]
           << "  " << Bound.Datum
            << '>';
  return stream;
  }; // ostream& operator<<(ostream& stream, cBound& Bound)

  ostream& operator<<(ostream& stream, cPartition& Partition) {
    stream << setfill(' ')
           << setw(7) << Partition.Name << "::"
           << '<'  << setw(9) << fixed << setprecision(3) << Partition.RLo
           << ", " << setw(9) << fixed << setprecision(3) << Partition.RHi
           << ", " << setw(5) << Partition.List->GetSize()
           << '>';
  return stream;
  }; // ostream& operator<<(ostream& stream, cPartition& Partition)

  ostream& operator<<(ostream& stream, cPartition* Partition) {
    stream << setfill(' ')
           << Partition->Name << "::"
           << '<'  << setw(9) << fixed << setprecision(3) << Partition->RLo
           << ", " << setw(9) << fixed << setprecision(3) << Partition->RHi
           << ", " << setw(5) << Partition->List->GetSize()
           << '>';
  return stream;
  }; // ostream& operator<<(ostream& stream, cPartition* Partition)
