/* 
 * File:   partition.cpp
 * Author: skidmarks
 * 
 * Created on March 3, 2009, 11:14 AM
 */

# include "../Headers/stdafx.h"

# include <cmath>
# include <cstdlib>
# include <ctime>
# include <iomanip>
# include <sstream>
# include "../Headers/Debug.h"
# include "../Headers/common.h"
# include "../Headers/create.h"
# include "../Headers/global.h"
# include "../Headers/utility.h"

//cCreate::cCreate(const cCreate& orig) {
//}

//cCreate::~cCreate() {
// insert code to remove Partition & Partition[i].List
//}

//******************************************************************************
//
//  NAME: BinaryOutput
//
//  Arguments
//    suffix  filename suffix
//
//  DESCRIPTION
//    Output the binary format of the input data.
//
//  ALGORITHM
//    The output is a nested set of structures with format:
//      (size <objects>)
//
//    The detailed format is:
//      <size <partition<size <List>>
//            <partition<size <List>>
//                      o o o
//            <partition<size <List>>
//      >
//
//      Where:
//         partition  == <low bound, high bound>
//         List       == <reference, reference .. reference>
//         low  bound == double
//         high bound == double
//         reference  == long
//         size       == long
//
//  NOTES
//
// UPDATE HISTORY
//     Date        Name         Description
// 10 Jul 2009  A. Schwarz      OIS
//******************************************************************************
void cCreate::BinaryOutput(string suffix) {
   struct {
      double Low;                                  // Low  Bound of Partition
      double High;                                 // High Bound of Paritition
      long   Size;                                 // Number of entries
   } Prefix;
   long      PSize        = Partitions->GetSize(); // Number of Partitions
   long      MaxEntries   = 0;                     // Maximum number of entries
   ofstream  stream;                               // Output Binary File

   if (Open(stream, File.Basename + suffix, ios::out | ios::binary)) exit(0x0101);

   for(long i = 0; i < PSize; i++)
      MaxEntries = max(MaxEntries, (*Partitions)[i].GetSize());

   long* List = new long[MaxEntries];

   stream.write((char*)&PSize, sizeof(PSize));

   for(long i = 0; i < PSize; i++) {
      Prefix.Size = (*Partitions)[i].GetSize();
      Prefix.Low  = (*Partitions)[i].RLo;
      Prefix.High = (*Partitions)[i].RHi;
      for(long j = 0; j < Prefix.Size; j++) List[j] = (*Partitions)[i][j].Datum;
      stream.write((char*)&Prefix, sizeof(Prefix));
      stream.write((char*)List, Prefix.Size * sizeof(long));
   }

   stream.close();

}; // cCreate::BinaryOutput

//******************************************************************************
//
//  NAME: CompareLong
//
//  Arguments
//    INPUT One  Pointer to Arg1 of comparand
//          Two  Pointer to Arg2 of comparand
//
//  DESCRIPTION
//    Used in qsort to determine the relationship between two values. See qsort
//    documentation.
//
//  ALGORITHM
//    One  < Two  return -1
//    One == Two  return  0
//    One  > Two  return +1
//
//******************************************************************************

 int CompareLong(const void* One, const void* Two) {
    const cRange* R1 = (*((cRange const **)One));
    const cRange* R2 = (*((cRange const **)Two));
    const long& Arg1 = R1->Datum;
    const long& Arg2 = R2->Datum;
    //if (cDebug::is_Debug()) 
    //   Log << "CompareLong(" << hex << setfill(' ')
    //       << One << ", " << Two << ") ";
    //   Log.flush();
    //   Log << "  " << R2 << " " << Arg2;
    //   Log.flush();
    //   Log << " "  << R1 << " " << Arg1 << endl;
    //   Log.flush();
    return (Arg1 < Arg2)? -1: (Arg1 == Arg2)? 0: 1;
}

//******************************************************************************
//
//  NAME: Create_DataList
//
//  Arguments
//    INPUT Stack Stack of user supplied data.
//
//  DESCRIPTION
//    An array of user supplied data is created. The user data allows the user
//    to designate an identifier to be associated with each range. Each partition
//    contains a list of the identifies which can be contained in the partition in
//    sorted order.
//
//  ALGORITHM
//    The passed array contain all of the identifiers that are associated with
//    the partition. Space is acquired from the heap and populated with this
//    data. The data is then sorted.
//
//******************************************************************************


cList*  cCreate::Create_DataList(cStack& Stack) {
   DebugEnter("Create_DataList");
//DEBUG    Stack.PrintLists();
    long const Size = Stack.GetSize();           // Number of user data in partition
    cCell*     Cell = &Stack.ResetNext();        // Current Cell
    cRange**   List = new cRange*[Size];         // User data in partition

    if (Size > 0) {
       for(long i = 0; i < Size; i++) {
          Cell    = &Cell->Right();
          List[i] = &Cell->Info();
          if (is_Debug()) {
             ostringstream stream(ios::out);
             stream << setw(5) << i << ": " << hex << setfill(' ') << &List[i] << ": " << List[i];
             DebugData(stream.str());
          }
       }
       qsort(&List[0], Size, sizeof(List[0]), CompareLong);
    }

    DebugExit("Create_DataList");
    return new cList(Size, List);
} // void cCreate::Create_DataList(cStack Stack)

//*****************************************************************************
// NAME: Detail
//
// INPUTS  none
//
// OUTPUT  none
//
// RETURN  none
//
// DESCRIPTION
//    Outputs a list of id's associated with each partition. The id's are 
//    defined in the ranges input to the partitioning process.
//
// UPDATE HISTORY
//     Date        Name         Description
// 19 Apr 2009  A. Schwarz      OIS
//*****************************************************************************

void cCreate::Detail() {
   DebugEnter("cCreate::Detail()");
   Stdout << '\f' << endl << "Partition Listing for "  << Name << " partition" << endl << endl;
   for(long i = 0; i < Partitions->GetSize(); i++) {
      Stdout << fixed << setprecision(4)
             << "["  << setw(13) << (*Partitions)[i].RLo
             << ", " << setw(13) << (*Partitions)[i].RHi
             << ") : ";
      for(long j = 0; j < (*Partitions)[i].GetSize();) {
         long k = j + 10;
         for(; j < min(k, (*Partitions)[i].GetSize()); j++) {
            Stdout << setw(7) << (*Partitions)[i][j].Info();
         }
         Stdout << endl;
         if (j < (*Partitions)[i].GetSize()) Stdout << setw(33) << "";
      }
      if (!(*Partitions)[i].GetSize()) Stdout << endl;
   }
   DebugExit("cCreate::Detail()");
}; // cCreate::Detail(cPartition_List& Partition)

//*****************************************************************************
// NAME: Histogram
//
// INPUTS  none
//
// OUTPUT  none
//
// RETURN  none
//
// DESCRIPTION
//    Provide a simple, visual representation of the number of id's in each
//    partition. A hash mark (#) is output for the proportional number of
//    id's in a partition. Tthe hash mark is rounded up to the nearest integer
//    (all partitions with an entry have at least 1 hash mark).
//
// UPDATE HISTORY
//     Date        Name         Description
// 19 Apr 2009  A. Schwarz      OIS
//*****************************************************************************

void cCreate::Histogram() {
   DebugEnter("cCreate::Histogram()");
   enum { MAXWIDTH = 50 };                       // Maximum marks/line
   long      EntryRange;                         // Bandwidth of entries
   long      MaxEntries   = -1;                  // Maximum Entries/Partition
   long      MinEntries   = LONG_MAX;            // Minimum Entries/Partition
   long      EntriesPerMark;
 
   for(long i = 0; i < Partitions->GetSize(); i++) {
      long NumEntries = (*Partitions)[i].GetSize();
      MaxEntries     = max(MaxEntries, NumEntries);
//      MinEntries     = min(MinEntries, NumEntries);
   }
//   EntryRange     = MaxEntries - MinEntries + 1;
   EntryRange     = MaxEntries + 1;
   EntriesPerMark = (EntryRange + MAXWIDTH - 1)/MAXWIDTH;
   Stdout << '\f' << endl  << "Histogram for "  << Name << " partition" 
                           << "  Entries per Hash Mark: " << EntriesPerMark
                           << endl << endl;

   for(long i = 0; i < Partitions->GetSize(); i++) {
      long Width = ((*Partitions)[i].GetSize() + EntriesPerMark - 1)/EntriesPerMark;
      Stdout << fixed << setprecision(4) << setfill(' ')
             << "["  << setw(13) << (*Partitions)[i].RLo
             << ", " << setw(13) << (*Partitions)[i].RHi
             << ") : " << setfill('#') << setw(Width) << "" << endl;
   }
   Stdout.fill(' ');
   DebugExit("cCreate::Histogram()");
}; // cCreate::Histogram(cPartition_List& Partition)

//*****************************************************************************
// NAME: Statistics
//
// INPUTS  none
//
// OUTPUT  none
//
// RETURN  none
//
// DESCRIPTION
//    Output general statistics associated with the partitioning type performed.
//    The statistics provide overall characteristics of the partitioning
//    process with a 'goodness' measure given by the percent of space wasted.
//    Zero percent space wasted is the best measure.
//
// UPDATE HISTORY
//     Date        Name         Description
// 19 Apr 2009  A. Schwarz      OIS
//*****************************************************************************

void cCreate::Statistics() {
   DebugEnter("cCreate::Statistics(" + Name + ")");
    clock_t   EndTime      = clock();            // Clock tick to here
    time_t    t            = time(NULL);         //
    string    Time(ctime(&t));                   // "www Mmm dd hh:mm:ss yyyy"
    double    MaxEntries   = -1.0;               // Maximum Entries/Partition
    double    MinEntries   = HUGE_VAL;           // Minimum Entries/Partition
    double    PercentUsed  = 0;                  // Percent of space used
    long      PSize        = Partitions->GetSize(); // Number of Partitions
    double    TotalEntries = 0.0;                // Sum of entries in all Parittions
 
    for(long i = 0; i < PSize; i++) {
       double NumEntries = (*Partitions)[i].GetSize();
       TotalEntries  += NumEntries;
       MaxEntries     = max(MaxEntries, NumEntries);
       MinEntries     = min(MinEntries, NumEntries);
       double RHi     = (*Partitions)[i].RHi;
       double RLo     = (*Partitions)[i].RLo;
       long   Size    = (*Partitions)[i].GetSize();
       double UsedSpace= 0;                      // Total space used
       if ( Size > 0 ) {
          double Difference = RHi - RLo;
          double Percent    = 100.0;
          if (Difference > 0.0) {
             for(long j = 0; j < Size; j++) {
                UsedSpace += min(RHi, (*Partitions)[i][j].RHi) - max(RLo, (*Partitions)[i][j].RLo);
                if (is_Debug()) 
                   Log << fixed << setprecision(3)
                       << "     [" << RHi << ", " << RLo << "]"
                       << " " << max(RLo, (*Partitions)[i][j].RLo)
                       << " " << min(RHi, (*Partitions)[i][j].RHi)
                       << " " << (*Partitions)[i]
                       << endl;
             }
             Percent    = (UsedSpace/Size * 100.)/Difference;
          }
          PercentUsed      += Percent;
          if (is_Debug()) 
             Log << fixed << setprecision(3)
             << "  "  << setw(9) << UsedSpace  
             << "  "  << setw(9) << Percent  
             << "  "  << setw(9) << PercentUsed
             << "  "  << setw(7) << Size
             << "  "  << (*Partitions)[i]
          << endl;
       }
    }
    double Bandwidth = (*Partitions)[(*Partitions).GetSize()-1].RHi - (*Partitions)[0].RLo;
    PercentUsed      = PercentUsed/PSize;

    Stdout << fixed << setprecision(2) << '\f'                                     << endl
           << "Statistics for " << Name << " partition"                    << endl << endl
           << "Date: " << Time.substr(8,2) << " " 
                       << Time.substr(4,3) << " " 
                       << Time.substr(20,4)                                        << endl
           << "Time: " << Time.substr(11,8)                                        << endl
           << "Elapsed Time: " << setprecision(3) << diffclock(EndTime, StartTime) << endl << endl
           << "Summary results for " << Filename << endl
           << "  Total Input Intervals:     " << setw(8) << Bounds->GetSize()/2    << endl
           << "  Total Partitions:          " << setw(8) << Partitions->GetSize()  << endl
           << "  Average Entries/Partition: " << setw(8) << TotalEntries/Partitions->GetSize() << endl
           << "  Maximum Entries/Partition: " << setw(8) << MaxEntries             << endl
           << "  Minimum Entries/Partition: " << setw(8) << MinEntries             << endl
           << "  Percent Used Space:        " << setw(8) << PercentUsed            << endl
           << "  Percent Wasted Space:      " << setw(8) << 100.0 - PercentUsed    << endl;
   DebugExit("cCreate::Statistics(" + Name + ")");
}; // cCreate::Statistics()
