# include "../Headers/StdAfx.h"
# include <ctime>
# include <ios>
# include "../Headers/generate.h"
# include "../Headers/global.h"
# include "../Headers/utility.h"

using namespace std;

//******************************************************************************
//  NAME: ComparesData
//
// INPUT   One  Pointer to the first Data Interval
//          Two  Pointer to the second Data Interval
//
// OUTPUT  none
//
//  RETURN  
//    One  < Two  return -1
//    One == Two  return  0
//    One  > Two  return +1
//
// DESCRIPTION
//    Used in qsort to sort the input data intervals by range. See qsort
//    documentation. This is a 'skeleton' function developed to allow access to
//    a public class function.
//
// UPDATE HISTORY
//     Date        Name         Description
// 27 May 2009  A. Schwarz      OIS
//******************************************************************************
int ComparesData(const void* One, const void* Two) {
   return cGenerate::ComparesData(One, Two);
} // ComparesData(const void* One, const void* Two)

//*****************************************************************************
//  NAME: ComparePercent
//
// INPUT   One  Pointer to the first Data Interval
//          Two  Pointer to the second Data Interval
//
// OUTPUT  none
//
//  RETURN  
//    One  < Two  return -1
//    One == Two  return  0
//    One  > Two  return +1
//
// DESCRIPTION
//    Used in qsort to sort the data intervals by percent. See qsort
//    documentation. This is a 'skeleton' function developed to allow access to
//    a public class function.
//
// UPDATE HISTORY
//     Date        Name         Description
// 27 May 2009  A. Schwarz      OIS
//*****************************************************************************

int ComparePercent(const void* One, const void* Two) {
   return cGenerate::ComparePercent(One, Two);
} // ComparePercent(const void* One, const void* Two)

//******************************************************************************
//  NAME: cGenerate::ComparesData
//
// INPUT   One  Pointer to the first Data Interval
//          Two  Pointer to the second Data Interval
//
// OUTPUT  none
//
//  RETURN  
//    One  < Two  return -1
//    One == Two  return  0
//    One  > Two  return +1
//
// DESCRIPTION
//    Used in qsort to sort the input data intervals by range. Compares two
//    data interval ranges.
//
// UPDATE HISTORY
//     Date        Name         Description
// 27 May 2009  A. Schwarz      OIS
//*****************************************************************************

int cGenerate::ComparesData(const void* One, const void* Two) {
   const sData& Arg1 = *(sData const *)One;
   const sData& Arg2 = *(sData const *)Two;
   if      (Arg1.Flow  < Arg2.Flow) return -1;
   else if (Arg1.Flow  > Arg2.Flow) return +1;
   else if (Arg1.Flow == Arg2.Flow) {
     if      (Arg2.FHigh  < Arg2.FHigh) return -1;
     else if (Arg1.FHigh  > Arg2.FHigh) return +1;
   }
   return 0;
} // cGenerate::ComparesData(const void* One, const void* Two)

//*****************************************************************************
// NAME: cGenerate::ComparePercent
//
// INPUT   One  Pointer to the first Data Interval
//          Two  Pointer to the second Data Interval
//
// OUTPUT  none
//
//  RETURN  
//    One  < Two  return -1
//    One == Two  return  0
//    One  > Two  return +1
//
// DESCRIPTION
//    Used in qsort to sort the input data intervals by range. Compares by
//    percent and sorts in descending order. (Descending order is needed
//    during output interval generation.)
//    
//
// UPDATE HISTORY
//     Date        Name         Description
// 27 May 2009  A. Schwarz      OIS
//*****************************************************************************

int cGenerate::ComparePercent(const void* One, const void* Two) {
   const sData& Arg1 = *(sData const *)One;
   const sData& Arg2 = *(sData const *)Two;
   return (Arg1.Percent < Arg2.Percent)? +1: (Arg1.Percent == Arg2.Percent)? 0: -1;
} // cGenerate::ComparePercent(const void* One, const void* Two)

//*****************************************************************************
// NAME: cGenerate::cGenerate
//
// INPUTS  Options Generate options from the commmand line
//         Files   Contains all the filenames contained in the command line
//
// OUTPUT  none
//
// RETURN  none
//
// DESCRIPTION
//    Constructor. Initializses class constants and variables. Note that
//    only the first seed in the pseudo-random number generator is used.
//
// UPDATE HISTORY
//     Date        Name         Description
// 27 May 2009  A. Schwarz      OIS
//*****************************************************************************

cGenerate::cGenerate(sGenerate Options, sFiles& File)
                                                          : Filename(File.Interval.c_str())
                                                          , IntervalFile(File.Generate.c_str())
                                                          , Default(0.0, 10000.0, 250.0, 100.)
                                                          , DefaultStat(0.0, 10000.0, 10000.0, 100.)
                                                          , InputStat(+LDBL_MIN, -LDBL_MIN, 0.0, 0.0)
                                                          , Statistic(512)
{
   for(long i = 0; i < 6; i++) Seed[i] = Options.Seed;
//   Seed[0]      = Options.Seed;
   Width        = Options.Width;
   Count        = Options.Count;

   if (!Rng.SetPackageSeed(Seed)) {
      cout << "INTERNAL ERROR: Pseudo-Random Number Generator seed rejected" << endl;
      exit(0x1012);
   }
   Rng.SetSeed(Seed);
   Rng.ResetStartStream();
}

//*****************************************************************************
// NAME: cGenerate::~cGenerate
//
// INPUTS  none
//
// OUTPUT  none
//
// RETURN  none
//
// DESCRIPTION
//    Class Destructor.
//
// UPDATE HISTORY
//     Date        Name         Description
// 27 May 2009  A. Schwarz      OIS
//*****************************************************************************

cGenerate::~cGenerate(void) {
}

//*****************************************************************************
// NAME: cGenerate::CheckInput
//
// INPUTS  none
//
// OUTPUT  none
//
// RETURN  true  Error detected
//         false inerrant
//
// DESCRIPTION
//    Validates the user supplied input data for correctness. If no user input
//    file is provided the system generates an (always correct) single entry
//    representing the interval to be used in generating cvorrect output.
//    Checks that no interval overlaps another. 
//
// NOTE:
//  1. At entry the user provided input is sorted by range. This allows an
//     assumption to be made that if any in any two successive intervals
//     the High bound of the earlier interval must be less than or equal to
//     the Low bound of it's successor interval.
//
// UPDATE HISTORY
//     Date        Name         Description
// 27 May 2009  A. Schwarz      OIS
//*****************************************************************************

bool   cGenerate::CheckInput() {                 // Check overlapping intervals
   bool       ErrorFlag = false;
   const long Size      = Statistic.GetSize();
   cDebug::DebugEnter("cGenerate::CheckInput");
   for(long i = 0; i < Size - 1; i++) {
      sData* DataI = &Statistic[i];
      for(long j = i+1; i < Statistic.GetSize(); j++) {
         sData* DataJ = &Statistic[j];
         if (is_Debug()) {
            ostringstream stream(ios::out);
            stream << *DataI << "  " << *DataJ;
            DebugData(stream.str());
         }
         if (DataJ->Flow > DataI->FHigh) break;
         ErrorFlag = true;
         Log << "ERROR: Overlapping Intervals [" 
             << DataI->Flow << ", " << DataI->FHigh << "] and ["
             << DataJ->Flow << ", " << DataJ->FHigh << "]" << endl;
      }
   }
   cDebug::DebugExit("cGenerate::CheckInput");
   return ErrorFlag;
} // cGenerate::CheckInput()

//*****************************************************************************
// NAME: cGenerate::CreateIntervals
//
// INPUTS  none
//
// OUTPUT  none
//
// RETURN  true  Error detected
//         false No errors
//
// DESCRIPTION
//    Creates a set of intervals with percents to be used in generating the 
//    exemplar input file for partitioning. At termination a 100% utilizaton
//    percent is represented in the interval set used in the generation function.
// 
//    The algorithm proceeds by comparing two successive intervals. If the
//    intervals are not adjacent:
//
//       I1[low, high]
//       I2[low, high]
//
//    are adjacent when I1.high == I2.low, I1.low < I2.high.
//
//    then an interval is constructed so that I1 is adjacent to I3 which is
//    adjacent to I2:
//
//       I1[low, high]
//       I3[low, high]
//       I2[low, high]
//
//    I1.high == I3.low & I3.high = I2.low.
//
//    This comparison stops when the last input interval is compared. 
//
//    The resultant created intervals are assigned a generation percentage
//    figure governed by:
//
//     1. A uniform distribution is assumed for the generation intervals.
//     2. The allocation per generation interval is made relative toe the
//        size of the interval relative to the total size of all generation
//        intervals.
//
//    At termination the generation interval set is dense and the total
//    generation percentage is equal to 100% (100% of the uniform pseudo-
//    random number output is used).
//
// NOTE:
//  1. The generation intervals are sorted by range prior to entry.
//  2. The algorithm applies only if less than 100% of the generation percentage
//     is used. If 100% of the output of the pseudo-random number generator
//     is used, then the algorithm is not used. As a byproduct, this allows
//     the user to specify non-contiguous ranges for generation of partitioning
//     ranges.
//
// UPDATE HISTORY
//     Date        Name         Description
// 27 May 2009  A. Schwarz      OIS
//*****************************************************************************

bool   cGenerate::CreateIntervals() {            // Generate intervals for 100% coverage
   static const
   double DELTA      = 1.0E-4;                   // Minimum percent granularity
   double Percent    = 0.0;                      // Cumulative percent
   double TotalWidth = 0.0;                      // Sum of created interval widths
   long   MaxNdx     = Statistic.GetSize();      // Current high-water mark
   bool   ErrorFlag  = false;

   cDebug::DebugEnter("cGenerate::CreateIntervals");

   if (InputStat.Percent < (100.0 - DELTA)) {
      for(long i = 0; i < MaxNdx - 1; i++) {     // Create required intervals
         if (is_Debug()) {
            ostringstream stream(ios::out);
            stream << setw(4) << i << ": "
                   << Statistic[i] << " "
                   << Statistic[i+1];
            DebugData(stream.str());
         }
         if (Statistic[i].FHigh < Statistic[i+1].Flow) {
            sData& DataI   = Statistic[i];
            sData& DataJ   = Statistic[i+1];
            double Width   = DataJ.Flow - DataI.FHigh;
            sData Data(DataI.FHigh, DataJ.Flow, Width, 0.0);
            if (is_Debug()) {
               ostringstream stream(ios::out);
               stream << "   " << Data;
               DebugData(stream.str());
            }
            Statistic      = Data;
            TotalWidth    += Width;
         }
      }
      if (MaxNdx == Statistic.GetSize()) {
         ErrorFlag = true;
         Log << "ERROR: Less than 100% distribution and unable to create intervals" << endl;
      } else {
         Percent     = (100.0 - InputStat.Percent);
         for(long i = MaxNdx; i < Statistic.GetSize(); i++)
            Statistic[i].Percent =  Percent * (Statistic[i].Width/TotalWidth);
      }
   }
   cDebug::DebugExit("cGenerate::CreateIntervals");
   return ErrorFlag;
} // cGenerate::CreateIntervals()

//*****************************************************************************
// NAME: cGenerate::Generate
//
// INPUTS  none
//
// OUTPUT  none
//
// RETURN  none
//
// DESCRIPTION
//    Generate partitioning intervals. The output of this function are the
//    intervals to be used by the partitioning algorithm with the distribution
//
//     1. The algorithm uses the pseudo-random number generator as an input
//        percentage, 
//     2. The distribution percentages are searched.
//     3. When the cumulative distribution percentages exceed the generated
//        percentage, the search stops.
//     4. The generation distribution used is the largest distribution with
//        a cumulative percentage greater than or equal to the generated
//        percentage.
//     5. The range start (low) is calculated by:
//        a. Using the pseudo-random number generator to calculate a 
//           percentage.
//        b. Multiplying this number by the Width of the distribution
//           interval.
//        c. Adding the distribution range start to the calculation in b.
//     6. The generation range stop (high) is calculated by:
//        a. Using the pseudo-random number generator to calculate a
//           percentage.
//        b. Multiplying this number by the distribution Width (given
//           in the distribution interval). This Width does not have
//           to be the interval Width.
//        c. Addiing the range start to b.
//        d. Taking the minimum of the range stop (high) and c.
//
// NOTE:
//  1. The input distribution intervals are sorted by percentage.
//  2. The pseudo-random number generator is a uniform random number
//     generator with are 2**53 period.
//  3. The output partitioning intervals are distributed according to 
//     the distributions contained in each distribution intervals.
//  4. There is no guarantee of contiguity on the output partitioning
//     intervals.
//
// UPDATE HISTORY
//     Date        Name         Description
// 27 May 2009  A. Schwarz      OIS
//*****************************************************************************

void   cGenerate::Generate() {                   // Generate output gizmo's
   DebugEnter("cGenerate::Generate");
   long const Size = Statistic.GetSize();
   for(long i = 0; i < Count; i++) {
      double Percent    = Rng.RandU01() * 100.0;
      double Cumulative = 0.0;
      long   j;
      for(j = 0; j < Size; j++) {
         if (is_Debug()) {
            ostringstream stream(ios::out);
            stream << setw(4) << i << ", " 
                   << setw(4) << j << setfill(' ')
                   << " " << setw(8) << setprecision(6) << Percent
                   << " " << setw(8) << setprecision(6) << Cumulative
                   << " " << Statistic[j];
            DebugData(stream.str());
         }
         if (Percent <= Statistic[j].Percent + Cumulative) break;
         Cumulative   += Statistic[j].Percent;
      }
      long Ndx = min(j, Size - 1);
      double Flow  = Statistic[j].Flow + (Statistic[j].FHigh - Statistic[j].Flow)*Rng.RandU01();
      double FHigh = Flow + min(Statistic[j].Width, (Statistic[j].FHigh - Flow))*Rng.RandU01();
      Stdout << setw(11) << setprecision(3) << fixed << Flow  << ", "
             << setw(11) << setprecision(3) << fixed << FHigh << ", "
             << setw(9) << i
             << endl;
   }
   DebugExit("cGenerate::Generate");
} // cGenerate::Generate()

//*****************************************************************************
// NAME: cGenerate::Parse
//
// INPUTS  ParseLine Input line to be parsed
//
// OUTPUT  none
//
// RETURN  true  Errors found
//         false No errors
//
// DESCRIPTION
//    Parses: (Floyd-Evans parsing method)
//      number,number,number,number representing [low, high, width, percentage]
//      number,number,,number       representing [low, high,      , percentage]
//
//    It does this by iterating through each of the expected input values
//    and parsing a number (::Parse). Each return is compared against the
//    expected value with errors in input reported in the log file. Of note
//    is that a field with no width (',,') is not a number and ::Parse reports
//    this as a failure.
//
//    If the input distribution interval is successfully read, then check are
//    made to ensure that each input value is correct and that the cumulative
//    distribution percentage does not exceed 100%.
//
//    The maximum and minimum range values (low, high) are determined during
//    input.
//
// UPDATE HISTORY
//     Date        Name         Description
// 27 May 2009  A. Schwarz      OIS
//*****************************************************************************

bool   cGenerate::Parse(string& ParseLine) {     // Parse input line
   enum eValue { eFlow                           // Lower interval bound
               , eFHigh                          // Upper interval bound
               , eWidth                          // Maximum width of generated intervals
               , ePercent                        // Percent of output in this range
               };

   static const
   string       Field[] = { "Lower Boundary", "Upper Boundary", "Width", "Percent" };

   sData Data;
   bool  ErrorFlag = false;
   bool  WidthMissingFlag = false;               // Width Missing
   long  i;                                      // indexing variable
   long  Start = 0;
   long  Stop  = 0;
   long  Substr= 0;

   cDebug::DebugEnter("cGenerate::Parse");

   for(i = eFlow; (i <= ePercent); i++) {        // Parse values (top-down parse)
      ErrorFlag = ::Parse(ParseLine.substr(Substr), Start, Stop);
      Start += Substr;
      Stop  += Substr;
      Substr = Stop + 1;
      if (is_Debug()) {
         ostringstream stream(ios::out);
         stream << "[" << Start << ", " << Stop << ", " << Substr << "] 0x"
                << hex << (unsigned)ParseLine[Stop] << " :: "
                << ParseLine.substr(Start, (Substr - Start));
         DebugData(stream.str());
      }
      if (ParseLine[Stop] == ',') ParseLine[Stop] = ' ';
      
      if (ErrorFlag) {
        if (Start == Stop) {                  //Is field blank
          if (ParseLine[Stop] != ' ' ) break;
          if (i == eWidth) WidthMissingFlag = true;
          else {
            ErrorFlag = true;
            Log << "ERROR: " << Field[i] << " value missing" << endl;
            break;
          }
        } else if (ParseLine[Stop] < ' ') break;
      }
   }

   if (!ErrorFlag) {
     if (i < ePercent) {
       ErrorFlag = true;
       Log << "ERROR: Missing " << Field[i+1] << endl;
     } 
     for(i = Stop; (ParseLine[i] == ' '); i++);
     if (ParseLine[i] > ' ') {
       ErrorFlag = true;
       Log << "ERROR: Extra data following Identifier" << endl;
     }
   }

   if (!ErrorFlag) {
      istringstream Line(ParseLine);
      Line >> skipws >> Data.Flow >> Data.FHigh >> Data.Width >> Data.Percent;
      if (WidthMissingFlag) {                    // ParseLine = [low, high,, percent]
         Data.Percent = Data.Width;
         Data.Width   = (Data.FHigh - Data.Flow);
      }
      ErrorFlag = true;
      if (Data.Percent <= 0.0) {
         Log << "ERROR: Percent must be > 0.0 '" << Data.Percent << "'" << endl;
      } else if (Data.Percent > 100.0) {
         Log << "ERROR: Percent must be <= 100.0 '" << Data.Percent << "'" << endl;
      } 
      if ((Data.FHigh   == -LDBL_MIN) ||
          (Data.Flow    == -LDBL_MIN) ||
          (Data.Width   == -LDBL_MIN)) {
         Log << "ERROR: Upper bound, Lower Bound or Width not successfully input" << endl;
      } else if (Data.Flow  > Data.FHigh) {
         Log << "ERROR: Lower Bound must be less than Upper Bound" << endl;
      } else if (Data.Flow == Data.FHigh) {
         Log << "ERROR: Lower Bound equal Upper Bound" << endl;
      } else if (Data.Width < 0.0 ) {
         Log << "ERROR: Width must be positive" << endl;
      } else if (Data.Width == 0.0) {
         Log << "ERROR: Zero length Width not allowed" << endl;
      } else if (Data.Width > (Data.FHigh - Data.Flow)) {
         Log << "ERROR: Width can not be greater than Interval Width" << endl;
      } else {
         Statistic          = Data;              // Append new entry
         InputStat.Flow     = min(Data.Flow,  InputStat.Flow);
         InputStat.FHigh    = max(Data.FHigh, InputStat.FHigh);
         InputStat.Percent += Data.Percent;
         ErrorFlag = false;
         if (InputStat.Percent > 100.0) {
            ErrorFlag = true;
            Log << "ERROR: Total percent must be <= 100.0" << endl;
         }
      }
   }
   cDebug::DebugExit("cGenerate::Parse");
   return ErrorFlag;
}; // cGenerate::Parse(string ParseLine)

//*****************************************************************************
// NAME: cGenerate::ReadIntervals
//
// INPUTS  none
//
// OUTPUT  none
//
// RETURN  true  Error found
//         false No error found
//
// DESCRIPTION
//    Inputs and parses the distribution intervals. After all intervals are
//    input, the Width of the interval ranges is calculated. This 'Width'
//    assumes that the ranges are contiguous.
//
// UPDATE HISTORY
//     Date        Name         Description
// 27 May 2009  A. Schwarz      OIS
//*****************************************************************************

bool cGenerate::ReadIntervals() {
   cDebug::DebugEnter("cGenerate::ReadIntervals");
   enum { MAXLINESIZE = 384 };
   bool  ErrorFlag = false;
   char  Line[MAXLINESIZE];
   long  LineNo    = 0;
   while(!Stdin.eof()) {
      long i = 0;
      LineNo++;
      Stdin.getline(Line, MAXLINESIZE);
      string Input(Line);
//DEBUG      Log << setw(6) << LineNo << ": " << Input << endl;
      string ParseLine = StripComments(Input);
      if (ParseLine.length() <= 0) continue;
      ErrorFlag     |= Parse(ParseLine);
   }
   InputStat.Width   = InputStat.FHigh - InputStat.Flow;

   if (!LineNo) {
      Log << "ERROR: Input Interval file empty" << IntervalFile << endl;
   } else if (ErrorFlag) {
      Log << "ERROR: One or more errors detected in input" << endl;
   }
   cDebug::DebugExit("cGenerate::ReadIntervals");
   return ErrorFlag;
}; // cGenerate::ReadIntervals()

//*****************************************************************************
// NAME: cGenerate::Run
//
// INPUTS  none
//
// OUTPUT  none
//
// RETURN  true  Error found
//         false No error found
//
// DESCRIPTION
//    Driver for inputing the generation distribution intervals and generating
//    the partition intervals. Ensures that entry conditions for each function
//    are satisfied (the previous function was inerrant and that the input
//    distribution intervals are sorted correctly).
//
// UPDATE HISTORY
//     Date        Name         Description
// 27 May 2009  A. Schwarz      OIS
//*****************************************************************************

bool cGenerate::Run() {
   cDebug::DebugEnter("cGenerate::Run");
   bool ErrorFlag = false;
   if (IntervalFile[0] == '\0') {
      Statistic          = Default;
      Statistic[0].Width = Width;
      InputStat          = DefaultStat;
   }
   if (!(ErrorFlag = ReadIntervals()))
      qsort(&Statistic[0], Statistic.GetSize(), sizeof(Statistic[0]), ComparesData);

   if (!ErrorFlag) {
      ErrorFlag = true;
      if (!CheckInput()) {
         if (!CreateIntervals()) {
            qsort(&Statistic[0], Statistic.GetSize(), sizeof(Statistic[0]), ComparesData);
            WriteHeader();
            qsort(&Statistic[0], Statistic.GetSize(), sizeof(Statistic[0]), ComparePercent);
            Generate();
            ErrorFlag = false;
         }
      }
   }
   cDebug::DebugExit("cGenerate::Run");
   return ErrorFlag;
}; // cGenerate::Run()

//*****************************************************************************
// NAME: cGenerate::WriteHeader
//
// INPUTS  none
//
// OUTPUT  none
//
// RETURN  none
//
// DESCRIPTION
//    Output the header for the generated partition interval file and outputs
//    the input generating partitions.
//
// UPDATE HISTORY
//     Date        Name         Description
// 27 May 2009  A. Schwarz      OIS
//*****************************************************************************

void cGenerate::WriteHeader() {                  // Output banner
   time_t    t    = time(NULL);
   string    Time(ctime(&t));                    // "www Mmm dd hh:mm:ss yyyy"
   string    IFile = (IntervalFile[0] == '\0')? "None": IntervalFile;

   Stdout << "#########################################################"  << endl
          << "# Filename:      " << Filename                              << endl
          << "# Date:          " << Time.substr(4, 6) << ", " << Time.substr(20, 4) << endl
          << "# Time:          " << Time.substr(11, 8)                    << endl
          << "#                  Generation Parameters"                   << endl
          << "# Seed:          " << Seed[0]                               << endl
          << "# Intervals:     " << Count                                 << endl
          << "# Interval File: " << IFile                                 << endl
          << "#                  Generation Statistics"                   << endl
          << "# Low Bound High Bound  Max Width Percent"                  << endl;
   for(long i = 0; i < Statistic.GetSize(); i++)
      Stdout << "# " << setw(9) << Statistic[i].Flow
             << ", " << setw(9) << Statistic[i].FHigh
             << ", " << setw(9) << Statistic[i].Width
             << ", " << setw(6) << Statistic[i].Percent                   << endl;
   Stdout << "#########################################################"  << endl
                                                                          << endl
          << "#  Low Bound   High Bound  Identifier"                      << endl;
}; // cGenerate::WriteHeader()

  
    //----------------------------------------
    //               Streams
    //----------------------------------------

ostream& operator<<(ostream& stream, cGenerate::sData& Data) {
   stream << "<["  << setw(11) << setprecision(3) << fixed << Data.Flow 
          << ", "  << setw(11) << setprecision(3) << fixed << Data.FHigh 
          << "], " << setw(11) << setprecision(3) << fixed << Data.Width
          << ", "  << setw(11) << setprecision(3) << fixed << Data.Percent
          << ">";
   return stream;
}
