/* 
 * File:   irange.cpp
 * Author: Arthur Schwarz
 * 
 * Created on April 13, 2009, 1:23 PM
 */

# include "../Headers/stdafx.h"

# include <cstdlib>
# include <cmath>
# include <string>
# include <ctime>
# include <ios>
# include <iostream>
# include <cstdio>
# include "../Headers/Debug.h"
# include "../Headers/global.h"
# include "../Headers/irange.h"
# include "../Headers/utility.h"

using namespace std;

    //----------------------------------------
    //        CLASS DATA DEFINITIONS
    //----------------------------------------
long const iRange_List::BUFFERSIZE = 10000;      // Input buffer size
long const iRange_List::ERRORCOUNTLIMIT = 100;   // Maximum I/O Errors allowed before termination
//char       sRange::Line[512];                    // Input Line

    //----------------------------------------
    //           DATA DEFINITIONS
    //----------------------------------------

    //----------------------------------------
    //              CODE
    //----------------------------------------


//******************************************************************************
//
//  NAME: CompareBounds
//
//  Arguments
//    INPUT One  Pointer to Arg1 of comparand
//          Two  Pointer to Arg2 of comparand
//
//  DESCRIPTION
//    Used to sort the SortArray. The sorted SortArray is used in Create_IdealParition
//    to determine the best choice for disjoint partitions. The sort arrangement
//    is critical in allowing the correct binary comparison between two elements
//    in the SortArray (SortArray[i - 1] and SortArray[i]).
//
//  ALGORITHM
//    One  < Two  return -1
//    One == Two  return  0
//    One  > Two  return +1
//
//******************************************************************************

 int CompareBounds(const void* One, const void* Two) {
    const cBound& Arg1 = *(cBound const *)One;
    const cBound& Arg2 = *(cBound const *)Two;
    if      ( Arg1.Bound <  Arg2.Bound) return -1;
    else if ( Arg1.Bound >  Arg2.Bound) return +1;
    else if ( Arg1.Bound == Arg2.Bound) {
      if ( Arg1.Type  == Arg2.Type)     return 0;
      else if ((Arg1.Type == cBound::Low) && (Arg2.Type == cBound::High)) return -1;
      else                              return +1;
    }
 }; // int CompareBounds(const void* One, const void* Two)

//*****************************************************************************
// NAME: iRange_List
//
// INPUTS  Path  Full path to the input file.
//         
// OUTPUT  none
//
// RETURN  none
//
// DESCRIPTION
//    Constructor. Opens a temporary file for storage of binary range data.
//    The temporary file is used to support allocation of the complete set
//    of range data in memory, in cRange* Range.
//
// UPDATE HISTORY
//     Date        Name         Description
// 21 Apr 2009  A. Schwarz      OIS
//*****************************************************************************

iRange_List::iRange_List(string Path) : Bounds(NULL), ErrorCount(0)
                                      , Ndx(-1), Ranges(NULL), Size(0) {
    Ranges         = new cRange[BUFFERSIZE];     // create buffer stuff
    time_t    t    = time(NULL);
    char*     Time = ctime(&t);                  // "www Mmm dd hh:mm:ss yyyy"
    for(long i = 4; Time[i]; i++) {              // Skip 'www' 
        if (Time[i] == ' ') Time[i] = '_'; }     // Change ' ' to '_'
    string File     = &Time[4];
    TempFileName    = Path + '/' + File;
    try {
       TempStream.open(TempFileName.c_str(), ios::in | ios::out | ios::binary | ios::trunc);
    } catch (...) {
       cout << "ERROR: Unable to open temporary file" << endl;
       exit(0x2000);
    }
}; // iRange_List::iRange_List()

//*****************************************************************************
// NAME: iRange_List
//
// INPUTS  orig  'this' pointer to an iRange_List object.
//
// OUTPUT  none
//
// RETURN  none
//
// DESCRIPTION
//    Copy constructor. It should never, never be used.
//
// UPDATE HISTORY
//     Date        Name         Description
// 21 Apr 2009  A. Schwarz      OIS
//*****************************************************************************

iRange_List::iRange_List(const iRange_List& orig) {
   exit(0x2020);
}; // iRange_List::iRange_List(const irange& orig)

//*****************************************************************************
// NAME: ~iRange_List
//
// INPUTS  none
//
// OUTPUT  none
//
// RETURN  none
//
// DESCRIPTION
//    Destructor. Deletes the range of input values and closes and deletes
//    the temporary binary file.
//
// UPDATE HISTORY
//     Date        Name         Description
// 21 Apr 2009  A. Schwarz      OIS
//*****************************************************************************

iRange_List::~iRange_List() {
    if (!Ranges) delete [] Ranges;
    if (TempStream.is_open()) TempStream.close();
    try {
      remove(TempFileName.c_str());
    } catch (...) { }
}; // iRange_List::~iRange_List()

//*****************************************************************************
// NAME: ChkIn
//
// INPUTS  Range 
//
// OUTPUT  none
//
// RETURN  true  Error detected
//         false No error detected
//
// DESCRIPTION
//    Check Range for validity. Each inpug interval, [Low, High}, is checked
//    for correctness. Failures are messaged and a failure indication returned.
//
// UPDATE HISTORY
//     Date        Name         Description
// 26 Apr 2009  A. Schwarz      OIS
//*****************************************************************************

bool   iRange_List::ChkIn(cRange Range) {
   bool ErrorFlag = true;
   if (!Stdin) {
      if (Stdin.fail()) Log << "ERROR: Input file bad" << endl;
   } else if (Range.RLo > Range.RHi) {
      Log << "ERROR: Range Low > Range High" << endl;
   } else if (Range.RLo == -LDBL_MIN) {
      if (Range.RHi == -LDBL_MIN) Log << "ERROR: Unable to input Range Low and High" << endl;
      else                        Log << "ERROR: Unable to input Range Low"          << endl;
   } else if (Range.RHi == -LDBL_MIN) Log << "ERROR: Unable to input Range High"    << endl;
     else ErrorFlag = false;
   return ErrorFlag;
}; // iRange_List::ChkIn(sRange)

//*****************************************************************************
// NAME: Create_Bounds
//
// INPUTS  Ranges List of intervals, [Low, High]  
//
// OUTPUT  none
//
// RETURN  Maximum bandwidth of all ranges (max(High) - min(Low))
//
// DESCRIPTION
//    For each bound in the input interval 2 entries are created in the Bounds
//    List, one for Low and one for High. The bounds are tagged with whether
//    the entry is low or high, and the resultant bounds array sorted by value
//    such that lower values precede higher values, and where the values are
//    equal, the low bound precedes the high bound. The resultant array is 
//    used by the partition creation functions.
//
// UPDATE HISTORY
//     Date        Name         Description
// 26 Apr 2009  A. Schwarz      OIS
//*****************************************************************************

double iRange_List::Create_Bounds() {
   double  RLo = HUGE_VAL;                       // Lowest value in range
   double  RHi = -HUGE_VAL;                      // Highest value in range
   Bounds      = new cBound_List(2*GetSize(), "Sorted Array");

   DebugEnter("iRange_List::Create_Bounds");

   for(long i = 0; i < GetSize(); i++) {
      RLo = (RLo > Ranges[i].RLo)? Ranges[i].RLo: RLo;
      RHi = (RHi < Ranges[i].RLo)? Ranges[i].RHi: RHi;
      (*Bounds)[2*i    ].Bound = Ranges[i].RLo;
      (*Bounds)[2*i    ].Type  = cBound::Low;
      (*Bounds)[2*i    ].Datum = &Ranges[i];
      (*Bounds)[2*i + 1].Bound = Ranges[i].RHi;
      (*Bounds)[2*i + 1].Type  = cBound::High;
      (*Bounds)[2*i + 1].Datum = &Ranges[i];
   }

   qsort(&(*Bounds)[0], Bounds->GetSize(), sizeof((*Bounds)[0]), CompareBounds);

   DebugExit("iRange_List::Create_Bounds");

   return (RHi - RLo);
}; // iRange_List::Create_Bounds()

//*****************************************************************************
// NAME: Flush
//
// INPUTS  none
//
// OUTPUT  none
//
// RETURN  none
//
// DESCRIPTION
//    Flushes the Range list to the temporary file, if required. Deletes the
//    Range list then repopulates it with the contents of the temporary binary
//    file. At the termination of this function:
//      1: All ranges are stored into a single, internal, array.
//      2: No  range is erronous.
//      3: The temporary binary file is closed (but not deleted).
//
//    At this point we have the following guarantess:
//       1: Next guarantees TempStream is open on Range overflow.
//       2: Either Tempstream is open or there is no Range overflow.
//
// UPDATE HISTORY
//     Date        Name         Description
// 21 Apr 2009  A. Schwarz      OIS
//*****************************************************************************

void iRange_List::Flush() {                      // Flush buffer, recreate & input all ranges
   Size += Ndx + 1;
   if (Size > BUFFERSIZE) {                      // File contains data 
      if (Ndx > -1) TempStream.write((char*)Ranges, sizeof (cRange) * (Ndx + 1));
      TempStream.seekg(0);
      delete [] Ranges;
      Ranges = NULL;
      try {
         Ranges = new cRange[Size];
      } catch (...) {
         Stdout << "ERROR: Not enough memory for input ranges [" << Size << "]" << endl;
         cout   << "ERROR: Not enough memory for input ranges [" << Size << "]" << endl;
         abort();
      }
      try {
         TempStream.read((char*)Ranges, sizeof (cRange) * Size);
      } catch (...) {
         Stdout << "ERROR: Unable to read all temporary file data [" << Size << "]" << endl;
         cout   << "ERROR: Unable to read all temporary file data [" << Size << "]" << endl;
         abort();
      }
   }
   if (TempStream.is_open()) TempStream.close();
   Ndx = 0;
}; // iRange_List::Flush()

//*****************************************************************************
// NAME: ciRange_3::Parse
//
// INPUTS  none
//
// OUTPUT  none
//
// RETURN  true  Successful
//         false Unsuccessful
//
// DESCRIPTION
//   
//   
//
// UPDATE HISTORY
//     Date        Name         Description
// 21 Apr 2009  A. Schwarz      OIS
//*****************************************************************************

bool   ciRange_3::Parse(string& ParseLine) {     // Parse input line
   enum { eFLow, eFHigh, eID };
   static const
   string       Field[] = { "Lower Boundary", "Upper Boudary", "Identifier" };

   bool     ErrorFlag = false;
   long     i;                                // Indexing variable
   long     Start = 0;
   long     Stop  = 0;
   long     Substr= 0;

   cDebug::DebugEnter("ciRange_3::Parse");

   for(i = eFLow; (i <= eID); i++) {              // Parse values (top-down parse)

      if ((ErrorFlag = ::Parse(ParseLine.substr(Substr), Start, Stop))) break;
      Start += Substr;
      Stop  += Substr;
      Substr = Stop + 1;
      if (ParseLine[Stop] == ',' ) ParseLine[Stop] = ' ';
      if (ParseLine[Stop] == ',') {
         ErrorFlag = true;
         Log << "ERROR: " << Field[i] << " must have a value" << endl;
         break;
      }
   }
   
   if (!ErrorFlag) {
      for(i = Substr; (ParseLine[i] == ' '); i++);
      if (ParseLine[i] > ' ') {
         ErrorFlag = true;
         Log << "ERROR: Extra data following Identifier" << endl;
      }
   } 
   if (!ErrorFlag) {
      cRange Range;                              // Input  Buffer
      istringstream Line(ParseLine);
      Line >> skipws >> Range.RLo >> Range.RHi >> Range.Datum;
      if (!ChkIn(Range)) *this = Range;
      else ErrorFlag = true;
      if (Range.Datum == LONG_MIN) Log << "Warning: Identifier not successfully input" << endl;
   }
   cDebug::DebugExit("ciRange_3::Parse");
   return ErrorFlag;
}; // ciRange_3::Parse(string ParseLine)


//*****************************************************************************
// NAME: ciRange_3::Read()
//
// INPUTS  none
//
// OUTPUT  none
//
// RETURN  true  Successful
//         false Unsuccessful
//
// DESCRIPTION
//   
//   
//
// UPDATE HISTORY
//     Date        Name         Description
// 21 Apr 2009  A. Schwarz      OIS
//*****************************************************************************

bool ciRange_3::Read() {
   DebugEnter("ciRange_3::Read()");
   enum { MAXLINESIZE = 384 };
   char  Line[MAXLINESIZE];
   bool  ErrorFlag = false;
   long  LineNo    = 0;
   ErrorCount      = 0;                          // No errors at startup

   while (!Stdin.eof() && !Stdin.fail()) {
      long i = 0;
      LineNo++;
      Stdin.getline(Line, MAXLINESIZE);
      Log << setw(6) << LineNo << ": " << Line << endl;
      while(Line[i++]) if (Line[i-1] < ' ') Line[i-1] = ' '; // 0x01 - 0x1F with ' '
      string Input(Line);
      string ParseLine = StripComments(Input) + '\0';
      if (ParseLine.length() <= 1) continue;
      if (!Parse(ParseLine)) {
      } else {
         ErrorCount++;
         if (ErrorCount >= ERRORCOUNTLIMIT) {
            Log << "ERROR: Input terminated. Too many errors [" << ErrorCount << endl;
            break;
         }
      }
   }
   Flush();
   DebugExit("ciRange_3::Read()");
   return (ErrorCount == 0);
}; // ciRange_3::Read()


    //----------------------------------------
    //          Processing Streams
    //----------------------------------------

inline ostream& operator<<(ostream& stream, sRange& Range) {
   stream << setw(9) << Range.RLo << " " << setw(9) << Range.RHi;
   return stream;
}; // ostream& operator<<(ostream& stream, sRange& Range)

inline ostream& operator<<(ostream& stream, sRange_3& Range) {
   stream << setw(9) << Range.RLo << " " << setw(9) << Range.RHi << " " << setw(9) << Range.Datum;
   return stream;
}; // ostream& operator<<(ostream& stream, sRange_3& Range)

