# include "../Headers/stdafx.h"
# include <cstring>

# include "../Headers/option.h"

//*****************************************************************************
// NAME: ChkArg
//
// INPUTS  Option   Definition of Command Line parameterization
//         Argument Option argument
//         ArgFlag  true if argument exists
//
// OUTPUT  ArgFlag  set to false if argument consumed
//         Ndx      Passthrough variable to GetOpts
//         Option   Updated with an input argument
//
// RETURN  0  No error detected
//         1  Errors found
//
// DESCRIPTION
//    Determines if an input argument is required and exists or is optional
//    and exists and consumes the argument by:
//    1: Transfering the argument pointer to the option template (Option.Val)
//    2: Setting the ArgFlag to 'false'
//    
//    If an argument is required and does not exist and error message is 
//    output and an error returned. In all other cases the appropriate
//    option processing function is called (Option.Function).
//
// UPDATE HISTORY
//     Date        Name         Description
// 19 Apr 2009  A. Schwarz      OIS
//*****************************************************************************

long cOption::ChkArg(char* Argument, long& Ndx, bool& ArgFlag) {
   long ErrorCount = 0;

   Empty = false;
   long signature = 4 * (Flag && cOption::eArgument     == cOption::eArgument)
                  + 2 * (Flag && cOption::eArgument_Req == cOption::eArgument_Req)
                  + 1 * (ArgFlag);
   switch(signature) {               //   Arg      Arg
                                     // Allowd Req Found
   case 0x00:                        //   No   No  No
   case 0x01:                        //   No   No  Yes
      ErrorCount = Function(*this);
      break;
   case 0x02:                        //   No   Yes No   (error)
      ErrorCount = 1;
      cout << "ERROR: Required argument not found -" << LongName << endl;
      break;
   case 0x03:                        //   No   Yes Yes 
      Ndx++;
      ArgFlag = false;
      Val     = Argument;
      ErrorCount = Function(*this);
      break;
   case 0x04:                        //   Yes  No  No
      break;
   case 0x05:                        //   Yes  No  Yes
      ArgFlag = false;
      Val     = Argument;
      ErrorCount = Function(*this);
      break;
   case 0x06:                        //   Yes  Yes No   (error)
      ErrorCount = 1;
      cout << "ERROR: Required argument not found -" << LongName << endl;
      break;
   case 0x07:                        //   Yes  Yes Yes
      Ndx++;
      ArgFlag = false;
      Val     = Argument;
      ErrorCount = Function(*this);
      break;
   default:
      ErrorCount = 1;
      cout << "INTERNAL ERROR: Signature wrong(" << signature << ")" << endl;
   }
   return ErrorCount;
}; // cOption::ChkArg(char& Argument)

//*****************************************************************************
// NAME: ChkRequired
//
// INPUTS  None
//
// OUTPUT  None
//
// RETURN true   No Errors found
//        false  Error found
//
// DESCRIPTION
//   .Check each option to see if it is required and has an entry. If it is
//    required and does not have an entry it is an error.
//
// NOTES
//
// UPDATE HISTORY
//     Date        Name         Description
//  1 Jul 2009  A. Schwarz      OIS
//*****************************************************************************
bool     cOption_List::ChkRequired() {
   long ErrorCount = 0;

   for(long i= 0; i < GetSize(); i++)
      if ((Option[i]->Flag & cOption::eRequired) &&
          (Option[i]->Empty)) {
          ErrorCount++;
          cout << "ERROR: Required option missing: -" << Option[i]->LongName << endl;
      }
   return (ErrorCount == 0);
}; // cOption_List::ChkRequired()

//*****************************************************************************
// NAME: CommandLine
//
// INPUTS  Argc   number of arguments in Argv
//         Argv   List of pointers ot char strings represented each argument
//         Option List of options
//
// OUTPUT  Filename Output file for process
//
// RETURN  true:  no fatal errors found
//         false: one or more fatal errors found
//
// DESCRIPTION
//    Command line options and arguments are processed. As part of option
//    process a list of non-option arguments are saved for processing. The
//    first such argument is assumed to be the output file name and an
//    attempt is made to open this file. The rest of the arguments are 
//    ignored (and deserve their ignominious end).
//    
//    If the user specified help (--HELP) or if any errors are discovered,
//    then output file processing is not done.
//
//    It is assumed that the C++ kernel returns arguments (argv[]) which
//    have all embedded blanks removed.
//
// UPDATE HISTORY
//     Date        Name         Description
// 19 Apr 2009  A. Schwarz      OIS
//*****************************************************************************

bool  cOption_List::CommandLine(long const Argc, char* Argv[], string& Filename) {
    string** Filenames = new string*[20];

    for(long i = 0; i < 20; i++) Filenames[i] = NULL;

    long ErrorCount = GetOpts(Argc, Argv, Filenames);

    if (ErrorCount) {
        cout << "Hey bubba, you had " << ErrorCount << " errors. Fix'em!" << endl << endl;
        return (0x1);
    } else {
        if (Filenames[0]) {
            Filename = *Filenames[0];
        }
        long i = 1;
        for(; i < sizeof(Filenames)/sizeof(Filenames[0]); i++) {
            if (!Filenames[i]) break;
            ErrorCount++;
            cout << "Warning " << "command line option ignored " << *Filenames[i] << endl;
        }
    }
    return (ErrorCount)? true: false;
}; // cOption_List::CommandLine(long const Argc, char* Argv[])

//*****************************************************************************
// NAME: GetLong
//
// INPUTS  Argv     Command line option
//         Argument Pointer to command line argument (-x <argumend>)
//         Option   Template for command line options
//
// OUTPUT  Ndx      Passthrough index into command line options/arguments
//         Option   Updated with pointer to argument
//
// RETURN  0  No error detected
//         1  Errors found
//
// DESCRIPTION
//    Processes a single long form option. Compares the input option up to
//    any '=' with each name in the option template - inefficient but
//    efficency is misplaced at this point. When an option is found matching
//    a template long form name a determination is made if an argument may
//    associated with it (either '=' is found or the Argument is non-null or
//    does not begin with '-'). Argument processing is then performed if this
//    is the first time that this option has been processed.
//    
//    Only one option with an argument can be processed in a single call. If 
//    several occurrences of options with the same name are found (either in 
//    a single string or through multiple calls) then there must be an option
//    template for each instance or duplicate instances of the same option will
//    be reported.
//
// NOTES
//    1. Several options with the same name can be stored in the option template.
//       The order of option appearance in the template is significant.
//
// UPDATE HISTORY
//     Date        Name         Description
// 19 Apr 2009  A. Schwarz      OIS
//*****************************************************************************

long   cOption_List::GetLong(char  Argv[], char* Argument, long& Ndx) {
   bool found      = false;
   bool ArgFlag    = (Argument) && (Argument[0] != '-');
   long FalseNdx   = 0;

   for(long i = 0; i < GetSize(); i++) {
//DEBUG      cout << Argv << ": " << Option[i] << endl;
      size_t Size = strlen(Option[i]->LongName);
      if (!strncmp(&Argv[1], Option[i]->LongName, Size))
         if (Option[i]->Empty) {
           if (Argv[Size+1] == '=' ) {
              bool ArgFlag    = true;
              return Option[i]->ChkArg(&Argv[Size+2], FalseNdx, ArgFlag);
           } else {
              bool ArgFlag    = (Argument) && (Argument[0] != '-');
              return Option[i]->ChkArg(Argument, Ndx, ArgFlag);
           }
         } else found = true;
   }

   if (!found) cout << "ERROR: Option not found " << Argv << endl;
   else        cout << "ERROR: Duplicate option " << Argv << endl;

   return 1;
}; // cOption_List::GetLong(char  Argv[], char* Argument, long& Ndx)

//*****************************************************************************
// NAME: GetOpts
//
// INPUTS  Argc   number of arguments in Argv
//         Argv   List of pointers ot char strings represented each argument
//         Option List of options
//
// OUTPUT  Option List of options with values
//         List   List of arguments (not associated with options)
//
// RETURN  number of errors found
//
// DESCRIPTION
//   A single scan is made of the input arguments. The C/C++ kernel parses
//   command line arguments by pointin Argv to blank bracketed values on
//   the line. Argv[0] contains the entire input line. Argv[1 .. Argc-1] is 
//   processed as command line input.
//   
//   If the current Argv value has a leading '-' then it is considered a
//      short form option.
//   If the current Argv value has a leading '--' then it is considered a
//      long form option.
//   Otherwise it is considered as an independent value and stored in a list
//   of such values.
//   
//   NOTES:
//    1 Multiple options with the same name are allowed (-x -x).
//    2 Options can be input in any order (-x -y) == (-y -x)
//    3 Inputs requiring a specific ordering of arguments and options are not
//      treated properly.
//      <program> -x <value> argument2 -y <value> argument2
//          is seen as identical to 
//      <program> -x <value> -y <value> argument1 argument2
//          or any other arrangement.
//    4 <value> must follow an option (-x) and must separated by a space
//    5 Embedded spaces are not allowed.
//    6 Long names are treated identically to short names with the addition
//      that 'NAME=<argument>' is allowed.
//   
//*****************************************************************************

long cOption_List::GetOpts(long const Argc, char* Argv[], string* List[]) {
    long    ErrorCount = 0;                      // Count of errors
    long    k = 0;                               // List index
    for(long i = 1; i < Argc; i++) {
        if (!Argv[i]) { break;
        } else if (Argv[i][0] != '-') {
            List[k++] = new string(Argv[i]);
        } else if (Argv[i][1] == '\0') {         // User error
           ErrorCount++;
           cout << "ERROR: did not find option" << endl;
        } else if (Argv[i][1] == '-' ) {         // Long Name processing
           ErrorCount += GetLong(Argv[i], Argv[i+1], i);
        } else {                                 // Short Name processing
           ErrorCount += GetShrt(Argv[i], Argv[i+1], i);
        }
    }
    return ErrorCount;
} // cOption_List::GetOpts(long const Size, char const* Argv[], char** Arg)

//*****************************************************************************
// NAME: GetShrt
//
// INPUTS  Argv     Command line option
//         Argument Pointer to command line argument (-x <argumend>)
//         Option   Template for command line options
//
// OUTPUT  Ndx      Passthrough index into command line options/arguments
//         Option   Updated with pointer to argument
//
// RETURN  0  No errors found
//        >1  Number of errors found
//
// DESCRIPTION
//    Processes multiple short form option. Compares the input option with each 
//    name in the option template - inefficient but efficency is misplaced at 
//    this point. When an option is found matching a template short form name a 
//    argument processing is performed. Looping continues until the current
//    list of short form arguments is exhausted (-abcd implies a list of four
//    short form arguments).
//    
//    Only one option with an argument can be processed in a single call. If 
//    several occurrences of options with the same name are found (either in 
//    a single string or through multiple calls) then there must be an option
//    template for each instance or duplicate instances of the same option will
//    be reported.
//
// NOTES
//    1. Several options with the same name can be stored in the option template.
//       The order of option appearance in the template is significant.
//
// UPDATE HISTORY
//     Date        Name         Description
// 19 Apr 2009  A. Schwarz      OIS
//*****************************************************************************

long   cOption_List::GetShrt(char  Argv[], char* Argument, long& Ndx) {
   long   i;
   long   k = 0;
   bool   ArgFlag    = (Argument) && (Argument[0] != '-');
   long   ErrorCount = 0;
   size_t Size       = strlen(Argv);

   for(long j = 1; Argv[j]; j++) {
      bool found = false;
//DEBUG      cout << endl;
      for(i = 0; i < GetSize(); i++) {
//DEBUG         cout << Argv[j] << ": " << Option[i] << endl;
         if (Argv[j] == Option[i]->ShortName) {
            if (Option[i]->Empty) ErrorCount += Option[i]->ChkArg(Argument, Ndx, ArgFlag);
            else found = true;
            break;
         }
      }
      if (found) {
        ErrorCount++;
        cout << "ERROR: duplicate option for -" << Argv[j] << endl;
      }
   }
   return ErrorCount;
}; // cOption_List::GetShrt(char  Argv[], char* Argument, long Ndx)

//*****************************************************************************
// NAME: ParseOption
//
// INPUTS  Str    input string option
//
// OUTPUT  Count  number of items in Value
//         Value  array of output strings
//
// RETURN  false  without error
//         true   with error (Values are unreliable)
//
// DESCRIPTION
//    Parses an option of the form '(value, ...)' and returns each value as
//    a string. 
//
//    A value is any object delimited by the set '(', ')', ',', ' '
//
// NOTES
//   1: The output array of strings must be large enough to hold all values
//
// UPDATE HISTORY
//     Date        Name         Description
// 29 Apr 2009  A. Schwarz      OIS
//*****************************************************************************

bool cOption::ParseOption(string Str, long& Count, string* Value) {
   enum eAlphabet { ill                          // Illegal character
                  , del                          // Delimeter
                  , an                           // Alphanumeric
                  };

   static const enum
   eAlphabet Alphabet[256] = {
/*
nul  soh  stx  etx  eot  enq  ack  bel  bs   hs   lf   vt   ff   cr   so   si*/
del, ill, ill, ill, ill, ill, ill, ill, ill, ill, ill, ill, ill, del, ill, ill,

/*
dle  dc1  dc2  dc3  dc4  nak  syn  etb  can  em   sub  esc  fs   gs   rs   us*/
ill, ill, ill, ill, ill, ill, ill, ill, ill, ill, ill, ill, ill, ill, ill, ill,

/*
sp    !    "    #    $    %    &    '    (    )    *    +    ,    -    .    /*/
del, ill, ill, ill, ill, ill, ill, ill, ill, del, ill, an , del, an , an,  ill,

/*
 0    1    2    3    4    5    6    7    8    9    :    ;    <    =    >    ?*/
an , an , an , an , an , an , an , an , an , an , ill, ill, ill, ill, ill, ill,

/*
 @    A    B    C    D    E    F    G    H    I    J    K    L    M    N    O*/
ill, an , an , an , an , an , an , an , an , an , an , an , an , an , an , an ,

/*
 P    Q    R    S    T    U    V    W    X    Y    Z    [    \    ]    ^    _*/
an , an , an , an , an , an , an , an , an , an , an , ill, ill, ill, ill, ill,

/*
 `    a    b    c    d    e    f    g    h    i    j    k    l    m    n    o*/
ill, an , an , an , an , an , an , an , an , an , an , an , an , an , an , an ,

/*
 p    q    r    s    t    u    v    w    x    y    z    {    |    }    ~  del*/
an , an , an , an , an , an , an , an , an , an , an , ill, ill, ill, ill, ill,

/* the remaining 128 characters are not part of the legal ASCII character set*/
   };
   long j         = 0;
   char Chr       = '\0';
   Count          = 0;

   if (Str[j] == '\0') return false;
   if (Str[j] == '=') j++;
   if (Str[j] == '(') j++;

   do {
      long k;
      for(; Str[j]; j++) if (Alphabet[Str[j]] != del) break;
      if (Alphabet[Str[j]] == ill) return true;
      if (Str[j]) {
         k = j + 1;
         for(; Str[k]; k++) if (Alphabet[Str[k]] != an) break;
      } else k = j;
      if (Alphabet[Str[k]] == ill) return true;
      Chr = Str[k + 1];
      Str[k]   = '\0';
      Value[Count++] = &Str[j];
      j        = k + 1;
   } while (Chr);
   return false;
}; // cOption::ParseOption(char * Str, long Count, string* Value 

    //----------------------------------------
    //            Debug Streams
    //----------------------------------------

ostream& operator<<(ostream& stream, cOption& Option) {
    static const string Flag[] = { "Optional Option, No Argument"
                                 , "Optional Option, Optional Argument"
                                 , "Optional Option, Required Argument"
                                 , "Optional Option, Required Argument"
                                 , "Required Option, No Argument"
                                 , "Required Option, Optional Argument"
                                 , "Required Option, Required Argument"
                                 , "Required Option, Required Argument"
                                 };
    string const Empty = (Option.Empty)? "Not Used": "Used";
    string const Val   = (Option.Val)? Option.Val: "No Value";
    stream << "<"  << Option.ShortName
           << ", " << Empty
           << ", " << Flag[(long)Option.Flag]
           << ", " << Val
           << ">";
    return stream;
}; // ostream& operator<<(ostream& stream, cOption& Option)
