// WinFrequency.cpp : Defines the entry point for the console application.
//
//******************************************************************************** 
//* File:  partition.cpp
//* Author: Arthur Schwarz
//*
//* Created on April 6, 2009, 2:54 PM
//********************************************************************************


//********************************************************************************
// PURPOSE:
//    For a given list of overlapping intervals, this program generates a set of
//    disjoint intervals with each generated interval containing a list of input
//    intervals fully contained within the generated interval. 
//
//    The output list of generated intervals can be used to construct an array in
//    which a binary search can be performed to find an the set of intervals 
//    containing a given value.
//
// DESCRIPTION:
//    Three options are provided:
//    1. Generate a set of overlapping intervals.
//    2. Construct a set of disjoint intervals.
//    3. Construct a set of fixed size partitions.
//    
//    Generate a set of overlapping intervals:
//    Using a large period (2**191) uniform random number (generator RNG )and
//    user inputs defining each of a set of input intervals, an overlapping set
//    of output intervals are generated.  
//    
//    Construct a set of disjoint intervals:
//    Given an unordered input set of intervals, an output ordered set of 
//    disjoint intervals is constructed. Each output interval is in a half-open
//    interval, I[low, high). Each constructed interval contains a list of
//    input intervals which are wholly contained in the output interval.
//    
//    Construct a set of fixed size partitions:
//    Given a user input specifying the width of each output partition, a 
//    set of output partitions is generated covering contiguously covering 
//    the total range of all input intervals [min(I.low), max(I.high)]. this
//    set of partitiions is populated with input intervals.
//
//    Overall functioning:
//
//                        --------------
//                        | Statistic  |
//                        --------------
//                               |
//                               |   o ------------------- o
//                               |   |                     |
//                         -g    v   |                     v
//    --------------      --------------               -------------
//    || Command  || =o=> || Generate || ======> o     | Interval  |
//    --------------  "   --------------         "     -------------
//                    "                          "           |
//                    "                          "           |
//                    "                          "           v                -I
//                    "                          "     --------------       -------------
//                    o =======================> o ==> || iRange   || =o==> ||  Ideal  ||
//                                                     --------------  "    -------------
//                                                                     "
//                                                                     "      -s
//                                                                     "    --------------
//                                                                     o => ||  Fixed   ||
//                                                                          --------------
//    LEGEND
//    Processes   Descriptions
//    Command     Command Line Process
//    Generate    Generate Intervals (optional)
//    iRange      Input Intervals
//    Ideal       Ideal (Variable) Partitions (optional)
//    Fixed       Fixed Size Partitions (optional)
//    ---         Data Flow
//    ===         Command Flow
//    o           Connector
//    
//    Files
//    Statistics  Statistics for Interval Generation (can be optioned  out)
//    interval    Input ranges for Parititioning (required)
//    
//
// ARCHITECTURE:
//    The following operations are performed:
//
// Step  Description                           Driver (class.name)
//    0. Process Driver.                       main
//    1. Command line processing.              cOption_List.CommandLine
//    2. Generation of input intervals.        cGenerate.Run
//    3. Input of intervals.                   ciRange_3.Read
//    4. Construction of disjoint partitions.  cIdealPartition.Run  (virtual Partition.Run)
//    5. Construdtion of fixed partitions.     cFixedPartition.Run  (virtual Partition.Run)
//    
//    The general philosophy is that if a required step fails, then the
//    following steps are not executed. In precedence order:
//    
//      Step 1 < Step 2 < Step 3 < Step 4 or Step 5
//    
//    With no dependency between Step 4 and Step 5.
//    
//    As much as possible, all reversible steps are balanced at the same
//    contextual level. That is:
//    
//      if a file is opened, then in is closed at the same contextual level.
//      if space is allocated from the heap, then it is released at the
//         same contextual level, if possible, or in the class destructor.
//    
//    All inputs are checked for validity. If anything wrong gets through
//    the input sieve, this program is in error! Except for heap overflow
//    and open failures, the program will always be successfully completed
//    if it's input is valid. There are no known theoretical or program
//    issues.
//    
//    The preferred means of passing information is through input arguments.
//    This is violated in partition creation in that the output routines,
//    Detail(), Statistics(), and Histogram() use (without attribution) data
//    created during partition creation. I view this as a defect in development.
//    Global data is eschewed as much as practical, however see global.cpp,
//    and utility.cpp. Global.h contains the I/O stream objects and utility.cpp
//    uses a file global array, Alphabet, used in several functions. This
//    file, containing main, uses file global data.
//    
//    Copius debug statements are provided, and will be output only if
//    debug (-D) is selected in the input line. One note, if debug is
//    selected then the constructed output partitions are exhaustively
//    checked for validity. In particular,  all input intervals must 
//    be represented within all partitions containing any values within
//    the interval, and all partitions must have intervals contained in
//    them which satisfy partition partitipation requirements.
//    
//    Two output files are generated:
//    <output>     Generated output
//    <output>.log Input listings, error & debug messages
//    
//    The files used are:
//      cFixedPartition Processing of fixed sized partitions (create descendent)
//      cIdealPartition Processing of 'perfect' sized partitions (create descendent)
//      common          Collection of classes common to processing
//      create          Partition create base class.
//      Debug           Debug output functions
//      generate        Generate input intervals for partition processing
//      global          Global data
//      irange          Interval input processing
//      option          Command line processing
//      RngStream       Pseudo-Random Number functions (P. L'Ecuyer)
//      stdafx          Microsoft Visual Studio pre-compiled headers
//      utility         General utility functions
//      WinFrequency    Main driver and command line option functions
//    
// NOTES:
//    1. Some general development guidelines:
//       a. cDebug is the base class for most class.
//          i. This forces debugging to be developed with the class.
//         ii. <class>.DData() is required.
//        iii. ostream& operator<<(ostream& stream, <class>& obj); is present.
//       b. Objects and function definitions are kept separate. As required
//          there will be a separate use of private:, protected:, public:
//          for objects and for functions. Objects will (in general) precede
//          functions. An exemplar class template is:
//            <class> name : cDebug {
//            private:
//               <specifiers> variable;
//            protected:
//               <specifiers> variable;
//            public:
//               <specifiers> variable;
//            private:
//               <specifiers> <function>;
//            protected:
//               void DData() {  }
//               <specifiers> <function>;
//            public:
//               <specifiers> <function>;
//            }
//       c. A list containing a <class> will be named <class>_List.
//       d. A list will have an indexing operator definition:
//            <class&> operator[](long Ndx) { }
//    
//********************************************************************************
# include "../Headers/stdafx.h"

# include <cctype>
# include <cstdio>
# include <stdlib.h>
# include <cstring>
# include <ostream>
# include <sstream>
# include <ctime>
# include "../Headers/Debug.h"
# include "../Headers/common.h"
# include "../Headers/create.h"
# include "../Headers/generate.h"
# include "../Headers/global.h"
# include "../Headers/irange.h"
# include "../Headers/option.h"
# include "../Headers/utility.h"
# include "../Headers/cFixedPartition.h"
# include "../Headers/cIdealPartition.h"


using namespace std;
 
    //----------------------------------------
    //           DATA DECLARATIONS
    //----------------------------------------

cOption cOption_List::ErrorOption = cOption();

bool     IdealPartitionFlag   = false;           // Construct Ideal Partitions
bool     FixedPartitionFlag   = false;           // Construct Fixed Partitions
bool     GenerateIntervalFlag = false;           // Generate intervals
bool     ValidateFlag         = false;           // Validate Generate Partitions
double   FixedPartitionRangeSize = 0.0;          // Size of every partition range

    //----------------------------------------
    //          Forward References
    //----------------------------------------

long   DebugOption(cOption& Option);      // -D  // Precess Option
long   GenOption  (cOption& Option);      // -g  // Precess Option
long   InputOption(cOption& Option);      // -f  // Precess Option
long   HelpOption (cOption& Option);      // -h  // Precess Option
long   IdealOption(cOption& Option);      // -I  // Precess Option
long   SizeOption (cOption& Option);      // -s  // Precess Option
long   ValidateOption(cOption& Option);   // -v  // Precess Option

string GetPath(char* argv);
void   Init();                                   // Initialize data/operations
long   Open(const char* Filename);

    //----------------------------------------
    //           DATA DEFINITIONS
    //----------------------------------------
enum { eDEBUG = 0
     , eINPUT = eDEBUG + 1
     , eGEN   = eINPUT + 1
     , eHELP  = eGEN   + 1
     , eIDEAL = eHELP  + 1
     , eSIZE  = eIDEAL + 1
     };
#define INPUT 1
#define HELP  2
cOption  Options[] =  { cOption('D', "-DEBUG",    cOption::eOptional    , DebugOption)
                      , cOption('f', "-FILE",     cOption::eRequired | cOption::eArgument_Req, InputOption)
                      , cOption('g', "-GENERATE", cOption::eOptional | cOption::eArgument_Req, GenOption )
                      , cOption('h', "-HELP" ,    cOption::eOptional    , HelpOption )
                      , cOption('I', "-IDEAL",    cOption::eOptional    , IdealOption)
                      , cOption('s', "-SIZE" ,    cOption::eArgument_Req, SizeOption )
                      , cOption('v', "-VALIDATE", cOption::eOptional    , ValidateOption)
                      }; // OptionList

cOption_List OptionList(sizeof(Options)/sizeof(Options[0]), Options);
sGenerate    GenOpt;                             // Generate Interval Options

    //----------------------------------------
    //              CODE
    //----------------------------------------

//*****************************************************************************
// NAME: main
//
// INPUTS  argc   Count of options in the command line
//         argv   List of pointer to character strings representing options
//
// OUTPUT  none
//
// RETURN  0  Successful
//         1  Unsuccessful
//
// DESCRIPTION
//   Driver program; Executes each step in processing if the preceding step
//   succeeds without error.
//
// UPDATE HISTORY
//     Date        Name         Description
// 19 Apr 2009  A. Schwarz      OIS
//*****************************************************************************


int main(int argc, char* argv[])
{
   clock_t  FixedEndTime         = 0;            // Fixed Partition Creation end time
   clock_t  FixedStartTime       = 0;            // Fixed Partition Creation start time
   clock_t  GenerateEndTime      = 0;            // Generate Interval end time
   clock_t  GenerateStartTime    = 0;            // Generate Interval start time
   clock_t  IdealEndTime         = 0;            // Ideal Partition Creation end time
   clock_t  IdealStartTime       = 0;            // Ideal Partition Creation start time
   clock_t  InputEndTime         = 0;            // Interval Input end time
   clock_t  InputStartTime       = 0;            // Interval Input start time
   clock_t  TotalEndTime         = 0;            // Progrem end time
   clock_t  TotalStartTime       = 0;            // Program start time
   clock_t  ValidateFixedEndTime = 0;            // Validate Partition end time
   clock_t  ValidateFixedStartTime= 0;           // Validate Partition start time
   clock_t  ValidateIdealEndTime  = 0;           // Validate Partition end time
   clock_t  ValidateIdealStartTime= 0;           // Validate Partition start time
   bool ErrorFlag = false;

   TotalStartTime=  clock();

   Init();
   if (OptionList.CommandLine(argc, argv, File.Output)) {     // Process command line
   } else if (Options[eHELP].Empty) {
      if (!OptionList.ChkRequired()) exit(0x0002);
      if (File.Output == "") {
         cout << "ERROR: Output file must be supplied" << endl;
         return 3;
      }
      File.Basename = Basename(File.Output);
      File.Log = File.Basename + ".log";
      if (Open(Log, File.Log, ios::out)) exit(0x0001);
      if (GenerateIntervalFlag) {
         GenerateStartTime = clock();
         if (Open(Stdout, File.Interval, ios::out)) return 1;   // Open Generate output file
         if (File.Generate.length() > 0)
            if (Open(Stdin, File.Generate, ios::in)) return 1;  // Open Generate statistics file
         cGenerate Generate(GenOpt, File);
         ErrorFlag = Generate.Run();
         if (Stdin.is_open())  Stdin.close();
         if (Stdout.is_open()) Stdout.close();
         GenerateEndTime = clock();
      }
      if (!ErrorFlag) {
         if (Open(Stdin,  File.Interval, ios::in)) return 1;   // Open Range File
         if (Open(Stdout, File.Output, ios::out)) return 1;
         ciRange_3  Range(GetPath(argv[0]));
         InputStartTime = clock();
         if (Range.Read()) {                        // Read ranges
            InputEndTime = clock();
            double Bandwidth = Range.Create_Bounds();
            if (IdealPartitionFlag) {
               IdealStartTime = clock();
               cIdealPartition Partition(Options[eINPUT].Val, "Ideal", Bandwidth, Range.GetBounds());
               Partition.Run();
               Partition.Statistics();
               Partition.Detail();
               Partition.Histogram();
               IdealEndTime = clock();
               ValidateIdealStartTime = IdealEndTime;
               if (ValidateFlag) Partition.Validate();
               ValidateIdealEndTime = clock();
            }
            if (FixedPartitionFlag) {
               FixedStartTime = clock();
               cFixedPartition Partition(Options[eINPUT].Val, "Fixed", Bandwidth, Range.GetBounds());
               Partition.Run(FixedPartitionRangeSize);
               Partition.Statistics();
               Partition.Detail();
               Partition.Histogram();
               FixedEndTime = clock();
               ValidateFixedStartTime = FixedEndTime;
               if (ValidateFlag) Partition.Validate();
               ValidateFixedEndTime = clock();
            }
         }
      } else return 1;
   }
   if (Stdout.is_open()) Stdout.close();
   if (Stdin.is_open())  Stdin.close();
   if (Log.is_open())    Log.close();

   TotalEndTime = clock();

   cout << "Total    Execution Time: " << fixed << setprecision(3) << diffclock(TotalEndTime, TotalStartTime) << endl
        << "Total    Generate Time:  " << fixed << setprecision(3) << diffclock(GenerateEndTime, GenerateStartTime) << endl
        << "Total    Input Time:     " << fixed << setprecision(3) << diffclock(InputEndTime, InputStartTime) << endl
        << "Total    Ideal Time:     " << fixed << setprecision(3) << diffclock(IdealEndTime, IdealStartTime) << endl
        << "Validate Ideal Time:     " << fixed << setprecision(3) << diffclock(ValidateIdealEndTime, ValidateIdealStartTime) << endl
        << "Total    Fixed Time:     " << fixed << setprecision(3) << diffclock(FixedEndTime, FixedStartTime) << endl
        << "Validate Fixed Time:     " << fixed << setprecision(3) << diffclock(ValidateFixedEndTime, ValidateFixedStartTime) << endl;
   return 0;
}; // main(int argc, char* argv[])

    //----------------------------------------
    //           OPTION FUNCTIONS
    //----------------------------------------

//*****************************************************************************
// NAME: DebugOption
//
// INPUTS  Option  Definition of Command Line parameterization
//
// OUTPUT  none
//
// RETURN  0  No error detected
//
// DESCRIPTION
//    Turn Debug on.
//
// UPDATE HISTORY
//     Date        Name         Description
// 19 Apr 2009  A. Schwarz      OIS
//*****************************************************************************

long  DebugOption(cOption& Option) {     // -D
    cDebug::setFlag(true);
    return 0;
}; // DebugOption(cOption& Option)

//*****************************************************************************
// NAME: GenOption
//
// INPUTS  Option  Definition of Command Line parameterization
//
// OUTPUT  none
//
// RETURN  0 No errors found
//         1 Error found
//
// DESCRIPTION
//    The user inputs the count of intervals to generate and optionally
//    provides the seed to use for the Uniform Random Generator. The seed
//    must be less than 4,294,967,087 and positive, the count must be 
//    positive.
//
//    The format of the input is:
//    -g COUNT
//    -g (COUNT,SEED)
//    -g (COUNT,FILENAME)
//    -g (COUNT,SEED,WIDTH)
//    -g (COUNT,SEED,FILENAME)
//    --GENERATE=COUNT
//    --GENERATE=(COUNT,SEED)
//    --GENERATE=(COUNT,FILENAME)
//    --GENERATE=(COUNT,SEED,WIDTH)
//    --GENERATE (COUNT,SEED,FILENAME)
//
// UPDATE HISTORY
//     Date        Name         Description
// 29 Apr 2009  A. Schwarz      OIS
//*****************************************************************************

long   GenOption(cOption& Option) {        // -g 
   static  const
   long    MaxSeed = 4294967087;
   static  const
   double  DefaultWidth = 250.0;
   string  Val(Option.Val + '\0');
   string  Value[10];                            // Input options
   long    Count;                                // Number of options
   
   if(Option.ParseOption(Val, Count, Value)) return 1;

   GenOpt.Width = DefaultWidth;

   if (!Count) {
      cout << "ERROR: Required arguments missing for -g '" << Option.Val << "'" << endl;
      return 1;
   } else if (Count > 3) {
      cout << "ERROR: Too many arguments for -g '" << Option.Val << "'" << endl;
      return 1;
   }

   // Process input options

   istringstream stream(Value[0]);
   stream >> GenOpt.Count;                       // Find COUNT
   if (GenOpt.Count <= 0) {
      cout << "ERROR: Count must be > 0 for -g '" << Option.Val << "'" << endl;
      return 1;
   }
   if (Count >= 2) {                             // Find SEED or FILENAME
      if ( (Value[1][0] == '+') || (Value[1][0] == '-') || isdigit(Value[1][0])) {
         istringstream stream(Value[1]);
         stream >> GenOpt.Seed;
         if (GenOpt.Seed <= 0) {
            cout << "ERROR: Seed must be > 0 for -g " << Value[1] << endl;
            return 1;
         } else if (GenOpt.Seed > MaxSeed) {
            cout << "ERROR: Seed must be <= 4,294,967,087 for -g " << Value[1] << endl;
            return 1;
         }
      } else {
         FILE* file = fopen(Value[1].c_str(), "r");
         if (file) {
            File.Generate = Value[1];
            fclose(file); 
         } else {
            cout << "ERROR: Unable to open generate file for -g '" << Value[1] << "'" << endl;
            return 1;
         }
         if (Count == 3) {
            cout << "ERROR: too many arguments for -g '" << Value[1] << "'" << endl;
            return 1;
         }
      }
   }

   if (Count == 3) {                             // Find WIDTH or FILENAME
      if (!isalpha(Value[2][0])) {
         try {
            istringstream stream(Value[2]);
            stream >> GenOpt.Width;
            if (GenOpt.Width <= 0) {
               cout << "ERROR: Width must be > 0 for -g " << Value[2] << endl;
               return 1;
            }
         } catch (...) {
            cout << "ERROR: Illegal Width for -g " << Value[2] << endl;
            return 1;
         }
      } else {
         FILE* file = fopen(Value[2].c_str(), "r");
         if (file) {
            File.Generate = Value[2];
            fclose(file);
         } else {
            cout << "ERROR: Unable to open generate file for -g '" << Value[1] << "'" << endl;
            return 1;
         }
      }
   }

   GenerateIntervalFlag = true;                  // Generate intervals
   return 0;
}; // GenOption(cOption& Option)


//*****************************************************************************
// NAME: InputOption
//
// INPUTS  Option  Definition of Command Line parameterization
//
// OUTPUT  none
//
// RETURN  0 No errors found
//         1 Error found
//
// DESCRIPTION
//    The user inputs the full path name of the file to be opened as an 
//    input argument. Opening the file is defered.
//
// UPDATE HISTORY
//     Date        Name         Description
//  8 May 2009  A. Schwarz      OIS
//*****************************************************************************

long  InputOption(cOption& Option) {     // -f
    File.Interval = Option.Val;
    return 0;
}; // Input(cOption& Option)

//*****************************************************************************
// NAME: HelpOption
//
// INPUTS  Option  Definition of Command Line parameterization
//
// OUTPUT  none
//
// RETURN  0  No error detected
//
// DESCRIPTION
//    The program help file is output
//
// UPDATE HISTORY
//     Date        Name         Description
// 19 Apr 2009  A. Schwarz      OIS
//*****************************************************************************

long  HelpOption (cOption& Option) {     // -h
    Option.Empty = false;
    cout << " NAME:"
         << "    partition - Partition ranges into disjoint regions"                     << endl
         << ""                                                                           << endl
         << " SYNOPSIS:"                                                                 << endl
         << "    partition [-D] [-g COUNT] -f FILENAME [-s SIZE] [-h] [-I] FILENAME"          << endl
         << "    partition [-D] [-g (COUNT,SEED)]          -f FILENAME [-s SIZE] [-h] [-I] [-v] FILENAME" << endl
         << "    partition [-D] [-g (COUNT,SEED,WIDTH)]    -f FILENAME [-s SIZE] [-h] [-I] [-v] FILENAME" << endl
         << "    partition [-D] [-g (COUNT,SEED,FILENAME)] -f FILENAME [-s SIZE] [-h] [-I] [-v] FILENAME" << endl
         << "    partition [-D] [-g (COUNT,FILENAME)]      -f FILENAME [-s SIZE] [-h] [-I] [-v] FILENAME" << endl
         << ""                                                                           << endl
         << " DESCRIPTION:"                                                              << endl
         << "   Partition accepts a list of overlapping intervals and constructs a"      << endl
         << "   list of disjoint ranges and/or a list of fixed size ranges."             << endl
         << ""                                                                           << endl
         << " OPTIONS:"                                                                  << endl
         << "   -D, --DEBUG     Output Debug information"                                << endl
         << "   -f, --FILE      Input filename of ranges"                                << endl
         << "   -g, --GENERATE  Generate a filename of ranges"                           << endl
         << "   -h, --HELP      Help"                                                    << endl
         << "   -I, --IDEAL     Output ideal (disjoint) ranges"                          << endl
         << "   -s, --SIZE      Size of fixed sized ranges (and output ranges)"          << endl
         << "   -v, --VALIDATE  Check generated partitions for correctness"              << endl
         << ""                                                                           << endl
         << " GENERATE DESCRIPTION"                                                      << endl
         << " Generate the input file (-f FILENAME) using a pseudo-random number"        << endl
         << " generator. The generate function produces COUNT items, each item a valid"  << endl
         << " interval. If an input generate file is not used then the following defaults" << endl
         << " are used for the output:"                                                  << endl
         << "   [0.00, 10,000.00) is the [lowest,largest) bound"                         << endl
         << "   [x, x+250)        is the interval for all items output"                  << endl
         << ""                                                                           << endl
         << " That is, no output bound is less than 0.00 or greater than 10,000.00."     << endl
         << " The WIDTH specifies the maximum WIDTH of any interval. The actual WIDTH"   << endl
         << " is generated using a pseudo-random number generator (RNG). Assuming:"      << endl
         << "    [x, y] is the range of interval values"                                 << endl
         << "    WIDTH  is the maximum width, and"                                       << endl
         << "    x <= start <= y where 'start' is some point within the range"           << endl
         << " then"                                                                      << endl
         << "    [start, start + RNG()*(y - x)] is the generated interval"               << endl
         << ""                                                                           << endl
         << " GENERATE OPTIONS"                                                          << endl
         << "  1. COUNT is required"                                                     << endl
         << "  2. COUNT, SEED and WIDTH must be positive numbers"                        << endl
         << "  3. COUNT and SEED must be integers."                                      << endl
         << "  4. SEED must be 0 < SEED <= 4,294,967,087"                                << endl
         << "  5. SEED  default is '12345'."                                             << endl
         << "  6. WIDTH can be a floating point number."                                 << endl
         << "  7. WIDTH default is '250' or 12% of total range."                         << endl
         << "  8. WIDTH must be less than 1/2 the maximum range (upper bound - lowest bound)"<< endl
         << "  9. FILENAME must not be a number. '4' or '4.1' are not allowed."          << endl
          << ""                                                                          << endl
         << " GENERATE FILE"                                                             << endl
         << " Subject to the restrictions given in GENERATE OPTIONS, a generate input"   << endl
         << " file can specify the start bound and WIDTH of a set of intervals."         << endl
         << ""                                                                           << endl
         << "   # Comment (to end of line)"                                              << endl
         << "   lower bound, upper bound, WIDTH, percent"                                << endl
         << "   |            |            |      |"                                      << endl
         << "   |            |            |      o percent of output"                    << endl
         << "   |            |            |"                                             << endl
         << "   |            |            o maximum WIDTH of interval"                   << endl
         << "   |            |"                                                          << endl
         << "   |            o upper bound of range"                                     << endl
         << "   |"                                                                       << endl
         << "   o lower bound of range"                                                  << endl
         << ""                                                                           << endl
         << " The interval containing all intervals is"                                  << endl
         << "   [min(lower bounds), max(upper bounds)]"                                  << endl
         << ""                                                                           << endl
         << " The WIDTH is optional and can be omitted by leaving the field empty."      << endl
         << " If WIDTH is not provided then the WIDTH defaults to the interval width"    << endl
         << " (upper bound - lower bound). The WIDTH can never extend the interval upper"<< endl
         << " bound and is limited to the min(WIDTH, (upper bound - start bound))"       << endl
         << ""                                                                           << endl
         << " The bounds can be given in any order. The sum of all percents must be"     << endl
         << " less than or equal to 100%. If the sum of all percents is less than 100%"  << endl
         << " then the residual percentage is uniformly assigned to intervals which"     << endl
         << " have not been specified."                                                  << endl
         << ""                                                                           << endl
         << " Ranges in intervals can not overlap."                                      << endl
         << ""                                                                           << endl
         << " EXAMPLE"                                                                   << endl
         << " 0.0,  10.0,, 20.0  # [0.0, 10.0) 20% of output"                            << endl
         << " 3.0,  15.0,, 33.0  $ ILLEGAL, overlapping ranges"                          << endl
         << ""                                                                           << endl
         << "                    # WIDTH defaults to (upper bound - lower bound)"        << endl
         << " 1.0,  10.0,, 20.0  # [ 1.0, 10.0) 20% of output"                           << endl
         << " 11.0, 15.0,, 10.0  # [11.0, 15.0) 10% of output"                           << endl
         << " 50.3, 70,,   22    # [50.3, 70.0) 22% of output"                           << endl
         << "                    # The remaining 48% of output are uniformly"            << endl
         << "                    # distributed in the following ranges"                  << endl
         << " 10,   11,,    1.32 # [10.0, < 11.0) approximately  1.32%"                  << endl
         << " 15,   50.3,, 46.68 # [15.3, < 50.3) approximately 46.68%"                  << endl
         << ""                                                                           << endl
         << " INPUT FILE FORMAT:"                                                        << endl
         << "   # Comment (to end of line)"                                              << endl
         << "   lower bound, upper bound, identifier"                                    << endl
         << "   |            |            |"                                             << endl
         << "   |            |            o unique identifier"                           << endl
         << "   |            |"                                                          << endl
         << "   |            o upper bound of range (floating point or integer)"         << endl
         << "   |"                                                                       << endl
         << "   o lower bound of range (floating point or integer)"                      << endl
         << ""                                                                           << endl
         << " OUTPUT FORMAT:"                                                            << endl
         << "  Section  I Statistics"                                                    << endl
         << "  Section II Partitions"                                                    << endl
         << ""                                                                           << endl
         << " Section I"                                                                 << endl
         << "   Statistics for \"Ideal or Fixed\" Partition"                             << endl
         << "   Date:          Date Output File createc"                                 << endl
         << "   Time:          Time Output File created"                                 << endl
         << "   Elapsed Time:  Time to create partition and generate statistics"         << endl
         << ""                                                                           << endl
         << "   Summary results for <input file name>"                                   << endl
         << "     Total Input Intervals:         Number of Intervals in the input file"  << endl
         << "     Total Partitions:              Number of constructed partitions"       << endl
         << "     Average Entries/Partition:     Average number of references / partition" << endl
         << "     Maximum Entries/Partition:     Maximum number of references / partition" << endl
         << "     Minimum Entries/Partition:     Minimum number of references / partition" << endl
         << "     Percent Used Space:            Used Interval Space / Total Interval Space"<< endl
         << "     Percent Wasted Space:          1 - Percent Used Space"                 << endl
         << ""                                                                           << endl
         << "  Section II"                                                               << endl
         << "   Partition Listing for Ideal partition"                                   << endl
         << "   [ lower bound, upper bound) entry, entry, ..."                           << endl
         << ""                                                                           << endl
         << "   where the region is closed on the left and open on the right"            << endl
         << "   and the identifier list contains all range identifiers which"            << endl
         << "   are fully contained in the range (Ideal Partitions) or are"              << endl
         << "   partially contained in the range (Fixed Partitions)."                    << endl
         << ""                                                                           << endl
         << "   Three digits to the right of the decimal point are output. This leads"   << endl
         << "   to partitions in which the bounds displayed appear to be the same,"      << endl
         << "   that is, more than three digits to the right of the decimal point are"   << endl
         << "   needed to show uniqueness."                                              << endl
         << ""                                                                           << endl
         << "   If the end of one input interval is the same as the beginning of the"    << endl
         << "   next interval, then the output is displayed as:"                         << endl
         << "     [x,  y)  x  lower bound, y upper bound on the open interval"           << endl
         << "     [y,  y)  y  the upper bound and lower bound on the open interval"      << endl
         << "     [y1, z)  y1 the lower bound, z the upper bound on the open interval"   << endl
         << ""                                                                           << endl
         << "   and x <  y   on [x,  y)"                                                 << endl
         << "       y == y   on [y,  y) and [x, y]"                                      << endl
         << "       y <  y1  on [y1, z)"                                                 << endl
         << "       y1 < z   on [y1, z)"                                                 << endl
         << ""                                                                           << endl
         << "   and y1 will be constructed to differ from y in the 54nd bit of the"      << endl
         << "   mantissa in an IEEE 754 formated double precision floating point"        << endl
         << "   number."                                                                 << endl
         << "     y1 = y1 + delta "                                                      << endl
         << "       where delta is scaled to be '1' in the last bit of the mantissa"     << endl
         << ""                                                                           << endl
         << "   The interval [y, y) is a singular point in the partition in that it"     << endl
         << "   has zero-length (on the open interval) but a real existence. When"       << endl
         << "   constructing an order table for sorting it is important to note that"    << endl
         << "   x, y, and y1 are distinct (x < y < y1) motivating searches to be made"   << endl
         << "   on the lower bound in preference to the upper bound."                    << endl
         << ""                                                                           << endl
         << " NOTES"                                                                     << endl
         << "   1. The output FILENAME is required."                                     << endl
         << "   2. The input FILENAME is required"                                       << endl
         << "   3. For long form options the following is legal"                         << endl
         << "      --LONG_FORM=<argument>"                                               << endl
         << "      --LONG_FORM <argument>"                                               << endl
         << "   4. For short form options the following is legal"                        << endl
         << "      -x=<argument>"                                                        << endl
         << "      -x <argument>"                                                        << endl
         << "   5. Options can be given in any order"                                    << endl
         << "   6. For short form options, options can be combinded"                     << endl
         << "      -D -I and -DI are treated the same"                                   << endl
         << "   7. For the combined form of short form arguments at most one"            << endl
         << "      option can have an argument"                                          << endl
         << "      -fD <argument>   is legal"                                            << endl
         << "      -fDs <argument>  is not legal"                                        << endl
         << "      -fDs <argument> <argument> is not legal"                              << endl
         << "   8. The GENERATE filename must not be a number"                           << endl
         << endl;
    return 0;   
}; // Help (cOption& Option)

//*****************************************************************************
// NAME: IdealOption
//
// INPUTS  Option  Definition of Command Line parameterization
//
// OUTPUT  none
//
// RETURN  0  No error detected
//
// DESCRIPTION
//    Ideal Partition processing is enabled.
//
// UPDATE HISTORY
//     Date        Name         Description
// 19 Apr 2009  A. Schwarz      OIS
//*****************************************************************************

long  IdealOption(cOption& Option) {     // -I
    IdealPartitionFlag = true;
    return 0;
}; // Ideal(cOption& Option)

//*****************************************************************************
// NAME: SizeOption
//
// INPUTS  Option  Definition of Command Line parameterization
//
// OUTPUT  none
//
// RETURN  0  No error detected
//         1  Input size not legal
//
// DESCRIPTION
//    Fixed partition processing is enabled and the partition size is converted
//    from Ascii to binary.
//
// UPDATE HISTORY
//     Date        Name         Description
// 19 Apr 2009  A. Schwarz      OIS
//*****************************************************************************

long  SizeOption (cOption& Option) {     // -s
    long ErrorCount    = 0;
    FixedPartitionFlag = true;
    istringstream stream(Option.Val); 
    double Size = -1.0;
    try{
        stream >> Size;
        FixedPartitionRangeSize = Size;
//DEBUG        cout << "Size = " << Size << " Val "   << Option.Val << endl;
        if (FixedPartitionRangeSize <= 0.0) {
           ErrorCount         = 1;
           FixedPartitionFlag = false;
           cout << "ERROR: Partition size not > 0.0 [" << Size << "]" << endl;
        }
    } catch (...) {
        ErrorCount         = 1;
        FixedPartitionFlag = false;
        cout << "ERROR: -s <size> not a number '" << Option.Val << "'" << endl;
    }
    return ErrorCount;
}; // Size (cOption& Option)


//*****************************************************************************
// NAME: ValidateOption
//
// INPUTS  Option  Definition of Command Line parameterization
//
// OUTPUT  none
//
// RETURN  0  No error detected
//         1  Error detected
//
// DESCRIPTION
//    Enables the validate flag to check the validity of the generated
//    partitions.
//
// UPDATE HISTORY
//     Date        Name         Description
//  7 Jul 2009  A. Schwarz      OIS
//*****************************************************************************

long   ValidateOption(cOption& Option){   // -v  // Precess Option
   ValidateFlag = true;
   return 0;
}; // ValidateOption(cOption& Option)

    //----------------------------------------
    //           GENERAL FUNCTIONS
    //----------------------------------------

//*****************************************************************************
// NAME: GetPath
//
// INPUTS  Option  Definition of Command Line parameterization
//
// OUTPUT  none
//
// RETURN  Path string (extracted from dommand line)
//
// DESCRIPTION
//
// UPDATE HISTORY
//     Date        Name         Description
// 19 Apr 2009  A. Schwarz      OIS
//*****************************************************************************

string GetPath(char* argv) {
   size_t i;
   for(i = strlen(argv)
      ;(i > 0) && ((argv[i] != '\\') && (argv[i] != '/'))
      ; i--);
   argv[i] = '\0';
   for(i = 0; argv[i]; i++) if (argv[i] == '\\') { argv[i] = '/'; }

   string Path(argv);
   return Path;
}; // GetPath(char* argv)

//*****************************************************************************
// NAME: Init
//
// INPUTS  none
//
// OUTPUT  none
//
// RETURN  none
//
// DESCRIPTION
//    Placeholder to perform process initialization
//
// UPDATE HISTORY
//     Date        Name         Description
// 19 Apr 2009  A. Schwarz      OIS
//*****************************************************************************

void Init() {                                    // Initialize data/operations
}; // Init()

