# include "../Headers/stdafx.h"

# include <ios>
# include <fstream>
# include <sstream>
# include "../Headers/Debug.h"
# include "../Headers/utility.h"

//********************************************
//             Shared Data
//********************************************

sFiles       File;                              // Filenames

//********************************************
//             Local Data
//********************************************

enum eAlphabet { ill                             // Illegal character
               , del                             // Delimiter ('(', ')', ',', <hul>, <cr>)
               , num                             // Numeric [0-9]
               , sgn                             // Sign ('+', '-')
               , xpo                             // Exponent ('E', 'e')
               , dec                             // Decimal point ('.')
               , spc                             // Space (' ')
               };

static const enum
eAlphabet Alphabet[256] = {
/*
nul  soh  stx  etx  eot  enq  ack  bel  bs   hs   lf   vt   ff   cr   so   si*/
del, ill, ill, ill, ill, ill, ill, ill, ill, ill, ill, ill, ill, del, ill, ill,

/*
dle  dc1  dc2  dc3  dc4  nak  syn  etb  can  em   sub  esc  fs   gs   rs   us*/
ill, ill, ill, ill, ill, ill, ill, ill, ill, ill, ill, ill, ill, ill, ill, ill,

/*
sp    !    "    #    $    %    &    '    (    )    *    +    ,    -    .    /*/
spc, ill, ill, ill, ill, ill, ill, ill, del, del, ill, sgn, del, sgn, del, ill,

/*
 0    1    2    3    4    5    6    7    8    9    :    ;    <    =    >    ?*/
num, num, num, num, num, num, num, num, num, num, ill, ill, ill, ill, ill, ill,

/*
 @    A    B    C    D    E    F    G    H    I    J    K    L    M    N    O*/
ill, ill, ill, ill, ill, xpo, ill, ill, ill, ill, ill, ill, ill, ill, ill, ill,

/*
 P    Q    R    S    T    U    V    W    X    Y    Z    [    \    ]    ^    _*/
ill, ill, ill, ill, ill, ill, ill, ill, ill, ill, ill, ill, ill, ill, ill, ill,

/*
 `    a    b    c    d    e    f    g    h    i    j    k    l    m    n    o*/
ill, ill, ill, ill, ill, xpo, ill, ill, ill, ill, ill, ill, ill, ill, ill, ill,

/*
 p    q    r    s    t    u    v    w    x    y    z    {    |    }    ~  eol*/
ill, ill, ill, ill, ill, ill, ill, ill, ill, ill, ill, ill, ill, ill, ill, ill,

/* the remaining 128 characters are not part of the legal ASCII character set*/
   };

//*****************************************************************************
// NAME: Basename
//
// INPUTS  Filename  Path and filename
//
// OUTPUT  none
//
// RETURN  basename  filename with Path removed
//
// DESCRIPTION
//   Remove the path preceding the filename.  
//
// UPDATE HISTORY
//     Date        Name         Description
// 27 May 2009  A. Schwarz      OIS
//*****************************************************************************

string Basename(string Filename) {
   long     i    = 0;
   long     Size = Filename.length();
   for(;;) {
      for(; i < Size; i++) if (Filename[i] == '.') break; // find first dot
      if (i >= Size) return Filename.substr(0, Size);

      long j = i + 1;
      for(; j < Size; j++) if (Filename[j] == '/') break; // find '/'
      if (j >= Size) return Filename.substr(0, Size - 1);
      i = j + 1;
   }
   return "";
}; // Basename(string Filename)

//*****************************************************************************
// NAME: diffclock
//
// INPUTS  End       Ending clock time
//         Start     Starting clock time
//
// OUTPUT  none
//
// RETURN  double    Seconds and fractional seconds as a double
//
// DESCRIPTION
//    Converts the difference between start and end times into seconds.
//
// NOTE:
//
// UPDATE HISTORY
//     Date        Name         Description
// 10 Jun 2009  A. Schwarz      OIS
//*****************************************************************************

double diffclock(clock_t End, clock_t Start) { 
   return double(End - Start) / CLOCKS_PER_SEC; 
}; // diffclock(clock_t End, clock_t Start)

//*****************************************************************************
// NAME: is_Num
//
// INPUTS  Str       String containing a number
//
// OUTPUT  Start     First non-blank character in number
//         Stop      First non-numeric character
//
// RETURN  true      number found
//         false     number not found
//
// DESCRIPTION
//   Determine if the first token in a string is a number. A legal number is
//   an integer with or without a leading sign, or a floating point number
//   with or without a leading sign with or without a signed xpoonent.
//
//   The algorithm:
//    1. Removes leading blanks.
//    2. Skips over any leading sign.
//    3. Accepts the integral part of a number.
//    4. Accepts the decimal point in a floating point number.
//    5. Accepts the fractional part of a number
//    6. Accepts the exponsent sign.
//    7. Accepts the exponent.
//
//    The algorithm terminates on the first character which is not expected
//    or not a number. If the number is vacuous or the terminating character
//    is not a delimeter, the function returns 'false'.
//
// NOTE:
//  1. The following are numbers.
//      '+1'
//      ' 1'
//      '123.456'
//      '.456'
//      '123.456e-123
//      '1e23
//  2. On exit, Stop is the location of the character which terminates the
//     number (one more than the last valid numeric character).
//
// UPDATE HISTORY
//     Date        Name         Description
// 27 May 2009  A. Schwarz      OIS
//*****************************************************************************

bool is_Num(const string& Str, long& Start, long& Stop) {
   long  i = 0;
   long  j = 0;

   cDebug::DebugEnter("is_Num(" + Str + ")");

   while(Str[j] == ' ') j++;                     // Strip leading blanks
   Start = j;                                    // Number Start

   if (Alphabet[Str[j]] == sgn) j++;             // Skip leading '+' or '-'
   i = j;                                        // Start of number w/o sign
   while(Alphabet[Str[j]] == num) j++;           // Skip leading integer
   if (Str[j] == '.') j++;                       // Skip decimal point
   while(Alphabet[Str[j]] == num) j++;           // Skip fractional part
   if (Alphabet[Str[j]] == xpo) {                // Parse exponent
     j++;
     if (Alphabet[Str[j]] == sgn) j++;           // Skip '+' or '- sign
     while(Alphabet[Str[j]] == num) j++;         // Skip exponent
   }
   Stop = j;
   if (cDebug::is_Debug()) {
      ostringstream stream(ios::out);
      stream << "[" << Start << ", " << Stop << "] 0x"
                << hex << (unsigned)Str[Stop];
      cDebug::DebugData(stream.str());
   }

   cDebug::DebugExit("is_Num(" + Str + ")");
   if ((Stop - i) <= 0)            return false;
   if (Alphabet[Str[Stop]] != del) return false;
   return true;
}; // is_Num(const string& Str, long& Start, long& Stop)

//*****************************************************************************
// NAME: Parse
//
// INPUTS  Str       String containing a number
//
// OUTPUT  Start     First non-blank character in number
//         Stop      First non-numeric character
//
// RETURN  true      number found
//         false     number not found
//
// DESCRIPTION
//   If the input string is not a number, determine the reason for failure
//   and issue an appropriate diagnostic message.
//
//
// UPDATE HISTORY
//     Date        Name         Description
// 27 May 2009  A. Schwarz      OIS
//*****************************************************************************

bool   Parse(const string& Str, long& Start, long& Stop) { // Parse input line
   bool  ErrorFlag = true;

   cDebug::DebugEnter("Parse");

   if (is_Num(Str, Start, Stop)) ErrorFlag = false;
   else {
      ostringstream stream(ios::out);
      switch ( Alphabet[Str[Stop]] ) {            // Why did we stop?
      case ill:                                  // Illegal character
         stream << "ERROR: Illegal alphanumeric character '" << Str[Stop] << "'" << endl;
         break;
      case del:                                  // End-of-Field/End-of-Line
         break;
      case num:                                  // Numeric [0-9]
         stream << "INTERNAL ERROR: numeric detected as field terminator '" << Str[Stop] << "'" << endl;
         break;
      case sgn:                                  // Sign ('+', '-')
         stream << "ERROR: too many signs '" << Str[Stop] << "' in number" << endl;
         break;
      case xpo:                                  // Exponent ('E', 'e')
         stream << "ERROR: too many exponents symbols '" << Str[Stop] << "' in  number" << endl;
         break;
      case spc:                                  // Space (' ')
         ErrorFlag = false;
         break;
      default:
         stream << "INTERNAL ERROR: Illegal switch value '0x" << hex << Alphabet[Str[Stop]] << "'"<< endl;
         break;
      }
      if (ErrorFlag) {
         if (Log.is_open()) Log  << stream.str();
         else               cout << stream.str();
      }
   }

   cDebug::DebugData(Str.substr(0,Stop));
   cDebug::DebugExit("Parse");
   return ErrorFlag;
}; // Parse(string Str)

//*****************************************************************************
// NAME: StripComments
//
// INPUTS  Line  input line
//
// OUTPUT  none
//
// RETURN  string with comments removed, or
//         null string ("")
//
// DESCRIPTION
//    If the entire line is a comment or blank return "".
//    Embedded comments are removed from the return string.
//
// NOTE
//    1: Sneaky-pete coding is used to pass the input line back to
//       the I/O user through Range.Line. This allows the original
//       input line to be output and the ranges to be recovered.
//
// UPDATE HISTORY
//     Date        Name         Description
// 11 May 2009  A. Schwarz      OIS
//*****************************************************************************

string StripComments(const string& Line) {
   cDebug::DebugEnter("StripComments");
   long   gcount = Line.length();
   long   Last   = 0;
   long   Ndx    = 0;
   string Output("");

   if (gcount > 0) {
      while(Line[Ndx++] == ' ');
      Ndx--;
      if (Line[Ndx] != '#') {
         for(Last = Ndx + 1; Last < gcount; Last++) // Find extent of live data
            if (Line[Last] == '#') break;
         Output = Line.substr(Ndx, (Last - Ndx + 1));
      }
   }
   cDebug::DebugData(Output);
   cDebug::DebugExit("StripComments");
   return Output;
}; // cGenerate::StripComments(const string& Line)

//*****************************************************************************
// NAME: Open
//
// INPUTS  stream   Input stream to open
//         Filename Name of file to be opened
//         Mode     ios mode
//
// OUTPUT  none
//
// RETURN  0 No errors found
//         1 Error found
//
// DESCRIPTION
//    Close the input stream if open, clear any stream flags and then attempt
//    to open the string for the new file. Issue a diagnostic message for any
//    failures detected during the open.
//
// UPDATE HISTORY
//     Date        Name         Description
// 15 May 2009  A. Schwarz      OIS
//*****************************************************************************

bool   Open(ifstream& stream, const string& Filename, ios_base::openmode Mode) {
   bool ErrorFlag  = true;
   if (stream.is_open()) stream.close();
   stream.clear();
   try {
     stream.open(Filename.c_str(), Mode);
     stream.exceptions(ios_base::badbit | ios_base::failbit);
     if (stream.fail()) {
       cout << "ERROR: " << Filename << " had fatal errors" << endl;
     } else if (!stream.is_open()) {
       cout << "ERROR: Unable to open " << Filename << endl;
     } else if (stream.eof()) {
       cout << "ERROR: File is empty " << Filename << endl;
     } else ErrorFlag  = false;
   } catch (exception& ex ) {
      cout << "ERROR: " << ex.what() << endl;
   }
   stream.exceptions(ifstream::goodbit);
   return ErrorFlag;
}; // Open(ofstream& stream, const string& Filename, ios_base::open_mode Mode)

//*****************************************************************************
// NAME: Open
//
// INPUTS  stream   Output stream to open
//         Filename Name of file to be opened
//         Mode     ios mode
//
// OUTPUT  none
//
// RETURN  0 No errors found
//         1 Error found
//
// DESCRIPTION
//    Close the input stream if open, clear any stream flags and then attempt
//    to open the string for the new file. Issue a diagnostic message for any
//    failures detected during the open.
//
// UPDATE HISTORY
//     Date        Name         Description
// 27 May 2009  A. Schwarz      OIS
//*****************************************************************************

bool   Open(ofstream& stream, const string& Filename, ios_base::openmode Mode) {
   bool ErrorFlag  = false;
   if (stream.is_open()) stream.close();
   stream.open(Filename.c_str(), Mode);
   if (stream.fail()) {
      ErrorFlag = true;
      cout << "ERROR: " << Filename << " had fatal errors" << endl;
   } if (!stream.is_open()) {
      ErrorFlag = true;
      cout << "ERROR: Unable to open " << Filename << endl;
   } 
   return ErrorFlag;
}; // Open(ofstream& stream, const string& Filename, ios_base::open_mode Mode)
